#include <libld/ld_bfd.h>
#include <macros.h>
#include <assert.h>

#include <libcrt/string.h>
#include <ar/ar.h>

inline static size_t ar_atoi(char *s)
{
  size_t res = 0;

  CONCEPTION_ASSERT(s != NULL);

  for (; (*s >= '0') && (*s <= '9') ; s++)
    res = 10*res + (*s - '0');

  return res;
}

int ar_load (struct ld_module_struct *module)
{
  struct ld_module_struct *cur_module = module;
  char   *image_start;
  addr_t  offset       = 0;
  char   *string_table = NULL;

  ASSERT_FATAL(module && module->raw_image_start
	       && (module->raw_image_size > SARMAG));

  image_start  = (char *)module->raw_image_start;
  image_start += SARMAG;

  /* Explanation for the "-1" : offset must be aligned on a 2 bytes
     boundary, so it may be later incremented. We must take this into
     account. Note that we can't lose any data, since each data is
     always more than sizeof(struct ar_hdr) > 1 bytes long */
  while (offset < module->raw_image_size - SARMAG -1) {
    char *c;
    struct ar_hdr *mb_hdr;
    size_t         mb_size;
    char          *mb_data;
    char          *mb_name;

    /* Aligns to an even boundary */
    offset = ALIGN_SUP(offset, 2);

    /* Get the ar_hdr header */
    mb_hdr  = (struct ar_hdr*)(image_start + offset);
    mb_data = image_start + offset + sizeof(struct ar_hdr);
    
    /* Set the temporary name */
    mb_name = mb_hdr->ar_name;

    /* Checks validity of header */
    if (strncmp(mb_hdr->ar_fmag, ARFMAG, 2)) {
      WARNING("Invalid header in archive %s at offset 0x%x. Ignoring it.\n",
	      module->name, offset);
      return -1;
    }

    /* Evaluates the size of this archive member */
    mb_size = ar_atoi(mb_hdr->ar_size);

#define GOTO_NEXT_MEMBER() ({ offset += sizeof(struct ar_hdr) + mb_size ; \
                              continue; })

    /* Treats the special ar members */
    if (mb_hdr->ar_name[0] == '/') {

      /* Ignores the symbol string table */
      if (mb_hdr->ar_name[1] == ' ')
	GOTO_NEXT_MEMBER();

      /* When this member is the archive string table, remember it */
      if (mb_hdr->ar_name[1] == '/') {
	/* if a preceding string table was already found => severe error */
	ASSERT_RETURN_VAL(string_table == NULL, -1);
	string_table = mb_data;
	GOTO_NEXT_MEMBER();
      }

      /* Or else, this member name is given by an offset into the
         archive string table. This assumes that the archive string
         table has already been found */
      ASSERT_RETURN_VAL(string_table != NULL, -1);
      mb_name = string_table + ar_atoi(& mb_hdr->ar_name[1]);
    }

    /* Actually, mb_name is wrong in the archive : the terminating
       '\0' is replaced by '/' */
    for (c = mb_name; *c != '/' ;c++)
      ;
    *c = '\0';

    /* When we are here, we have mb_name, mb_data, mb_size correctly
       defined */
    DEBUG_PRINT("%s(%s): addr=0x%x, sz=%d.\n",
		module->name, mb_name, (unsigned)mb_data, mb_size);
    
    /* Adds this archive member to the list of modules */
    ASSERT_RETURN_VAL((cur_module = ld_register_module(mb_name,
						       (addr_t)mb_data,
						       mb_size, cur_module))
		      != NULL, -1);

    /* Next iteration */
    GOTO_NEXT_MEMBER();
  }

  return 0;
}
