/* cmdline.c - Copyright (c) 2002 Julien Munier & Thomas Petazzoni
   Command line parsing for the module parameters

   the Kid Operating System (http://kos.enix.org)
   @(#) $Id: cmdline.c,v 1.1.1.1 2003/04/27 12:49:48 thomas Exp $ */

#include <mm/mm.h>
#include <mod.h>
#include <libcrt/stdio.h>
#include <macros.h>
#include <assert.h>

/* The module parameters are given after the (kernel) argument in
   Grub.
   
   i.e. (kernel_path_substring) module params...

   The syntax expects a semicolon separated list of module name and
   parameters pairs. Those are separated by a colon. Every space is
   stripped, if you want to give a parameter with space you need to
   put the string inside double-quotes.

   i.e. ... wolfgang:root=/dev/hda0,type=ext2;net:hostname="foo bar"; ...
   
   WARNING: - spaces not belonging to a double-quoted string are
              squizzed !!
            - kernel_path_substring is expected NOT to contain any
              space !!
   NOTE: The final ";" is optional
*/
int parse_cmdline(char *cmdline, kernel_parameter_t *kp);

#define isspace(c) (((c) == ' ') || ((c) == '\n') || ((c) == '\t'))

int parse_cmdline(char *cmdline, kernel_parameter_t *kp)
{
  char *cur;
  char *params_begin;
  char *dest;
  int len = 0;
  int nb_modules = 0;
  int strip_spaces;
  module_parameter_t *mp;
  int i;

  /* Skip any leading space before kernel_path_substring */
  for(params_begin = cmdline ;
      (*params_begin != '\0') && (isspace(*params_begin)) ;
      params_begin++)
    continue;
  /* Skip kernel_path_substring (we assume it does not contain any space) */
  for(;
      (*params_begin != '\0') && (! isspace(*params_begin)) ;
      params_begin++)
    continue;

  /* Compute the length of the parameter string */
  strip_spaces = TRUE;
  for(cur = params_begin ; *cur != '\0' ; cur++)
    {
      if(isspace(*cur) && strip_spaces)
	continue;
      if(*cur == ':')
	nb_modules++;
      if(*cur == '"')
	strip_spaces = !strip_spaces;
      len++;
    }

  if(!strip_spaces)
    /* Non terminated "..." string */
    return -1;

  /* Allocate space for the comdline structure */
  len += nb_modules * sizeof(module_parameter_t);
  mp = vmm_page_alloc(len, NULL);
  if(mp == NULL)
    return -1;

  /* Initialise destination strings' area */
  mp[0].modname = (char *) (mp + nb_modules * sizeof(module_parameter_t));
  dest          = mp[0].modname;

  for(i = 0, cur = params_begin ; *cur != '\0'; cur++)
    {
      if(isspace(*cur) && strip_spaces)
	{
	  continue;
	}

      else if(*cur == ':')
	{
	  *dest = '\0';
	  dest++;
	  mp[i].params = dest;
	}

      else if(*cur == '"')
	{
	  strip_spaces = !strip_spaces;
	  *dest = *cur;
	  dest++;
	}
      
      else if(*cur == ';')
	{
	  if (i >= nb_modules)
	    break;

	  i++;
	  *dest = '\0';
	  dest++;
	  mp[i].modname = dest;
	}

      else
	{
	  *dest = *cur;
	  dest++;
	}
    }
  if (i+1 != nb_modules)
    return -1;
  *dest = '\0';

  kp->module_parameters             = mp;
  kp->module_parameters_size        = len;
  kp->nb_modules_parameters_entries = nb_modules;

  return 0;
}
