/*
 * Copyright (C) 2000, Julien Munier, Thomas Petazzoni
 * http://kos.enix.org
 *
 * Main code of the loader
 *
 * @(#) $Id: loader.c,v 1.1.1.1 2003/04/27 12:49:48 thomas Exp $
 */

/* These lines are supposed to be the Kos Loader recently written from
   scratch (mejj) */

#include <boot/multiboot.h>
#include <libld/ld_bfd.h>
#include <assert.h>
#include <macros.h>
#include <types.h>

#include <debug/debug.h>
#include <libcrt/string.h>
#include <libcrt/stdio.h>
#include <mm/mm.h>

int parse_cmdline(char *cmdline, kernel_parameter_t *kp);

struct {
  unsigned short size;
  unsigned long  addr;
} __attribute__((packed)) gdtr;
unsigned long long *gdt;

#define CHECK_FLAG(flags,bit)	((flags) & (1 << (bit)))
inline static void check_if_modules(multiboot_info_t *mbi)
{
  if(!CHECK_FLAG(mbi->flags, 3)) {
    FAILED_VERBOSE("\nFatal Error : No modules were loaded...");
    return;
  }
}

/* Let gcc be happy + make sure launch_bootstrap is a real call far (see 
   end of cmain), especially with gcc 3.1 */
extern void launch_bootstrap(void);
void launch_bootstrap()
{
  int res;
  kernel_startup_routine_t kernel_bootstrap;
  exported_symbol_t* init_sym;
  struct ld_module_struct* init_module;
  int init_is_function;
  

  DEBUG_PRINT("Launch bootstrap. Setup video...\n");

  /* Map video memory into virtual allocated memory */
  kernel_params->video_memory = 
    (addr_t) vmm_page_alloc_at_phys_addr(VIDEO_MEM_MAP_SIZE, 0xb8000, 
					 /* Cache disable, write through */
					 PTE_WT | PTE_CD);
  

  if (! kernel_params->video_memory)
    FAILED_VERBOSE("Could not map video memory into virtual mapping\n");
  DEBUG_PRINT("Video mapped at 0x%x\n", (unsigned) kernel_params->video_memory);

  DEBUG_PRINT("Setup GDT...\n");

  /* Set up GDT for kernel bootstrap (new gdt will be set up by mm-x86) */

  gdt[0]   = 0x0;                 /* NULL descriptor */
  gdt[1]   = 0x00CF9B000000FFFF;  /* 4G, code, CPL0 */
  gdt[2]   = 0x00CF93000000FFFF;  /* 4G, data, CPL0 */

  gdtr.size = 3*sizeof(unsigned long long);
  gdtr.addr = (addr_t) gdt;

  asm("lgdt %0":: "m" (gdtr));
  asm("movw $0x10, %%ax\n\t"
      "movw %%ax, %%ds\n\t"
      "movw %%ax, %%es\n\t"
      "movw %%ax, %%fs\n\t"
      "movw %%ax, %%gs\n\t"
      "movw %%ax, %%ss\n\t":::"eax");

  DEBUG_PRINT("Looking for KERNEL_BOOTSTRAP\n");
  /* Jump to boostrap entry point */
  if (! ld_lookup_expsym(KERNEL_BOOTSTRAP_SYMBOL_STR,
			 & init_is_function,
			 & init_sym,
			 & init_module))
    FAILED_VERBOSE("(No symbol entry " KERNEL_BOOTSTRAP_SYMBOL_STR
		   " for kernel found !)\n");
  ASSERT_FATAL(init_module && init_sym && init_sym->addr);
  ASSERT_FATAL(init_module->kernel_module->magic == MAGIC_RELOC_EXECUTABLE);

  if (! init_is_function)
    FAILED_VERBOSE("(Kernel entry in module %s is not a function !)\n",
		   init_module->name);
  
  DEBUG_PRINT("End of looking (found at 0x%x in module %s)\n",
	      init_sym->addr, init_module->name);
  
  kernel_params->allocated_memory_base_phys_addr =
    pmm_page_alloc_min_phys_addr();
  kernel_params->allocated_memory_top_virt_addr =
    vmm_get_top_virt_addr();

  DEBUG_PRINT("base_phys=0x%x, top_phys=0x%x\n",
	      kernel_params->allocated_memory_base_phys_addr,
	      kernel_params->loader_allocatable_mem_size);
  DEBUG_PRINT("total_mem=0x%x, init_heap_phys_addr=0x%x, sz=0x%x\n",
	      kernel_params->total_mem_size,
	      kernel_params->init_heap_paddr,
	      kernel_params->init_heap_size);

  bochs_printl("Switch to kernel bootstrap (Adios from loader)...\n");
  kernel_bootstrap = (kernel_startup_routine_t)init_sym->addr;
  res = kernel_bootstrap(kernel_params);
 
  /* We should never get there */
  SEVERE_PRINT("(Bootstrap returned with result %d ?!)\n", res);
  
  HALT();
}

/* Make GCC happy */
void cmain (unsigned long magic, unsigned long addr);

void cmain (unsigned long magic, unsigned long addr)
{
  unsigned int i;
  multiboot_info_t *mbi;
  module_t         *mmod;
  addr_t stack_addr;
  
  cls();
  printl("(Welcome to KOS Loader - this is Wolfgang Release)\n");

#ifdef DEBUG_ON_SERIAL
  printl("(Init loader debug on serial port %d at %d bauds in 8N1 mode)\n",
	 DEBUG_SERIAL_PORT, DEBUG_SERIAL_SPEED);
  if (serial_init(DEBUG_SERIAL_PORT, DEBUG_SERIAL_SPEED, UART_8BITS_WORD,
		  UART_NO_PARITY, UART_1_STOP_BIT) != 0) {
    printl("Warning: Could not initialise serial port\n");
  }
#endif
  
  /* Am I booted by a Multiboot-compliant boot loader?  */
  if ( magic != MULTIBOOT_BOOTLOADER_MAGIC )
      FAILED_VERBOSE("Invalid magic number (not grub ?) : 0x%x\n",
		     (unsigned) magic);

  mbi = (multiboot_info_t *) addr;

  // Check for Modules
  printl("(Checking for modules ...");
  check_if_modules(mbi);
  printl(" Ok)\n");

  // Memory init
  printl("(Init memory ...");
  mm_init(mbi);
  printl(" Ok)\n");

  if(mbi->flags & (1<<2))
    if(parse_cmdline((char *) mbi->cmdline, kernel_params) != 0)
      FAILED_VERBOSE("Error while parsing module parameters, check syntax\n");

  // Builds the internal list of modules
  printl("(Building internal list of modules ");
  ld_modules_head = ld_modules_tail = NULL;
  for( i = 0, mmod = (module_t *) mbi->mods_addr ;
       i < mbi->mods_count ;
       i++, mmod++)
   {
     ld_register_module((char *) mmod->string,
			mmod->mod_start,
			mmod->mod_end - mmod->mod_start, NULL);
     printl(".");
   }
  printl(" Ok)\n");

  /* Load and make the relocations */
  ld_build();

  /* Set up stack for kernel bootstrap */
  DEBUG_PRINT("Set up stack for kernel bootstrap...\n");
  stack_addr = (addr_t) vmm_page_alloc(BOOTSTRAP_STACK_SIZE, NULL);
  if(!stack_addr)
    FAILED_VERBOSE("Couldn't allocate mem for bootstrap stack\n");

  kernel_params->bootstrap_stack_base_addr = stack_addr;

  /* gdt is at the bottom of the stack */
  gdt = (unsigned long long*) stack_addr;

  /* Switch to new stack (and jump to a clean function in order to
     have clean local variables */
  stack_addr += BOOTSTRAP_STACK_SIZE;
  asm("movl %0, %%esp\n\tmovl $0, %%ebp"::"r"(stack_addr));

  /* Make sure this is a real call far (not a jump) ! If this a jmp, then
     launch_bootstrap local variables are allocated on old stack => BING !
     """unfortunately""" gcc 3.1 aggressively transform calls to local
     static functions into jmp... */
  launch_bootstrap();
  FAILED_VERBOSE("Should not print this !\n");
}
