
#include <macros.h>
#include <libcrt/string.h>
#include <assert.h>
#include <mm/mm.h>

#include "_elf32.h"

struct _marshall_symbols_ret_value
{
  uint   nb_symbols;
  size_t str_size;
};

static void calc_symbols_marshall_size(struct elf_mod_info_s *elfmod,
				       Elf32_Shdr *symtab_section,
				       Elf32_Half section_owner /* or 0 == all sections */,
				       /*out*/struct _marshall_symbols_ret_value * result)
{
  Elf32_Shdr* strtab_section = elfmod->section + symtab_section->sh_link;
  Elf32_Sym*  symtab;
  char*        strtab;
  unsigned int i;

  result->nb_symbols = 0;
  result->str_size   = 0;

  symtab = (Elf32_Sym*)symtab_section->sh_addr;
  strtab = (char*)strtab_section->sh_addr;

  for ( i = 0 ;
	i < (symtab_section->sh_size / sizeof(Elf32_Sym)) ;
	i++)
    {
      char *symname       = strtab + symtab[i].st_name;
      Elf32_Shdr* section = elfmod->section + symtab[i].st_shndx;

      /* noname symbols are ignored */
      if (*symname == '\0')
        continue;

      /* ignore NOTYPE symbols */
      if (ELF32_ST_TYPE(symtab[i].st_info)
          == STT_NOTYPE)
        continue;

      /* register the *COM* symbols */
      if (symtab[i].st_shndx == SHN_COMMON)
	goto register_symbol;

      /* ignore symbols in reserved sections */
      if (symtab[i].st_shndx >= elfmod->header->e_shnum)
	continue;
      
      /* ignore symbols in non-alloc sections or in .init heap */
      if (((section->sh_flags & SHF_ALLOC) == 0)
	  || (section->sh_flags & SHF_KOS_ALLOC_IN_INIT_HEAP))
        continue;

      /* Consider all symbols if section_owner is undefined */
      if (section_owner == 0)
	goto register_symbol;
      
      /* Ignore sections different to section_owner if it is specified */
      if (symtab[i].st_shndx != section_owner)
	continue;

    register_symbol:
      result->nb_symbols ++;
      result->str_size   += strlen(symname) + /* \0 */1;
    }

  DEBUG_PRINT("[Calc] Exp or Def syms relative to section %d for symtab %s nb=%d, str_sz=%d\n",
	      section_owner,
	      elf32_get_section_name(elfmod, symtab_section),
	      result->nb_symbols, result->str_size);
}


static uint calc_impsyms_marshall_size(struct elf_mod_info_s *elfmod,
				       Elf32_Shdr *rel_section)
{
  Elf32_Shdr *section        = elfmod->section + rel_section->sh_info;
  Elf32_Shdr *symtab_section = elfmod->section + rel_section->sh_link;
  Elf32_Shdr *strtab_section = elfmod->section + symtab_section->sh_link;
  Elf32_Rel  *reltab;
  Elf32_Sym  *symtab;
  char*        strtab;
  unsigned int i;
  int          res = 0;
  
  reltab = (Elf32_Rel*)rel_section->sh_addr;
  symtab = (Elf32_Sym*)symtab_section->sh_addr;
  strtab = (char*)strtab_section->sh_addr;

  /* Make sure the section is to be allocated, and is not of the .init
     kind */
  if (((section->sh_flags & SHF_ALLOC) == 0)
      || (section->sh_flags & SHF_KOS_ALLOC_IN_INIT_HEAP))
    return 0;

  for ( i = 0 ;
        i < rel_section->sh_size / sizeof(Elf32_Rel);
        i++)
    {
      uint sym_id = ELF32_R_SYM(reltab[i].r_info);
      char *symname = strtab + symtab[sym_id].st_name;

      if ((symtab[sym_id].st_shndx == SHN_UNDEF)
          && (*symname != '\0'))
        res ++;
    }

  DEBUG_PRINT("[Calc] Impsym for SHT_REL %s nb=%d\n",
	      elf32_get_section_name(elfmod, rel_section),
	      res);
  return res;
}

static size_t calc_com_size(struct elf_mod_info_s *elfmod,
			    Elf32_Shdr *symtab_section,
			    offset_t offs_start)
{
  size_t dest_end   = offs_start;
  Elf32_Sym    *symtab;
  unsigned int  i;
  UNUSED(elfmod);

  symtab = (Elf32_Sym*)symtab_section->sh_addr;
  
  for ( i = 0 ;
	i < (symtab_section->sh_size / sizeof(Elf32_Sym)) ;
	i++)
    {
      /* filters the *COM* symbols */
      if (symtab[i].st_shndx != SHN_COMMON)
	continue;

      /* Take symbol's alignment constraint into account */
      dest_end  = ALIGN_SUP(dest_end, symtab[i].st_value);

      dest_end += symtab[i].st_size;
    }

  DEBUG_PRINT("[Calc] *COM* from symtab %s sz=%d\n",
	      elf32_get_section_name(elfmod, symtab_section),
	      dest_end - offs_start);
  return dest_end - offs_start;  
}

static size_t calc_alloc_section_size(struct elf_mod_info_s *elfmod,
				      Elf32_Shdr *section,
				      offset_t offs_start)
{
  size_t dest_start = offs_start;
  size_t dest_end;
#ifndef KEEP_KSYMS
  UNUSED(elfmod);
#endif
  
  /* Take section's alignment constraints into account */
  dest_start = ALIGN_SUP(dest_start, section->sh_addralign);
  dest_end   = dest_start;

  /* Increment size needed */
  dest_end += section->sh_size;

  DEBUG_PRINT("[Calc] load for section %s sz=%d\n",
	      elf32_get_section_name(elfmod, section),
	      dest_end - offs_start);

  /* If we're working on .bss, calc size for the *COM* section */
  if (section == elfmod->section + elfmod->bss_id)
    {
      Elf32_Half section_index;

      for (section_index = 0 ;
	   section_index < elfmod->header->e_shnum ;
	   section_index++)
	{
	  Elf32_Shdr *symtab_section = elfmod->section + section_index;
	  
	  if ((symtab_section->sh_type == SHT_SYMTAB)
	      && (symtab_section->sh_flags & SHF_KOS_USED_SYMTAB))
	    dest_end += calc_com_size(elfmod,
				      symtab_section,
				      dest_end);
	}
    }

  return dest_end - offs_start;
}

int elf32_alloc (ld_module_t *ldmod)
{
  Elf32_Half section_index;
  struct elf_mod_info_s *elfmod;
  size_t total_size = 0, header_size = 0, name_size = 0, strtab_size = 0, orig_init_heap_size;
  uint nb_defsyms = 0 , nb_exp_funcs = 0, nb_exp_vars = 0, nb_imp_syms = 0;
  struct kernel_module *km;
  vaddr_t addr;
  
  ASSERT_RETURN_VAL(ldmod && ldmod->operation_data, -1);
  elfmod = (struct elf_mod_info_s *) ldmod->operation_data;

  header_size += sizeof(struct kernel_module);

  name_size   += strlen(ldmod->name) + 1;
  header_size += name_size;
  
  /* Calc size for the marshalls of the defined/exported/imported symbols */
  for (section_index = 0 ;
       section_index < elfmod->header->e_shnum ;
       section_index++)
    {
      struct _marshall_symbols_ret_value marshall_size;
      Elf32_Shdr *section = elfmod->section + section_index;

      /* Relocation sections */
      if (elfmod->section[section_index].sh_type == SHT_REL)
	nb_imp_syms += calc_impsyms_marshall_size(elfmod, section);
      else if (elfmod->section[section_index].sh_type == SHT_RELA)
	FAILED_VERBOSE("Do not handle SHT_RELA sections yet\n");

      /* Otherwise: ignore non-symtab sections */
      else if (section->sh_type != SHT_SYMTAB)
	continue;

      /* Ignore sections where nothing is of interest */
      if (! (section->sh_flags & SHF_KOS_USED_SYMTAB))
	continue;

#ifdef KEEP_KSYMS
      /* Calc defined symbols marshall size */
      calc_symbols_marshall_size(elfmod,
				 section,
				 0,
				 & marshall_size);
      nb_defsyms  += marshall_size.nb_symbols;
      strtab_size += marshall_size.str_size;
#endif

      /* Calc exported functions marshall size */
      if (elfmod->exp_section_text_id != 0)
	{
	  calc_symbols_marshall_size(elfmod,
				     section,
				     elfmod->exp_section_text_id,
				     & marshall_size);
	  nb_exp_funcs += marshall_size.nb_symbols;
	  strtab_size  += marshall_size.str_size;
	}

      /* Calc exported variables marshall size */
      if (elfmod->exp_section_data_id != 0)
	{
	  calc_symbols_marshall_size(elfmod,
				     section,
				     elfmod->exp_section_data_id,
				     & marshall_size);
	  nb_exp_vars += marshall_size.nb_symbols;
	  strtab_size += marshall_size.str_size;
	}
    }

  /* Update header size */
  header_size += strtab_size;
  header_size += (nb_exp_funcs + nb_exp_vars)*sizeof(exported_symbol_t);
  header_size += nb_imp_syms*sizeof(imported_symbol_t);
  header_size += nb_defsyms*sizeof(defined_symbol_t);

  DEBUG_PRINT("Header size for module %s is %d (0x%x)\n", 
              ldmod->name, header_size, header_size);
  
  total_size = header_size;

  /* Calc size for all the SHF_ALLOC sections */
  orig_init_heap_size = ld_get_init_heap_size();
  for (section_index = 0 ;
       section_index < elfmod->header->e_shnum ;
       section_index++)
    {
      Elf32_Shdr *section = elfmod->section + section_index;
      if (section->sh_flags & SHF_ALLOC)
	{
	  if (section->sh_flags & SHF_KOS_ALLOC_IN_INIT_HEAP)
	    {
	      size_t init_sz = calc_alloc_section_size(elfmod,
						       section,
						       ld_get_init_heap_size());
	      ld_inc_init_heap_size(init_sz);
	    }
	  else
	    total_size += calc_alloc_section_size(elfmod,
						  section,
						  total_size);
	}
    }

  /* Remember size of init needed */
  elfmod->init_heap_size = ld_get_init_heap_size() - orig_init_heap_size;

  /* Commit the preparation -> alloc the kernel_param data */
  km = (struct kernel_module *) vmm_page_alloc(total_size, NULL);
  ASSERT_RETURN_VAL(km != NULL, -1);
  ldmod->kernel_module = km;
  DEBUG_PRINT("Alloc module at 0x%x\n", km);

  km->magic = MAGIC_RELOC_EXECUTABLE;
  km->size  = total_size;
  km->next  = 0;
  km->u.reloc_executable.exp_cc_funcs           = NULL;
  km->u.reloc_executable.nb_exp_cc_funcs        = 0;
  km->u.reloc_executable.exp_cc_vars            = NULL;
  km->u.reloc_executable.nb_exp_cc_vars         = 0;
  km->u.reloc_executable.init_routines          = NULL;
  km->u.reloc_executable.nb_init_routines       = 0;
  km->u.reloc_executable.cleanup_routines       = NULL;
  km->u.reloc_executable.nb_cleanup_routines    = 0;

  /* Init km fields */
  addr = (addr_t)km;

  addr += sizeof(struct kernel_module);
  addr += name_size;
  km->u.reloc_executable.strtab = (char*)addr;
  DEBUG_PRINT("Str tab at addr 0x%x, sz=%d\n",
              km->u.reloc_executable.strtab, strtab_size);
  
  addr += strtab_size;
  km->u.reloc_executable.exp_section_funcs    = (exported_symbol_t*)addr;
  km->u.reloc_executable.nb_exp_section_funcs = 0;
  DEBUG_PRINT("Exp syms funcs from .export.* at addr 0x%x\n",
              km->u.reloc_executable.exp_section_funcs);
  
  addr += nb_exp_funcs*sizeof(exported_symbol_t);
  km->u.reloc_executable.exp_section_vars    = (exported_symbol_t*)addr;
  km->u.reloc_executable.nb_exp_section_vars = 0;
  DEBUG_PRINT("Exp syms vars from .export.* at addr 0x%x\n",
              km->u.reloc_executable.exp_section_vars);

  addr += nb_exp_vars*sizeof(exported_symbol_t);
  km->u.reloc_executable.imp_syms = (imported_symbol_t*)addr;
  DEBUG_PRINT("Imp syms at addr 0x%x\n",
              km->u.reloc_executable.imp_syms);

  addr += nb_imp_syms*sizeof(imported_symbol_t);
  km->u.reloc_executable.def_syms = (defined_symbol_t*)addr;
  DEBUG_PRINT("Def syms at addr 0x%x\n",
              km->u.reloc_executable.def_syms);

  addr += nb_defsyms*sizeof(defined_symbol_t);
  km->u.reloc_executable.contents = addr;
  DEBUG_PRINT("Modules contents at addr 0x%x\n",
              km->u.reloc_executable.contents);

  return 0;
}
