
#include <macros.h>
#include <libcrt/string.h>
#include <assert.h>
#include <mm/mm.h>

#include "_elf32.h"

struct _marshall_symbols
{
  uint    nb_symbols;
  vaddr_t next_sym_vaddr;
  char*   next_str;
};

/* NOTE: the *COM* symbols should have been transfered in the .bss
   section */
static void marshall_symbols(struct elf_mod_info_s *elfmod,
			     Elf32_Shdr *symtab_section,
			     Elf32_Half section_owner  /* or 0 == all sections + defined_symbol_t */,
			     struct _marshall_symbols * result)
{
  Elf32_Shdr* strtab_section = elfmod->section + symtab_section->sh_link;
  Elf32_Sym*  symtab;
  char*        strtab;
  unsigned int i;
  size_t       str_size   = 0;
  uint         nb_symbols = 0;

  symtab = (Elf32_Sym*)symtab_section->sh_addr;
  strtab = (char*)strtab_section->sh_addr;
  
  for ( i = 0 ;
	i < (symtab_section->sh_size / sizeof(Elf32_Sym)) ;
	i++)
    {
      char *symname       = strtab + symtab[i].st_name;
      Elf32_Shdr* section = elfmod->section + symtab[i].st_shndx;

      /* noname symbols are ignored */
      if (*symname == '\0')
        continue;

      /* ignore NOTYPE symbols */
      if (ELF32_ST_TYPE(symtab[i].st_info)
          == STT_NOTYPE)
        continue;

      /* NOTE: the *COM* symbols should have been transfered in the .bss
	 section by the section_alloc of .bss */
      ASSERT_FATAL(symtab[i].st_shndx != SHN_COMMON);
      
      /* ignore symbols in reserved sections */
      if (symtab[i].st_shndx >= elfmod->header->e_shnum)
	continue;
      
      /* ignore symbols in non-alloc sections or in .init heap */
      if (((section->sh_flags & SHF_ALLOC) == 0)
	  || (section->sh_flags & SHF_KOS_ALLOC_IN_INIT_HEAP))
        continue;
      
      if (section_owner == 0)
	{
	  size_t len               = strlen(symname) + 1;
	  defined_symbol_t *defsym = (defined_symbol_t*)result->next_sym_vaddr;
	  defsym->addr = (offset_t)(symtab[i].st_value) + section->sh_addr;

	  strcpy(result->next_str, symname);
	  defsym->name = (char*)result->next_str;

	  result->next_str       += len;
	  str_size               += len;
	  result->next_sym_vaddr += sizeof(defined_symbol_t);
	  nb_symbols ++;
	}
      else if (symtab[i].st_shndx == section_owner)
	{
	  size_t len                = strlen(symname) + 1;
	  exported_symbol_t *expsym = (exported_symbol_t*)result->next_sym_vaddr;
	  expsym->addr    = (offset_t)(symtab[i].st_value) + section->sh_addr;
	  expsym->ref_cnt = 0;

	  strcpy(result->next_str, symname);
	  expsym->name = (char*)result->next_str;

	  result->next_str       += len;
	  str_size               += len;
	  result->next_sym_vaddr += sizeof(exported_symbol_t);
	  nb_symbols ++;
	}
    }

  result->nb_symbols += nb_symbols;
  DEBUG_PRINT("[Build] Exp or Def syms relative to section %d for symtab %s nb=%d, str_sz=%d\n",
	      section_owner,
	      elf32_get_section_name(elfmod, symtab_section),
	      nb_symbols, str_size);
}


static size_t build_com_section(struct elf_mod_info_s *elfmod,
				Elf32_Shdr *symtab_section,
				vaddr_t base_bss,
				vaddr_t base_com)
{
  size_t dest_end = base_com;
  Elf32_Sym    *symtab;
  unsigned int  i;

  symtab = (Elf32_Sym*)symtab_section->sh_addr;
  
  for ( i = 0 ;
	i < (symtab_section->sh_size / sizeof(Elf32_Sym)) ;
	i++)
    {
      /* filters the *COM* symbols */
      if (symtab[i].st_shndx != SHN_COMMON)
	continue;

      /* Do as if the symbol were transfered to the .bss section */
      symtab[i].st_shndx = elfmod->bss_id;

      /* Take symbol's alignment constraint into account */
      dest_end  = ALIGN_SUP(dest_end, symtab[i].st_value);
      symtab[i].st_value = dest_end - base_bss;

      dest_end += symtab[i].st_size;
    }

  DEBUG_PRINT("[Build] *COM* from symtab %s sz=%d\n",
	      elf32_get_section_name(elfmod, symtab_section),
	      dest_end - base_com);
  return dest_end - base_com;
}

static size_t build_section(struct elf_mod_info_s *elfmod,
			    Elf32_Shdr *section,
			    vaddr_t base_vaddr)
{
  vaddr_t dest_start = base_vaddr;
  vaddr_t dest_end;
  UNUSED(elfmod);
  
  /* Take section's alignment constraints into account */
  dest_start = ALIGN_SUP(dest_start, section->sh_addralign);
  dest_end   = dest_start;

  dest_end  += section->sh_size;

  DEBUG_PRINT("[Build] load for section %s sz=%d\n",
	      elf32_get_section_name(elfmod, section),
	      dest_end - base_vaddr);

  if (section == elfmod->section + elfmod->bss_id)
    {
      Elf32_Half section_index;

      /* Alloc the *COM* section */
      for (section_index = 0 ;
	   section_index < elfmod->header->e_shnum ;
	   section_index++)
	{
	  Elf32_Shdr *symtab_section = elfmod->section + section_index;
	  
	  if ((symtab_section->sh_type == SHT_SYMTAB)
	      && (symtab_section->sh_flags & SHF_KOS_USED_SYMTAB))
	    {
	      size_t sz = build_com_section(elfmod,
					    symtab_section,
					    dest_start,
					    dest_end);
	      dest_end         += sz;
	      section->sh_size += sz;
	    }
	}
    }

  /* Copy the section to its new location */
  if (section->sh_size > 0)
    switch(section->sh_type)
      {
      case SHT_PROGBITS:
	memcpy((char*) dest_start,
	       (char*) section->sh_addr,
	       section->sh_size);
	break;
      case SHT_NOBITS:
	memset((char*) dest_start,
	       0x0,
	       section->sh_size);
	break;
      default:
	FAILED_VERBOSE("Loading of section of type %d not supported\n",
		       (int)section->sh_type);
      }

  /* Update the section address */
  section->sh_addr = (Elf32_Addr)dest_start;

  return dest_end - base_vaddr;
}

int elf32_build (ld_module_t *ldmod)
{
  struct kernel_module *km = ldmod->kernel_module;
  struct elf_mod_info_s *elfmod = (struct elf_mod_info_s*)ldmod->operation_data;
  Elf32_Half section_index;
  vaddr_t strtab_vaddr, expfuncs_vaddr, expvars_vaddr, defsyms_vaddr;
  uint    nb_expfuncs = 0, nb_expvars = 0;
  vaddr_t contents_vaddr, init_base, orig_init_base;

  /* Copy name */
  strcpy(km->data, ldmod->name);

  /* Prepare alloc init .init heap */
  orig_init_base
    = init_base
    = ld_alloc_in_init_heap(elfmod->init_heap_size);

  /* Alloc all the SHF_ALLOC sections */
  contents_vaddr = km->u.reloc_executable.contents;
  for (section_index = 0 ;
       section_index < elfmod->header->e_shnum ;
       section_index++)
    {
      Elf32_Shdr *section = elfmod->section + section_index;
      if (section->sh_flags & SHF_ALLOC)
	{
	  if (section->sh_flags & SHF_KOS_ALLOC_IN_INIT_HEAP)
	    init_base += build_section(elfmod,
				       section,
				       init_base);
	  else
	    contents_vaddr += build_section(elfmod,
					    section,
					    contents_vaddr);
	}
    }

  /* Make sure nothing is overwritten */
  ASSERT_FATAL((init_base - orig_init_base)
	       == elfmod->init_heap_size);

  /* Build defined symbols marshall */
  strtab_vaddr   = (vaddr_t)km->u.reloc_executable.strtab;
  expfuncs_vaddr = (vaddr_t)km->u.reloc_executable.exp_section_funcs;
  expvars_vaddr  = (vaddr_t)km->u.reloc_executable.exp_section_vars;
  defsyms_vaddr  = (vaddr_t)km->u.reloc_executable.def_syms;

  for (section_index = 0 ;
       section_index < elfmod->header->e_shnum ;
       section_index++)
    {
      struct _marshall_symbols marshall_vaddr;
      Elf32_Shdr *section = elfmod->section + section_index;
      
      /* Ignore non-symtab sections */
      if (section->sh_type != SHT_SYMTAB)
	continue;

      /* Ignore sections where nothing is of interest */
      if (! (section->sh_flags & SHF_KOS_USED_SYMTAB))
	continue;

      /* Build defsym marshall */
#ifdef KEEP_KSYMS
      marshall_vaddr.next_str = (char*)strtab_vaddr;
      marshall_vaddr.next_sym_vaddr = defsyms_vaddr;
      marshall_symbols(elfmod,
		       section,
		       0,
		       & marshall_vaddr);
      strtab_vaddr  = (vaddr_t)marshall_vaddr.next_str;
      defsyms_vaddr = marshall_vaddr.next_sym_vaddr;
#endif

      /* Build exported functions marshall */
      marshall_vaddr.next_str = (char*)strtab_vaddr;
      marshall_vaddr.next_sym_vaddr = expfuncs_vaddr;
      marshall_vaddr.nb_symbols     = nb_expfuncs;
      if (elfmod->exp_section_text_id != 0)
	marshall_symbols(elfmod,
			 section,
			 elfmod->exp_section_text_id,
			 & marshall_vaddr);
      strtab_vaddr   = (vaddr_t)marshall_vaddr.next_str;
      expfuncs_vaddr = marshall_vaddr.next_sym_vaddr;
      nb_expfuncs    = marshall_vaddr.nb_symbols;

      /* Build exported variables marshall */
      marshall_vaddr.next_str = (char*)strtab_vaddr;
      marshall_vaddr.next_sym_vaddr = expvars_vaddr;
      marshall_vaddr.nb_symbols     = nb_expvars;
      if (elfmod->exp_section_data_id != 0)
	marshall_symbols(elfmod,
			 section,
			 elfmod->exp_section_data_id,
			 & marshall_vaddr);
      strtab_vaddr  = (vaddr_t)marshall_vaddr.next_str;
      expvars_vaddr = marshall_vaddr.next_sym_vaddr;
      nb_expvars    = marshall_vaddr.nb_symbols;
    }

  km->u.reloc_executable.nb_exp_section_funcs = nb_expfuncs;
  km->u.reloc_executable.nb_exp_section_vars = nb_expvars;

  /* Check that nothing has been overwritten */
  ASSERT_FATAL(strtab_vaddr
	       == (vaddr_t)km->u.reloc_executable.exp_section_funcs);
  ASSERT_FATAL(expfuncs_vaddr
	       == (vaddr_t)km->u.reloc_executable.exp_section_vars);
  ASSERT_FATAL(defsyms_vaddr == km->u.reloc_executable.contents);

  return 0;
}
