
#include <macros.h>
#include <libcrt/string.h>
#include <assert.h>
#include <mm/mm.h>

#include "_elf32.h"

/* NULL terminated array of section names that are to be allocated in
   the init region */
static char* init_heap_section_names[] =
{
  INIT_HEAP_SECTIONS_LIST,
  NULL
};

static bool_t check_section_is_in_init_heap (struct elf_mod_info_s* emod,
					     Elf32_Shdr *section)
{
  char *sname, **s;
  
  /* Already determined that section is in the .init heap */
  if (section->sh_flags
      & SHF_KOS_ALLOC_IN_INIT_HEAP)
    return TRUE;
  
  /* Else, must check it according to its name */
  sname = elf32_get_section_name(emod, section);
  if (! sname)
    return FALSE;
  
  /* Is the section name among the .init section names ? */
  for (s = init_heap_section_names ; s && *s ; s++)
    if (! strcmp(*s, sname))
      {
	/* Yes : mark it as such in the sh_flags for later */
	section->sh_flags |= SHF_KOS_ALLOC_IN_INIT_HEAP;
	return TRUE;
      }

  return FALSE;
}


static bool_t check_if_rel_uses_any_symtab(struct elf_mod_info_s *elfmod,
					   Elf32_Shdr *rel_section)
{
  Elf32_Shdr *section        = elfmod->section + rel_section->sh_info;
  Elf32_Shdr *symtab_section = elfmod->section + rel_section->sh_link;
  Elf32_Shdr *strtab_section;
  
  /* Sanity checks */
  ASSERT_FATAL(rel_section->sh_info < elfmod->header->e_shnum);
  ASSERT_FATAL(rel_section->sh_link < elfmod->header->e_shnum);
  
  /* Make sure the section is to be allocated, and is not of the .init
     kind */
  if ((section->sh_flags & SHF_ALLOC) == 0)
    return FALSE;
  
  /* Make sure the symtab is a symtab */
  ASSERT_FATAL(symtab_section->sh_type == SHT_SYMTAB);
  
  /* Make sure the sh_link of the symtab is a strtab */
  ASSERT_FATAL(symtab_section->sh_link < elfmod->header->e_shnum);
  strtab_section = elfmod->section + symtab_section->sh_link;
  ASSERT_FATAL(strtab_section->sh_type == SHT_STRTAB);
  
  /* Everything Ok : mark symtyab section as 'used' */
  symtab_section->sh_flags |= SHF_KOS_USED_SYMTAB;

  return TRUE;
}


int elf32_init (ld_module_t *module)
{
  Elf32_Half section_index;
  struct elf_mod_info_s *elfmod;

  /* Allocating structures to manage our module */
  module->operation_data = pmm_malloc( sizeof(struct elf_mod_info_s) );
  ASSERT_RETURN_VAL(module->operation_data != NULL, -1);
  elfmod = (struct elf_mod_info_s*) module->operation_data;
  elfmod->ldmod   = module;
  elfmod->header  = (Elf32_Ehdr*)module->raw_image_start;
  elfmod->section = (Elf32_Shdr*)(module->raw_image_start + elfmod->header->e_shoff);
  elfmod->init_heap_size = 0;
  elfmod->bss_id         = 0;
  elfmod->exp_section_text_id = 0;
  elfmod->exp_section_data_id = 0;

  /* Travel the section table */
  for (section_index = 0 ;
       section_index < elfmod->header->e_shnum ;
       section_index++)
    {
      Elf32_Shdr* section = elfmod->section + section_index;

      /* update sh_addr field */
      section->sh_addr
	= (Elf32_Addr)(module->raw_image_start + section->sh_offset);

      /* Mark any section to allocate in the .init heap */
      check_section_is_in_init_heap(elfmod, section);

      switch (section->sh_type)
	{
	case SHT_REL:
	  ASSERT_FATAL(section->sh_entsize == sizeof(Elf32_Rel));
	  /* Mark any used symbol table */
	  check_if_rel_uses_any_symtab(elfmod, section);
	  break;
	case SHT_RELA:
	  ASSERT_FATAL(section->sh_entsize == sizeof(Elf32_Rela));
	  /* Mark any used symbol table */
	  check_if_rel_uses_any_symtab(elfmod, section);
	  break;
	case SHT_SYMTAB:
	  ASSERT_FATAL(section->sh_entsize == sizeof(Elf32_Sym));

	  /* The .symtab section was allocated by the compiler =>
             consider it always used */
	  if (!strcmp(elf32_get_section_name(elfmod, section),
		      COMPILER_SYMTAB))
	    {
	      /* Make sure the sh_link of the symtab is a strtab */
	      Elf32_Shdr *strtab_section;
	      ASSERT_FATAL(section->sh_link < elfmod->header->e_shnum);
	      strtab_section = elfmod->section + section->sh_link;
	      ASSERT_FATAL(strtab_section->sh_type == SHT_STRTAB);

	      section->sh_flags |= SHF_KOS_USED_SYMTAB;
	    }
	  
	  break;
	case SHT_PROGBITS:
	  if ((section->sh_flags & SHF_ALLOC)
	      && (!strcmp(elf32_get_section_name(elfmod, section),
			  EXP_SECTION_TEXT)))
	    {
	      /* Make sure we did not already find another bss section */
	      ASSERT_FATAL(elfmod->exp_section_text_id == 0);
	      /* Remember this export text section */
	      elfmod->exp_section_text_id = section_index;	      
	    }
	  else if ((section->sh_flags & SHF_ALLOC)
		   && (!strcmp(elf32_get_section_name(elfmod, section),
			       EXP_SECTION_DATA)))
	    {
	      /* Make sure we did not already find another bss section */
	      ASSERT_FATAL(elfmod->exp_section_data_id == 0);
	      /* Remember this export text section */
	      elfmod->exp_section_data_id = section_index;	      
	    }
	  break;
	case SHT_NOBITS:
	  if (section->sh_flags & SHF_ALLOC)
	    {
	      /* Make sure we did not already find another bss section */
	      ASSERT_FATAL(elfmod->bss_id == 0);
	      /* Remember this .bss section */
	      elfmod->bss_id = section_index;
	    }
	  break;
	default:
	}
    }

  /* Make sure we were able to find a bss section */
  if (elfmod->bss_id == 0)
    FAILED_VERBOSE("No .bss section could be found !");

  return 0;
}
