
#include <macros.h>
#include <assert.h>

#include "_elf32.h"

#define RELOC_LOCAL  1<<1
#define RELOC_IMPORT 1<<2

/* MUST ALWAYS return address immediately after the last imported
   symbol recorded */
static vaddr_t relocate_section(struct elf_mod_info_s *elfmod,
				Elf32_Shdr* rel_section,
				int reloc_flags,
				imported_symbol_t *next_imp_sym)
{
  Elf32_Shdr *section        = elfmod->section + rel_section->sh_info;
  Elf32_Shdr *symtab_section = elfmod->section + rel_section->sh_link;
  Elf32_Shdr *strtab_section = elfmod->section + symtab_section->sh_link;
  Elf32_Rel  *reltab;
  Elf32_Sym  *symtab;
  char*       strtab;
  unsigned int j;

  if ((section->sh_flags & SHF_ALLOC) == 0)
    return (vaddr_t)next_imp_sym;

  reltab = (Elf32_Rel*)rel_section->sh_addr;
  symtab = (Elf32_Sym*)symtab_section->sh_addr;
  strtab = (char*)strtab_section->sh_addr;

  /* Parse relocation table */
  for ( j = 0 ;
	j < (rel_section->sh_size
	     / sizeof(Elf32_Rel)) ;
	j++) 
    {
      Elf32_Sym *sym;
      char   *symname;
      addr_t symaddr;
      addr_t reladdr;

      // calculate reladdr to relocate
      reladdr = section->sh_addr + reltab[j].r_offset ;
      sym     = symtab + ELF32_R_SYM(reltab[j].r_info);
      symname = strtab + sym->st_name;
      symaddr = 0;

      // whether string is NULL symaddr = to the beginning of the
      // related section
      if (*symname == '\0')
        {
          if ((reloc_flags & RELOC_LOCAL) == 0)
            continue;
	  ASSERT_FATAL(sym->st_shndx <= elfmod->header->e_shnum);
          symaddr = elfmod->section[sym->st_shndx].sh_addr;
        }
      else
	{
	  // if the symbol is an external reference (outside the
          // current module), we have to increment the ref_cnt for
          // this symbol
          if ( sym->st_shndx == SHN_UNDEF ) {
            exported_symbol_t *src_sym;

	    if ((reloc_flags & RELOC_IMPORT) == 0)
              continue;
	    
	    if (! ld_lookup_expsym(symname, NULL, & src_sym, NULL))
              FAILED_VERBOSE("Unresolved symbol \"%s\" in module %s.\n",
                             symname,
                             elfmod->ldmod->name);

	    ASSERT_FATAL(src_sym != NULL);
	    ASSERT_FATAL(src_sym->addr);

            if ((section->sh_flags & SHF_KOS_ALLOC_IN_INIT_HEAP) == 0)
              {
                /* Update the imported/exported syms tables */
		ASSERT_FATAL(next_imp_sym != NULL);
		ASSERT_FATAL((addr_t)next_imp_sym
			     < elfmod->ldmod->kernel_module->u.reloc_executable.contents);
                next_imp_sym->addr_to_relocate = reladdr;
                next_imp_sym->relocation_type  = ELF32_R_TYPE(reltab[j].r_info);
                next_imp_sym->src_sym          = src_sym;
                src_sym->ref_cnt++;

                next_imp_sym++;
              }

	    symaddr = src_sym->addr;

          } else {
	    Elf32_Shdr *src_section = elfmod->section + sym->st_shndx;

	    ASSERT_FATAL(sym->st_shndx <= elfmod->header->e_shnum);
	    
            if ((reloc_flags & RELOC_LOCAL) == 0)
              continue;

	    /* Make sure the destination symbol is not in .init */
	    if (((section->sh_flags & SHF_KOS_ALLOC_IN_INIT_HEAP) == 0)
		&& (src_section->sh_flags & SHF_KOS_ALLOC_IN_INIT_HEAP))
	      FAILED_VERBOSE("Symbol \"%s\" called from a non-init section in module %s\npoints to an init section\n",
			     symname, elfmod->ldmod->name);
	    
	    symaddr = sym->st_value + src_section->sh_addr;
	    
	  }
	}
      
      if (symaddr == 0)
	FAILED_VERBOSE("Unresolved symbol \"%s\" in module %s.\n",
                       symname, elfmod->ldmod->name);

      // apply relocation, depending of the type of the relocation
      switch ( ELF32_R_TYPE(reltab[j].r_info) )
        {
                
        case R_386_32: 
          *((unsigned *)reladdr) += symaddr;
          break;
          
        case R_386_PC32:     
          /* Relocation for a foreign symbol => relocation
             addr is an absolute addr */
          *((unsigned *)reladdr) += ( symaddr - (unsigned long)reladdr );
          break;
                
        default:
          FAILED_VERBOSE( "Relocation type #%d on symbol \"%s\" unknown\n",
                          ELF32_R_TYPE(reltab[j].r_info),
                          symname);
          break;
        }
    }

  return (vaddr_t)next_imp_sym;
}

static int do_relocate(struct elf_mod_info_s *elfmod, int reloc_flag)
{
  Elf32_Half section_index;
  vaddr_t    impsyms_end = (vaddr_t)elfmod->ldmod->kernel_module->u.reloc_executable.imp_syms;
  
  /* Travel the section table */
  for (section_index = 0 ;
       section_index < elfmod->header->e_shnum ;
       section_index++)
    {
      Elf32_Shdr* section = elfmod->section + section_index;

      if (section->sh_type == SHT_REL)
	impsyms_end = relocate_section(elfmod,
				       section,
				       reloc_flag,
				       (imported_symbol_t*)impsyms_end);
      else if (section->sh_type == SHT_RELA)
	FAILED_VERBOSE("SHT_RELA relocations not yet supported\n");
    }

  /* Make sure imported symbols were all considered */
  if (reloc_flag & RELOC_IMPORT)
    ASSERT_FATAL(impsyms_end == (vaddr_t)elfmod->ldmod->kernel_module->u.reloc_executable.def_syms);

  return 0;
}


int elf32_relocate_local(ld_module_t *module)
{
  return do_relocate((struct elf_mod_info_s*)module->operation_data,
		     RELOC_LOCAL);
}

int elf32_relocate_import(ld_module_t *module)
{
  return do_relocate((struct elf_mod_info_s*)module->operation_data,
		     RELOC_IMPORT);
}
