/*
 * Copyright (C) 2000, Julien Munier, Thomas Petazzoni
 * http://kos.enix.org
 *
 * Main code of the loader
 *
 * @(#) $Id: ld_build.c,v 1.1.1.1 2003/04/27 12:49:48 thomas Exp $
 */

/* These lines are supposed to be the Kos Loader recently written from
   scratch (mejj) */

#include <assert.h>
#include <macros.h>
#include <types.h>

#include <elf32/_elf32.h>
#include <ar/ar.h>
#include <debug/debug.h>
#include <libcrt/string.h>
#include <libcrt/stdio.h>
#include <mm/mm.h>

#include "ld_bfd.h"

#ifdef CYCLE_COUNT
static unsigned long long tsc;

#define __start_counter() ({ \
  asm volatile("rdtsc":"=A"(tsc)); \
})

#define __stop_counter(fmt,arg...) ({ \
  unsigned long long ntsc; \
  unsigned long high, low; \
  asm volatile("rdtsc":"=A"(ntsc)); \
  low = (ntsc - tsc) & 0xFFFFFFFF; \
  high = ((ntsc-tsc) >> 32) & 0xFFFFFFFF; \
  printl(fmt,##arg);\
  printl("TSC HIGH : %u, LOW : %u\n", \
         (unsigned)high, (unsigned)low); \
})

#else
#define __start_counter()
#define __stop_counter(fmt,arg...)
#endif

kernel_parameter_t *kernel_params;
struct ld_module_struct *ld_modules_head, *ld_modules_tail;
static struct ld_bfd_operations_struct *ld_operations[] = {
  & elf32_ops,
  & ar_bfd_ops,
  NULL
};


/* 0 on success */
inline static int assign_module_operation (struct ld_module_struct *module)
{
  int i;
  struct ld_bfd_operations_struct *op;

  ASSERT_FATAL(module != NULL);
  
  for (i = 0 ; (op = ld_operations[i]) != NULL ; i++) {
    ASSERT_FATAL(op->check_magic != NULL);
    
    if (op->check_magic(module)) {
      /* Ok, this module is to be handled by these operations */
      module->operation = op;
      return 0;
    }
  }
  
  return -1;  
}


/*
 * Adds the module at address addr to the linked list of modules,
 * after "after_which_module" if it is not NULL, or at the end of the
 * ll otherwise. Does also associate it to its corresponding
 * operation. If the module can't be hold by any operation, returns
 * NULL. Returns the resulted 'module struct' on success.
 */
ld_module_t * ld_register_module (char *name, addr_t addr,
				  size_t size,
				  ld_module_t *after_which_module)
{
  struct ld_module_struct *m;
  
  ASSERT_FATAL(name && addr);
  
  m = (struct ld_module_struct *)
        pmm_malloc(sizeof(struct ld_module_struct));
  ASSERT_FATAL(m != NULL);
  
  memset(m, 0x0, sizeof(struct ld_module_struct));
  strncpy(m->name, name, MOD_NAME_MAX_LEN);

  m->raw_image_start = addr;
  m->raw_image_size  = size;

#ifdef COMPUTE_CHECKSUM
  checksum(name, addr, size);
#endif

  if (assign_module_operation(m)) {
    WARNING("Unknown binary type for module %s. Module ignored.\n", m->name);
    return NULL;
  }
  

  if (after_which_module) {
    ld_module_t *mxchg;
    mxchg = after_which_module->next;
    after_which_module->next = m;
    m->next = mxchg;
  } else if (ld_modules_tail) {
    ld_modules_tail->next = m;
    ld_modules_tail       = m;
  } else {
    ld_modules_head = ld_modules_tail = m;
  }

  return m;
}


int ld_build (void)
{
  unsigned int i;
  ld_module_t      *current;
  
  ASSERT_FATAL(kernel_params != NULL);

  ld_init_expsym_hashtable();

  // Applies all the passes for all the operations on their respective
  // modules
  printl("(Applying passes ");
  i = 0;
  while (TRUE) {
    int at_least_one_operation = 0;

    __start_counter();

    for ( current = ld_modules_head ;
	  current ;
	  current = current->next ) {
      
	if (current->operation
	    && (i < (unsigned) current->operation->num_ops)
	    && current->operation->op[i].operation)
	  {
	    at_least_one_operation ++;
	    bochs_printl("[%s] invoking \"%s\" on module %s...\n",
			 current->operation->name
			 ?current->operation->name
			 :"(Unknown)",
			 current->operation->op[i].name
			 ?current->operation->op[i].name
			 :"(Unknown)",
			 current->name?current->name:"unknown");
	    if (current->operation->op[i].operation(current))
	      FAILED_VERBOSE("\"%s\" failed on module \"%s\".\n",
			     current->operation->op[i].name
			       ?current->operation->op[i].name
			       :"(unknown)",
			     current->name);
	  }

	printl(".");
    }
    
    __stop_counter("\nOps[%d] - ", i);
    
    if (! at_least_one_operation)
      break;
    i++;
  }
  printl(" Ok) Init heap size=%dkB\n", kernel_params->init_heap_size >> 10);


  // Builds the final linked list of all the modules
  for ( current = ld_modules_head ;
	current ;
	current = current->next ) 
    {
      
      if(current->kernel_module == NULL)
	continue;
      
      ASSERT_FATAL(current->kernel_module->magic = MAGIC_RELOC_EXECUTABLE);

      if (kernel_params->modules_tail) {
	kernel_params->modules_tail->next = current->kernel_module;
	kernel_params->modules_tail       = current->kernel_module;
      } else
	kernel_params->modules_head = kernel_params->modules_tail
	  = current->kernel_module;
      
    }


  /* Dump hash table stats for symbols */
  ld_stats_hashtable();

  return 0;
}
