
#include <macros.h>
#include <debug/debug.h>
#include <libcrt/string.h>
#include <mm/mm.h>
#include <assert.h>

#include "ld_bfd.h"

/* The key is calculated with the HASH_CMP key, and used as the key
   for the external chaining in the hash (to prevent the strcmp while
   traversing the list). The hash table is HASH_LEN in length. The
   item position is given by: key mod HASH_LEN. */

#define HASH_CMP 509 /* prime */
struct ld_modsym_hash_item_s {
  exported_symbol_t *sym;
  struct ld_module_struct *ldmod;
  int key;
  struct ld_modsym_hash_item_s *next;
};


#define HASH_LEN 17 /* prime */
static struct ld_hash_s {
  int nb_elem;
  struct ld_modsym_hash_item_s *list;
} hash_funcs[HASH_LEN], hash_vars[HASH_LEN];


static int str_hash(const char* str, const int len)
{
  const char *c;
  int key = 0;

  for (c = str ; c && *c ; c++)
    key = ((int)(key*256) + (int)(*c)) % len;

  return key;
}

void ld_init_expsym_hashtable()
{
  int i;

  /* Init hash table */
  for (i = 0 ; i < HASH_LEN; i++)
    {
      hash_funcs[i].nb_elem = 0;
      hash_funcs[i].list = NULL;
      hash_vars[i].nb_elem = 0;
      hash_vars[i].list = NULL;
    }
}

void ld_stats_hashtable()
{
  int i;

  for (i = 0 ; i < HASH_LEN ; i++)
    {
      DEBUG_PRINT("# hash[%d]: %d funcs, %d vars\n",
		  i, hash_funcs[i].nb_elem, hash_vars[i].nb_elem);
    }
  
}

static struct ld_modsym_hash_item_s *
__ld_lookup_expsym_hashtable(const char* str, const int hindex,
			     const int key, struct ld_hash_s *hash)
{
  struct ld_modsym_hash_item_s *hitem;

  for (hitem = hash[hindex].list ; hitem ; hitem = hitem->next)
    if ((hitem->key == key)
	&& (!strcmp(hitem->sym->name, str)))
      {
	return hitem;
      }

  return NULL;
}

static void __ld_register_exported_symbols(struct ld_module_struct *ldmod,
					   exported_symbol_t *exp_sym,
					   uint nb_syms,
					   int funcs_or_vars)
{
  kernel_module_t *mod;
  struct ld_hash_s *hash, *other_hash;

  DEBUG_PRINT("Register %d %s\n",
	      nb_syms, (funcs_or_vars)?"functions":"variables");
 
  mod = (kernel_module_t *)ldmod->kernel_module;
  if (! mod)
    return;

  if (mod->magic != MAGIC_RELOC_EXECUTABLE)
    return;

  if (funcs_or_vars)
    {
      hash       = hash_funcs;
      other_hash = hash_vars;
    }
  else
    {
      hash       = hash_vars;
      other_hash = hash_funcs;
    }

  for ( ;
       nb_syms ;
       nb_syms--, exp_sym++)
    {
      struct ld_modsym_hash_item_s *hitem;
      int key = str_hash(exp_sym->name, HASH_CMP);
      int hindex = key % HASH_LEN;

      if ((exp_sym->name == 0) || (*exp_sym->name == '\0'))
	FAILED_VERBOSE("Trying to export an invalid symbol\n");

      if (exp_sym->addr == 0)
	FAILED_VERBOSE("You try to export symbol %s that is not declared locally inside module %s\n",
		       exp_sym->name,
		       ldmod->name);
      DEBUG_PRINT("Export sym %s (bootstrap=%s)\n",
		  exp_sym->name, KERNEL_BOOTSTRAP_SYMBOL_STR);

      /* Make sure symbol does not already exist */
      hitem = __ld_lookup_expsym_hashtable(exp_sym->name, hindex,
					   key,
					   hash);
      if (hitem)
	{
	  FAILED_VERBOSE("Exported %s %s (module %s) already defined in module %s !\n",
			 funcs_or_vars?"function":"variable",
			 exp_sym->name,
			 mod->data,
			 hitem->ldmod->kernel_module->data);
	  return; /* never reached */
	}

      hitem = __ld_lookup_expsym_hashtable(exp_sym->name, hindex,
					   key,
					   other_hash);
      if (hitem)
	{
	  FAILED_VERBOSE("Exported %s %s (module %s) already defined as a %s in module %s !\n",
			 funcs_or_vars?"function":"variable",
			 exp_sym->name,
			 mod->data,
			 funcs_or_vars?"variable":"function",
			 hitem->ldmod->kernel_module->data);
	  return; /* never reached */
	}
      
      /* Ok : insert symbol in list */
      hitem = pmm_malloc(sizeof(struct ld_modsym_hash_item_s));
      if (! hitem)
	FAILED_VERBOSE("Loader allocation error\n");
      hitem->sym = exp_sym;
      ASSERT_FATAL(ldmod && ldmod->kernel_module);
      hitem->ldmod = ldmod;
      hitem->key = key;
      hitem->next = hash[hindex].list;
      hash[hindex].list = hitem;
      hash[hindex].nb_elem ++;
    }
}

void ld_register_exported_functions(struct ld_module_struct *owner,
				    exported_symbol_t *exp_funcs,
				    uint nb_funcs)
{
  __ld_register_exported_symbols(owner, exp_funcs, nb_funcs, 1);
}

void ld_register_exported_variables(struct ld_module_struct *owner,
				    exported_symbol_t *exp_vars,
				    uint nb_vars)
{
  __ld_register_exported_symbols(owner, exp_vars, nb_vars, 0);
}

/* 1 when found. 0 otherwise */
int ld_lookup_expsym(const char *str,
		     /*out*/int *is_function,
		     /*out*/exported_symbol_t **sym,
		     /*out*/struct ld_module_struct **ldmod)
{
  struct ld_modsym_hash_item_s *hitem;
  int key, hindex;
  int _is_function;

  RETURN_VAL_IF_FAIL(str, 0);

  key = str_hash(str, HASH_CMP);
  hindex = key % HASH_LEN;

  _is_function = 1;
  hitem = __ld_lookup_expsym_hashtable(str, hindex, key,
				       hash_funcs);
  if (! hitem)
    {
      _is_function = 0;
      hitem = __ld_lookup_expsym_hashtable(str, hindex, key,
					   hash_vars);
    }

  if (hitem)
    {
      if (is_function)
	*is_function = _is_function;
      if (ldmod)
	*ldmod = hitem->ldmod;
      if (sym)
	*sym = hitem->sym;
      return 1;
    }

  return 0;
}

