/*
 * Copyright (C) 2000, Julien Munier, Thomas Petazzoni
 * http://kos.enix.org                 
 * 
 * Memory Initialization Header
 * 
 * @(#) $Id: mm.h,v 1.1.1.1 2003/04/27 12:49:48 thomas Exp $
 */
#ifndef __KOS_LD_MM_H__
#define __KOS_LD_MM_H__

/**
 * @file mm.h
 *
 * There are 2 kinds of allocations inside the loader :
 *  - temporary for the loader's usage -> pmm_malloc() [@see
 *    _mm.h]. Data in this space is not known to the kernel, which may
 *    (should) safely overwrite them later on.
 *  - exported to the kernel -> pmm_page_alloc() [@see _mm.h] and
 *    vmm_page_alloc(). Data in this space is available and used by
 *    the  kernel. It contains the modules, the init heap, the kernel
 *    parameters, and more.
 *
 * pmm_malloc() allocates data in identity mapped physical memory (up
 * to 256MB => total_allocatable_mem_size() field of the
 * kernel_parameter), taking care not to overwrite some precious
 * regions (as marked by pmm_init()): the loader, page 0, the Bios,
 * and the Video. The memory allocated chuncks' physical addresses
 * increase, starting from physical address 4096, and pmm_free() is
 * available.
 *
 * pmm_page_alloc() allocates pages (4k) of physical memory starting
 * from the upper address of the identity mapping (up to 256MB =>
 * total_allocatable_mem_size() field of the kernel_parameter), and
 * taking care not to overwrite the same precious regions as for
 * pmm_malloc(). All the allocated chuncks are page-aligned. The
 * memory allocated pages' physical addresses decrease, starting from
 * the end of the identity mapping. pmm_page_alloc_min_phys_addr()
 * returns the lowest physical allocated address of this
 * region. pmm_free() should not be used for these chuncks.
 *
 * vmm_page_alloc() calls pmm_page_alloc(), and also maps each of the
 * allocated pages in a virtual memory address space, starting from
 * CORE_KERNEL_VIRTUAL_ADDR (768M by default). The virtual addresses
 * of the allocated pages in this virtual region increase (while their
 * physical addresses decrease ; see
 * pmm_page_alloc()). vmm_get_top_virt_addr() returns the highest
 * virtual allocated address in this region.
 */

#include <types.h>
#include <boot/multiboot.h>

/**
 * Init physical memory map for allocations (calls pmm_init()), in
 * order not to allocate memory inside the loader's code, the modules,
 * or the Bios/Video.
 */
void mm_init(struct multiboot_info *mbi);

/**
 * Allocate a chunk of size Bytes into physical memory (identity
 * mapped). size of the chunk is aligned to the upper 8B
 * boundary. Physical addresses INCREASE.
 */
void *pmm_malloc(unsigned int size);

/**
 * Allocate a chunk of size Bytes into physical memory and map it into
 * the CORE_KERNEL_VIRTUAL_ADDR space. size of the chunk is aligned to
 * the upper PAGE_SIZE (ie 4kB) boundary !!! Physical & virtual
 * addresses are page aligned, and physical addresses DECREASE.
 */
void * vmm_page_alloc(size_t size, addr_t *allocated_phys_addr);

/**
 * Get virtual address after the latest vmm_page_alloc().
 */
addr_t vmm_get_top_virt_addr(void);

/**
 * Get lowest physical address of the latest vmm_page_alloc()
 * allocation. It is also the lowest address of the latest
 * pmm_page_alloc() allocation (see _mm.h) which is called inside
 * vmm_page_alloc().
 */
addr_t pmm_page_alloc_min_phys_addr(void);


/* Flags for vmm_page_alloc_at_phys_addr */
#define PTE_VAL 0x01   /* 00000001b : Page is present */
#define PTE_WRT 0x02   /* 00000010b : Page is writable */
#define PTE_USR 0x04   /* 00000100b : User/supervisor */
#define PTE_WT  0x08   /* 00001000b : Write through */
#define PTE_CD  0x10   /* 00010000b : Cache disabled */
#define PTE_ACC 0x20   /* 00100000b : Accessed Page */
#define PTE_DIR 0x40   /* 01000000b : Dirty Bit */

/**
 * Map a block of contiguous memory located at at_phys in physical
 * memory into virtual memory. Used for video mem mapping.
 */
void * vmm_page_alloc_at_phys_addr(size_t size, addr_t at_phys, int flags);

#endif
