/*
 * Copyright (C) 2000, Julien Munier, David Decotigny, Thomas Petazzoni
 * http://kos.enix.org                 
 * 
 * Standard interface between loader and kernel modules
 * 
 * @(#) $Id: mod.h,v 1.1.1.1 2003/04/27 12:49:48 thomas Exp $
 */


#ifndef __KOS_LD_MOD_H__
#define __KOS_LD_MOD_H__

#ifdef __IN_LOADER__
#include <types.h>
#else
#include <kos/types.h>
#endif

#ifdef __cplusplus
extern "C" {
#endif /* __cplusplus */

#define TO_STR(s)  _TO_STR(s)
#define _TO_STR(s) #s

#define KERNEL_BOOTSTRAP_SYMBOL      ___kernel_bootstrap
#define KERNEL_BOOTSTRAP_SYMBOL_STR TO_STR(KERNEL_BOOTSTRAP_SYMBOL)

#define INIT_TEXT_SECTION_NAME   ".init.text"
#define INIT_DATA_SECTION_NAME   ".init.data"
#define CTOR_SECTION_NAME        ".ctor"
#define DTOR_SECTION_NAME        ".dtor"

/* Exported symbols tables allocated by loader : the tables are built
   from the symbols defined in these sections */
#define EXP_SECTION_TEXT ".exp_section.text"
#define EXP_SECTION_DATA ".exp_section.data"

/* Exported symbols tables allocated by compiler */
#define EXP_CC_STRSYMTAB_SECTION_NAME  ".exp_cc.strsymtab"
#define EXP_CC_FUNCTAB_SECTION_NAME    ".exp_cc.functab"
#define EXP_CC_VARTAB_SECTION_NAME     ".exp_cc.vartab"

#define __export_text \
  __attribute__((section(EXP_SECTION_TEXT)))
#define __export_data \
  __attribute__((section(EXP_SECTION_DATA)))
#define __init_text \
  __attribute__ ((section(INIT_TEXT_SECTION_NAME)))
#define __init_data \
  __attribute__ ((section(INIT_DATA_SECTION_NAME)))

typedef struct defined_symbol
{
  addr_t     addr;
  const char *name; /* asciiZ */
} defined_symbol_t;

/* Exported symbol entry */
typedef struct exported_symbol
{
  addr_t     addr;
  int        ref_cnt;
  const char *name; /* asciiZ */
} exported_symbol_t;

typedef struct imported_symbol
{
  addr_t          addr_to_relocate;
  unsigned int    relocation_type;
  exported_symbol_t *src_sym; /* link to the corresponding symbol exported
			       by another module */
} imported_symbol_t;

typedef struct init_cleanup_routine
{
  addr_t       routine;
  unsigned int level; // MUST be > 0
} init_cleanup_routine_t;

/*
 * Be careful : only for functions
 */
#define EXPORT_FUNCTION(sym)                                  \
extern const char __kexport_func_name_##sym[];                \
extern const struct exported_symbol __kexport_func_sym_##sym;   \
const char __kexport_func_name_##sym[]                        \
  __attribute__((section(EXP_CC_STRSYMTAB_SECTION_NAME))) = # sym;  \
const struct exported_symbol __kexport_func_sym_##sym           \
  __attribute__((section(EXP_CC_FUNCTAB_SECTION_NAME))) =           \
  { (addr_t)&sym, 0, __kexport_func_name_##sym}

/*
 * Be careful : usage GREATLY disapproved
 */
#define EXPORT_VARIABLE(sym)                                  \
extern const char __kexport_var_name_##sym[];                 \
extern const struct exported_symbol __kexport_var_sym_##sym;    \
const char __kexport_var_name_##sym[]                         \
  __attribute__((section(EXP_CC_STRSYMTAB_SECTION_NAME))) = # sym;  \
const struct exported_symbol __kexport_var_sym_##sym            \
  __attribute__((section(EXP_CC_VARTAB_SECTION_NAME))) =            \
  { (addr_t)&sym, 0, __kexport_var_name_##sym }

#define DECLARE_INIT_SYMBOL(sym,level)                  \
static const struct init_cleanup_routine __kctor_##sym  \
  __attribute__((section(CTOR_SECTION_NAME))) =         \
  { (addr_t)&(sym), (level) }

#define DECLARE_CLEANUP_SYMBOL(sym,level)               \
static const struct init_cleanup_routine __kdtor_##sym  \
  __attribute__((section(DTOR_SECTION_NAME))) =         \
  { (addr_t)&(sym), (level) }

#define DECLARE_KERNEL_BOOTSTRAP(sym)               \
  int KERNEL_BOOTSTRAP_SYMBOL(kernel_parameter_t *) \
    __attribute__ ((alias (#sym)));                 \
    EXPORT_FUNCTION(___kernel_bootstrap)

#define KERNEL_START(type_and_args)                 \
   int kernel_start(type_and_args);                 \
   EXPORT_FUNCTION(kernel_start);                   \
   int kernel_start(type_and_args)

//  Module header structure
//   +--------------------------+
//   | magic                    | <--- this (lowest address)
//   +--------------------------+
//   | size                     |
//   +--------------------------+
//   | ptr to next              |
//   +==========================+ ----- Begin of Reloc Executable -----
//   | ptr to contents          | Never NULL.
//   +--------------------------+
//   | ptr to exp_cc_funcs      | NULL <=> nb_exp_cc_funcs == 0
//   | nb_exp_cc_funcs          |
//   +--------------------------+
//   | ptr to exp_cc_vars       | NULL <=> nb_exp_cc_vars == 0
//   | nb_exp_cc_vars           |
//   +--------------------------+
//   | ptr to init_routines     | NULL <=> nb_init_routines == 0
//   | nb_init_routines         |
//   +--------------------------+
//   | ptr to cleanup_routines  | NULL <=> nb_cleanup_routines == 0
//   | nb_cleanup_routines      |
//   +--------------------------+
//   | ptr to exp_section_funcs | Never NULL. End of array == begin of exp_section_vars
//   +--------------------------+
//   | ptr to exp_section_vars  | Never NULL. End of array == begin of imp_syms
//   +--------------------------+
//   | ptr to imp_syms          | Never NULL. End of array == begin of def_syms
//   +--------------------------+
//   | ptr to def_syms          | Never NULL. End of marshall == begin of contents
//   +--------------------------+
//   | Module name (asciiZ)     | <--- this->data
//   +--------------------------+
//   | S                        | <--- this->strtab
//   |   T     (asciiZ)         |
//   |     R                    |
//   |    DEF+EXP FUNCS+VARS    |
//   +--------------------------+
//   | E                        | <--- exp_section_funcs
//   |   X                      |
//   |     P                    |
//   |  FUNCS (.exp_section.*)  |
//   +--------------------------+
//   | E                        | <--- exp_section_vars
//   |   X                      |
//   |     P                    |
//   |   VARS (.exp_section.*)  |
//   +--------------------------+
//   | I                        | <--- imp_syms
//   |   M                      |
//   |     P                    |
//   |       SYMS               |
//   +--------------------------+
//   | D                        | <--- def_syms
//   |   E                      |
//   |     F                    |
//   |       SYMS               |
//   +--------------------------+
//   |                          | <--- contents
//   .                          .
//   .  CONTENTS                .
//   .   text/data              .
//   .   ctors/dtors/exp syms   .
//   .                          .
//   |                          |
//   +--------------------------+
//                               <--- this + size

typedef struct kernel_module kernel_module_t;

typedef struct module_parameter
{
  char *modname;
  char *params;
} module_parameter_t;

typedef struct kernel_parameter
{
  // Parameters for memory management
  size_t total_mem_size;
  size_t loader_allocatable_mem_size;

  paddr_t allocated_memory_base_phys_addr; // PHYSICAL
  vaddr_t allocated_memory_top_virt_addr;  // VIRTUAL

  /* Beginning of the BOOTSTRAP_STACK_SIZE stack for the bootstrap module */
  vaddr_t bootstrap_stack_base_addr; // VIRTUAL
#define BOOTSTRAP_STACK_SIZE 2*4096

  /* Beginning of the text mode video memory */
  vaddr_t video_memory; // VIRTUAL
#define VIDEO_MEM_MAP_SIZE 4096

  /* init heap (ie where all the .init sections are located) addresses
     and size */
  paddr_t init_heap_paddr;
  vaddr_t init_heap_vaddr;
  size_t  init_heap_size;

  // Used to pass parameters to modules 
  module_parameter_t *module_parameters;
  size_t module_parameters_size;
  int    nb_modules_parameters_entries;

  // System_wide linked list of loaded modules
  kernel_module_t *modules_head, *modules_tail;
} kernel_parameter_t;


/* Init levels => executed by bootstrap module, IRQ disabled */
#define INIT_BASE_LEVEL 1024
#define INIT_NB_LEVELS  0x10000000 /* This should be enough ;) */
#define INIT_LEVEL0     (INIT_BASE_LEVEL+0) /* Base system init : idt,
                                               debug, lib */
#define INIT_LEVEL1     (INIT_BASE_LEVEL+1) /* Memory init : pmm, kmem,
                                               vmm, mm-x86 */
#define INIT_LEVEL2     (INIT_BASE_LEVEL+2) /* Task init : task-x86,
                                               task */
#define INIT_LEVEL3     (INIT_BASE_LEVEL+3) /* Daemons & services :
                                               babel, kgc, kswap */
#define INIT_LEVEL4     (INIT_BASE_LEVEL+4) /* Drivers */
#define INIT_LEVEL5     (INIT_BASE_LEVEL+5) /* Drivers */
#define INIT_LEVEL6     (INIT_BASE_LEVEL+6) /* Drivers */

/* Post init levels => executed by primary thread, IRQ enabled */
#define POST_INIT_BASE_LEVEL (INIT_BASE_LEVEL + INIT_NB_LEVELS)
#define POST_INIT_NB_LEVELS  0x10000000 /* This should be enough ;) */
#define POST_INIT_LEVEL0     (POST_INIT_BASE_LEVEL+0) 
#define POST_INIT_LEVEL1     (POST_INIT_BASE_LEVEL+1) 
#define POST_INIT_LEVEL2     (POST_INIT_BASE_LEVEL+2)
#define POST_INIT_LEVEL3     (POST_INIT_BASE_LEVEL+3)

#define CLEANUP_BASE_LEVEL 1024
#define CLEANUP_NB_LEVELS  0x10000000 /* This should be enough ;) */
#define CLEANUP_LEVEL0     (CLEANUP_BASE_LEVEL+0)
#define CLEANUP_LEVEL1     (CLEANUP_BASE_LEVEL+1)
#define CLEANUP_LEVEL2     (CLEANUP_BASE_LEVEL+2)

typedef int (*kernel_startup_routine_t)(kernel_parameter_t *kp);
typedef int (*init_routine_t)(kernel_parameter_t *kp);
typedef int (*cleanup_routine_t)(void);

struct kernel_module
{
  enum {MAGIC_RELOC_EXECUTABLE=1, MAGIC_KEYMAP} magic;
  size_t size;
  struct kernel_module *next;      

  union {

    /* Reloc executable */
    struct {
      vaddr_t contents; // text + data

      /*
       * Tables allocated by compiler
       */

      // NULL <=> nb_exp_cc_funcs == 0
      exported_symbol_t* exp_cc_funcs;
      uint               nb_exp_cc_funcs;

      // NULL <=> nb_exp_cc_vars == 0
      exported_symbol_t* exp_cc_vars;
      uint               nb_exp_cc_vars;

      // NULL <=> nb_init_routines == 0
      init_cleanup_routine_t* init_routines;
      uint                    nb_init_routines;

      // NULL <=> nb_cleanup_routines == 0
      init_cleanup_routine_t* cleanup_routines;
      uint                    nb_cleanup_routines;

      /*
       * Tables allocated by loader
       */

      /* Marshall of asciiZ elements
	 End of marshall == begin of exp_section_funcs */
      char* strtab; // Never NULL

      /* Array of the loader allocated table of exported functions
	 Never NULL. End of array == begin of exp_section_vars */
      exported_symbol_t* exp_section_funcs;
      uint               nb_exp_section_funcs;

      /* Array of the loader allocated table of exported variables
	 Never NULL. End of array == begin of imp_syms */
      exported_symbol_t* exp_section_vars;
      uint               nb_exp_section_vars;

      /* Array of imported_symbol_t elements
	 End of array == begin of def_syms */
      imported_symbol_t* imp_syms; // Never NULL

      /* Array of defined_symbol_t elements
	 End of marshall == begin of contents */
      defined_symbol_t* def_syms; // Never NULL

    } reloc_executable;
  
    struct {
      uint dummy;
    } keymap;
    
  } u;

  unsigned char data[0];
};

#ifdef __cplusplus
}
#endif /* __cplusplus */

#endif /* __KOS_LD_MOD_H__ */
