
#include <stdlib.h>
#include <unistd.h>
#include <stdio.h>
#include <string.h>

#include <libld/ld_bfd.h>
#include <mm/_mm.h> // For PAGE_SIZE macro
#include <modules/kos/system.h>

#include "_utils.h"

/* ************************************************************
 * Graph generation functions
 */
struct dep_node_s
{
  struct ld_module_struct* src_ldmod;
  int nb_ref;

  /* Only when detail_by_symbol is true */
  struct exported_symbol* src_sym;
  int is_function;

  struct dep_node_s *next;
};
#define get_dep_root(ldmod) \
  ((struct dep_node_s*)((ldmod)->operation_data))

static int gen_dependency_list(ld_module_t* ldmod,
			       int detail_by_symbol)
{
  kernel_module_t* mod;
  imported_symbol_t* imp_sym;

  if (!ldmod || !ldmod->kernel_module)
    return -1;
  mod = ldmod->kernel_module;

  if (mod->magic != MAGIC_RELOC_EXECUTABLE)
    return -1;

  get_dep_root(ldmod) = NULL;

  for (imp_sym = mod->u.reloc_executable.imp_syms ;
       (addr_t)imp_sym < (addr_t)mod->u.reloc_executable.def_syms ;
       imp_sym++)
    {
      int res, is_function;
      struct ld_module_struct* src_ldmod;
      struct dep_node_s* dep_node;

      if (! imp_sym->src_sym)
	abort();
      
      res = ld_lookup_expsym(imp_sym->src_sym->name, & is_function,
			     NULL, & src_ldmod);
      if (!res || !src_ldmod || !src_ldmod->kernel_module
	  || (src_ldmod->kernel_module->magic != MAGIC_RELOC_EXECUTABLE))
	{
	  abort();
	}

      if (detail_by_symbol) {
	for (dep_node = get_dep_root(ldmod) ;
	     dep_node ;
	     dep_node = dep_node->next) {
	  if (dep_node->src_sym == imp_sym->src_sym)
	    break;
	}
      } else {
	for (dep_node = get_dep_root(ldmod) ;
	     dep_node ;
	     dep_node = dep_node->next) {
	  if (dep_node->src_ldmod == src_ldmod)
	    break;	
	}
      }

      if (! dep_node) {
	dep_node = (struct dep_node_s*)malloc(sizeof(struct dep_node_s));
	if (! dep_node)
	  {
	    perror("malloc dep_node");
	    return -1;
	  }

	if (detail_by_symbol) {
	  dep_node->src_sym = imp_sym->src_sym;
	  dep_node->is_function = is_function;
	}
	dep_node->nb_ref    = 0;
	dep_node->src_ldmod = src_ldmod;
	
	dep_node->next = get_dep_root(ldmod);
	get_dep_root(ldmod) = dep_node;
      }
      
      dep_node->nb_ref ++;
    }

  return 0;
}

static void print_deps(FILE* f,
		       ld_module_t* ldmod,
		       int detail_by_symbol,
		       ld_module_t* owned_by)
{
  char from_mod[255], *c;
  int  got_sthg = 0;
  struct dep_node_s *dep_node;

  if (gen_dependency_list(ldmod, detail_by_symbol))
    return;

  c = strrchr(ldmod->kernel_module->data, '/');
  if (c)
    strncpy(from_mod, c+1, sizeof(from_mod));
  else
    strncpy(from_mod, ldmod->kernel_module->data, sizeof(from_mod));
  from_mod[sizeof(from_mod)-1] = '\0';
  c = strchr(from_mod, '.');
  if (c)
    *c='\0';

  fprintf(f, "\n   /* Dependencies for module %s (%sdetailed) */\n",
	  ldmod->kernel_module->data,
	  detail_by_symbol?"":"not ");

  for (dep_node = get_dep_root(ldmod) ; dep_node ; dep_node = dep_node->next){
    char to_mod[255];

    /* Filters out the symbols not owned by owned_by */
    if (owned_by && (dep_node->src_ldmod != owned_by))
      continue;

    if (! got_sthg)
      fprintf(f, "   \"%s\" [shape=house,style=bold,fontsize=28];\n", from_mod);
    got_sthg = 1;
    
    c = strrchr(dep_node->src_ldmod->kernel_module->data, '/');
    if (c)
      strncpy(to_mod, c+1, sizeof(to_mod));
    else
      strncpy(to_mod, dep_node->src_ldmod->kernel_module->data, sizeof(to_mod));
    to_mod[sizeof(to_mod)-1] = '\0';
    c = strchr(to_mod, '.');
    if (c)
      *c='\0';

    if (detail_by_symbol) {

      if (!dep_node->src_sym || !dep_node->src_sym->name)
	abort();

      fprintf(f, "   \"%s\" -> \"%s\" [label=\"%s (%d)\"%s];\n",
	      from_mod,
	      to_mod,
	      dep_node->src_sym->name,
	      dep_node->nb_ref,
	      dep_node->is_function?"":",color=red");
    } else {
      fprintf(f, "   \"%s\" -> \"%s\" [label=\"%d\"];\n",
	      from_mod,
	      to_mod,
	      dep_node->nb_ref);
    }
  }

  return;
}

#define PRINT_DOT_LAYOUT(_f) \
  ({ fprintf(_f, "   size=\"10.5,7.5\"\n"); \
     fprintf(_f, "   ratio=fill\n");   \
     fprintf(_f, "   rotate=90;\n");   \
     fprintf(_f, "   URL = \"http://kos.enix.org/\";\n"); })

// mod's imported symbols
void deps_imp(FILE* f,
	      ld_module_t* ldmod,
	      int detail_by_symbol)
{
  fprintf(f, "digraph KOS_imported_symbols {\n");
  PRINT_DOT_LAYOUT(f);
  print_deps(f, ldmod, detail_by_symbol, NULL);
  fprintf(f, "}\n");
}

// mod's exported symbols
void deps_exp(FILE* f,
	      struct ld_module_struct* owned_by,
	      int detail_by_symbol)
{
  struct ld_module_struct* ldmod;

  fprintf(f, "digraph KOS_exported_symbols {\n");
  PRINT_DOT_LAYOUT(f);

  /* Lookup the module if given */
  for (ldmod = ld_modules_head ; ldmod ; ldmod = ldmod->next)
    print_deps(f, ldmod, detail_by_symbol, owned_by);  

  fprintf(f, "}\n");
}

void deps_all(FILE* f,
	      int detail_by_symbol)
{
  struct ld_module_struct* ldmod;

  fprintf(f, "digraph KOS_cross_refs {\n");
  PRINT_DOT_LAYOUT(f);

  /* Lookup the module if given */
  for (ldmod = ld_modules_head ; ldmod ; ldmod = ldmod->next)
    print_deps(f, ldmod, detail_by_symbol, NULL);  

  fprintf(f, "}\n");
}
