/*
 * Copyright (C) 2000 - kos team
 * http://kos.enix.org
 *
 * Simulation loader for the host
 *
 * @(#) $Id: mod_check.c,v 1.1.1.1 2003/04/27 12:49:48 thomas Exp $
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/mman.h>
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include <libld/ld_bfd.h>
#include <mm/_mm.h> // For PAGE_SIZE macro
#include <modules/kos/system.h>

#include "_utils.h"

int debug               =0;
int view_console_output =1;
int view_bochs_output   =0;
int view_serial_output  =0;

/* ************************************************************
 * Helper functions
 */
static int register_file(char *path)
{
  int fd, r;
  struct stat stat;
  addr_t mod_start;

  DBG("Loading %s\n", path);
  fd = open (path, O_RDONLY);
  if (fd < 0)
    {
      perror("open module");
      return -1;
    }
  r = fstat(fd, & stat);
  if (r < 0)
    {
      perror("stat module");
      return -1;
    }
  mod_start = (addr_t) mmap(NULL, stat.st_size,
			    PROT_READ | PROT_WRITE,
			    MAP_PRIVATE, fd, 0);
  if (mod_start == 0)
    {
      perror("mmap module");
      return -1;
    }
  close(fd);
  
  ld_register_module(path,
		     mod_start,
		     stat.st_size, NULL);
  return 0;
}


static void print_usage(char* path)
{
  fprintf(stderr, "%s -B -S -c -D -I -P -m<file.map> -o<file.dot> -a -i<module> -e<module> modules...\n",
	  path);
  fprintf(stderr, "    -B : enable bochs output\n");
  fprintf(stderr, "    -S : enable serial output\n");
  fprintf(stderr, "    -c : disable console output\n");
  fprintf(stderr, "    -D : enable mod_check specific debug\n");
  fprintf(stderr, "    -I : print init order\n");
  fprintf(stderr, "    -P : print post init order\n");
  fprintf(stderr, "    -T : check start symbol module name\n");
  fprintf(stderr, "    -m<file.map> : generate bochs symbol file\n");
  fprintf(stderr, "  Dependency graph generation (-i, -e, -a mutually exclusive):\n");
  fprintf(stderr, "    -v : include the names of the symbols\n");
  fprintf(stderr, "    -o<file.dot> : output file name (default %s)\n",
	  DFLT_DOT);
  fprintf(stderr, "    -i<module> : give dependencies for module's imported symbols\n");
  fprintf(stderr, "    -e<module> : give dependencies for module's exported symbols\n");
  fprintf(stderr, "    -a : give all dependencies for all modules\n");
}


/* ************************************************************
 * main
 */
int main (int argc, char *argv[])
{
  unsigned int i;
  enum { DEPS_NOP, DEPS_IMP, DEPS_EXP, DEPS_ALL } deps_gen = DEPS_NOP;
  char *deps_outfname = DFLT_DOT;
  char *deps_module = NULL;
  int detail_by_symbol = 0;
  char *bochs_map_fname = NULL;
  int show_init_order = 0;
  int show_post_init_order = 0;
  int show_start_address = 0;

  /* Init kernel_params */
  kernel_params =
    (kernel_parameter_t *) malloc(sizeof(kernel_parameter_t));
  if (! kernel_params)
    {
      perror("malloc kernel_params");
      return -1;
    }
  kernel_params->modules_head = kernel_params->modules_tail = NULL;
  kernel_params->init_heap_size  = 0;
  kernel_params->init_heap_vaddr = 0;
  kernel_params->init_heap_paddr = 0;

  // Not used by mod_check
  kernel_params->total_mem_size = 0;
  kernel_params->loader_allocatable_mem_size = 0;
  
  // Builds the internal list of modules
  while ((i = getopt (argc, argv, "cBSDIPThHavi:e:o:m:")) != -1)
    switch(i)
      {
      case 'c':
	view_console_output = 0;
	break;
      case 'B':
	view_bochs_output = 1;
	break;
      case 'S':
	view_serial_output = 1;
	break;
      case 'D':
	debug = 1;
	break;
      case 'I':
	show_init_order = 1;
	break;
      case 'P':
	show_post_init_order = 1;
	break;
      case 'T':
	show_start_address = 1;
	break;
      case 'v':
	detail_by_symbol = 1;
	break;
      case 'i':
	if (deps_gen != DEPS_NOP)
	  {
	    fprintf(stderr,
		    "> -i option : Other -e/-a/-i option already given\n");
	    return -1;
	  }
	deps_gen = DEPS_IMP;
	deps_module = optarg;
	break;
      case 'e':
	if (deps_gen != DEPS_NOP)
	  {
	    fprintf(stderr,
		    "> -e option : Other -e/-a/-i option already given\n");
	    return -1;
	  }
	deps_gen = DEPS_EXP;
	deps_module = optarg;
	break;
      case 'a':
	if (deps_gen != DEPS_NOP)
	  {
	    fprintf(stderr,
		    "> -a option : Other -e/-a/-i option already given\n");
	    return -1;
	  }
	deps_gen = DEPS_ALL;
	break;
      case 'o':
	deps_outfname = optarg;
	break;
      case 'm':
	bochs_map_fname = optarg;
	break;
      case '-':
      case 'H':
      case 'h':
      case '?':
	if (isprint (optopt))
	  fprintf (stderr, "Unknown option `-%c'.\n", i);
	else
	  fprintf (stderr,
		   "Unknown option character `\\x%x'.\n",
		   i);
	print_usage(argv[0]);
	return 1;
      default:
	abort();
      }

  printl("(Welcome to KOS module checker - this is Wolfgang Release)\n");
  ld_init_expsym_hashtable();
  
  for (i = optind; i < argc; i++)
    if (register_file(argv[i]))
      return -1;
  
  /* Actually try to build */
  ld_build();

  /* Dump hash table stats for symbos */
  ld_stats_hashtable();

  if (bochs_map_fname)
    gen_bochs_symfile(bochs_map_fname);

  if (show_start_address)
    {
      exported_symbol_t* init_sym;
      struct ld_module_struct* init_module;
      int init_is_function;

      printf("Start symbol in module: ");
      if (ld_lookup_expsym(KERNEL_BOOTSTRAP_SYMBOL_STR,
			   & init_is_function,
			   & init_sym,
			   & init_module))
	{
	  printf("%s ", init_module->name);
	  if (! init_is_function)
	    printf("(not a function !) ");
	}
      else
	printf("<not found>");
      printf("\n");
    }
  if (show_init_order)
    {
      printf("Init module routines:\n");
      print_init_order(INIT_BASE_LEVEL, INIT_NB_LEVELS);
    }
  if (show_post_init_order)
    {
      printf("Post init module routines:\n");
      print_init_order(POST_INIT_BASE_LEVEL, POST_INIT_NB_LEVELS);
    }

  if (deps_gen != DEPS_NOP)
    {
      struct ld_module_struct* ldmod;
      FILE *f;

      /* Lookup the module if given */
      if (deps_module)
	{
	  for (ldmod = ld_modules_head ; ldmod ; ldmod = ldmod->next)
	    if (ldmod->kernel_module
		&& (! strcmp(deps_module, ldmod->kernel_module->data)))
	      break;
	  if (! ldmod)
	    {
	      fprintf(stderr, "Could not find module %s !\n", deps_module);
	      fprintf(stderr, "Available modules:\n");
	      for (ldmod = ld_modules_head ; ldmod ; ldmod = ldmod->next)
		fprintf(stderr, "   %s\n", ldmod->kernel_module->data);
	      return -1;
	    }
	}
      
      f = fopen(deps_outfname, "w");
      if (!f)
	{
	  perror("fopen deps file");
	  return -1;
	}
      switch (deps_gen)
	{
	case DEPS_IMP:
	  deps_imp(f, ldmod, detail_by_symbol);
	  break;
	case DEPS_EXP:
	  deps_exp(f, ldmod, detail_by_symbol);
	  break;
	case DEPS_ALL:
	  deps_all(f, detail_by_symbol);
	  break;
	default:
	  abort();
	}
      fclose(f);
    }

  return 0;
}
