/*
 * Copyright (C) 2000, Thomas PETAZZONI
 * http://kos.enix.org
 *
 * Debug functions for modules
 *
 * @(#) $Id: debug.c,v 1.1.1.1 2003/04/27 12:49:48 thomas Exp $
 */

#include <loader/mod.h>
#include <lib/stdio.h>

#include <arch/mm/mm.h>
#include <arch/task/task.h>
#include <kos/asm.h>
#include <kos/spinlock.h>
#include <lib/string.h>
#include <kos/system.h>

#include <debug/_debug.h>

kernel_parameter_t *kernel_params = NULL;

typedef struct 
{
  int is_capable;
  int is_enabled;
} debug_cap_t;

#define DEBUG_LAST     DEBUG_BOCHS

debug_cap_t debug_caps [] =
{
  { FALSE, FALSE },
  { FALSE, FALSE },
  { FALSE, FALSE },
  { FALSE, FALSE }
};

#ifdef DEBUG_SERIAL_SUPPORT
__init_text static int _init_serial_debug(unsigned short unit)
{
  printk("(Debug on serial port %d at %d bauds in 8N1 mode... ",
	 unit, DEBUG_SERIAL_SPEED);
  if (serial_init(unit, DEBUG_SERIAL_SPEED, UART_8BITS_WORD,
		  UART_NO_PARITY, UART_1_STOP_BIT) != 0) {
    printk("Could not initialise serial port\n");
    return -1;
  }
  printk("Ok)\n");

  return 0;
}
#endif

void __dbg_dumpmem(char *src, unsigned long size)
{
  unsigned int i;

  __dbg_printk("Dumpmem at 0x%x :\n", (unsigned)src);
  
  for ( i = 0 ; i < size ; i++ )
    {
      char c = *(src++);
      
      if (c)
	__dbg_printk("%c (0x%x) | ", c, c);
      else
	__dbg_printk("\\0 | ");
    }

  __dbg_printk("End of Dumpmem\n");
}


/* Returns the size of the decoded instruction, or -1 on error */
int disasm(vaddr_t vaddr, char *text)
{
  struct i386_context ctx;
  struct i386_decode instr;

  memset (&ctx, 0, sizeof (ctx));
  memset (&instr, 0, sizeof (instr));

  ctx.mode = CODE32;
  ctx.base = 0;
#define SEG_CODE_KERNEL_ID   1
#warning "TBI: move disasm/backtrace to arch/debug/"
  ctx.segment = SEG_CODE_KERNEL_ID;
  ctx.offset  = vaddr;

  *text = '\0';
  if (!i386_decode (&ctx, &instr) || !i386_decode_text (&ctx, &instr, text,
							SX_ATT))
    return -1;

  return instr.length;
}

int __dbg_disasm(vaddr_t vaddr, int ninstr)
{
  char buffer[64];
  int offset;

  __dbg_printk("# Debug first %d instr at 0x%x:\n", ninstr, vaddr);
  for(offset = 0 ; ninstr > 0 ; ninstr--)
    {
      offset += disasm(vaddr + offset, buffer);
      __dbg_printk("byte @0x%x: %s\n", vaddr+offset, buffer);
    }

  return 0;
}

int __dbg_lookup_sym(addr_t addr) {
  char buf[200];
  addr2sym(addr, buf, sizeof(buf));
  __dbg_printk("lookup_sym(0x%x) => %s.\n", addr, buf);
  return 0;
}


int __dbg_backtrace(struct thread* thread, struct cpu_state_s* cpu_state) {
  vaddr_t top_cpl0_stack;

  __dbg_printk("----------- Backtrace --------------\n");
  
  if (!thread) {
    WARNING("current thread not identified\n");
    top_cpl0_stack = (unsigned)~0;
  } else
    top_cpl0_stack = thread->cpl0_stack_base_addr
      + CPL0_KERNEL_STACK_VIRTUAL_SIZE;
  
  if (! bt_is_enabled())
    __dbg_printk("# Backtrace disabled #\n");
  else {
    char buf[200];
    int nested_level;
    struct bt_data btd;

    bt_init(& btd, cpu_state);
    for (nested_level = 0 ;
	 bt_next(& btd) && (nested_level < 6);
	 nested_level ++) {
      unsigned p0=0, p1=0, p2=0, p3=0;

      addr2sym(btd.next_instruction, buf, sizeof(buf));

      /* Avoids reading above the stack (or else: will get a #PF...) */
      // TBI : The same in USER mode
      if(btd.params_array < USER_SPACE_START)
	{
	  if (top_cpl0_stack - btd.params_array > 12)
	    p3 = (unsigned) *((unsigned *)(btd.params_array+12));
	  if (top_cpl0_stack - btd.params_array > 8)
	    p2 = (unsigned) *((unsigned *)(btd.params_array+8));
	  if (top_cpl0_stack - btd.params_array > 4)
	    p1 = (unsigned) *((unsigned *)(btd.params_array+4));
	  if (top_cpl0_stack - btd.params_array > 0)
	    p0 = (unsigned) *((unsigned *)(btd.params_array+0));
	}
      else
	{
	  TBI("Implement bt for user space stacks\n");
	  return 0;
	}
      
      __dbg_printk("bt[%d] = %s (0x%x, 0x%x, 0x%x, 0x%x)\n",
		   nested_level, buf, p0, p1, p2, p3);
    } /* for */
  } /* else */

  __dbg_printk("------------------------------------\n");
  return 0;
}

void __dbg_halt_handler()
{
  __dbg_enable_output(DEBUG_CONSOLE);
  WARNING("This handler may be overloaded\n");
  __dbg_backtrace(NULL, NULL);
}

int __dbg_printk (char *fmt, ...)
{
  va_list ap;
  int len;
  char __dbg_printk_buffer[256];

  va_start(ap,fmt);
  len = vsnprintf(__dbg_printk_buffer, 
		  sizeof(__dbg_printk_buffer),
		  fmt, ap);

  __dbg_printk_buffer[sizeof(__dbg_printk_buffer)-1] = '\0';

  if (debug_caps[DEBUG_CONSOLE].is_enabled)
    printk("%s", __dbg_printk_buffer);

#ifdef DEBUG_BOCHS_SUPPORT
  if (debug_caps[DEBUG_BOCHS].is_enabled)
    bochs_prints(__dbg_printk_buffer);    
#endif

#ifdef DEBUG_SERIAL_SUPPORT
  if (debug_caps[DEBUG_SERIAL0].is_enabled)
    serial_prints(0, __dbg_printk_buffer);

  if (debug_caps[DEBUG_SERIAL1].is_enabled)
    serial_prints(1, __dbg_printk_buffer);
#endif

  return len;
}


int __dbg_enable_output(int method)
{
  if(method < 0 || method > DEBUG_LAST)
    return -1;

#ifdef DEBUG_SERIAL_SUPPORT
  switch(method)
    {
    case DEBUG_SERIAL0:
      if (debug_caps[DEBUG_BOCHS].is_capable)
	return -1; /* bochs makes it hard to emulate serial line */

      /* Tries to init serial ports only when asked for */
      if (_init_serial_debug(0) == 0)
	debug_caps[DEBUG_SERIAL0].is_capable = TRUE;
      else
	debug_caps[DEBUG_SERIAL0].is_capable = FALSE;
      break;
    case DEBUG_SERIAL1:
      if (debug_caps[DEBUG_BOCHS].is_capable)
	return -1; /* bochs makes it hard to emulate serial line */

      if (_init_serial_debug(1) == 0)
	debug_caps[DEBUG_SERIAL1].is_capable = TRUE;
      else
	debug_caps[DEBUG_SERIAL1].is_capable = FALSE;
      break;
    default:
    }
#endif
  
  if(!debug_caps[method].is_capable)
    return -1;

  debug_caps[method].is_enabled = TRUE;

  return 0;
}

int __dbg_disable_output(int method)
{
  if(method < 0 || method > DEBUG_LAST)
    return -1;
  
  if(!debug_caps[method].is_capable)
    {
      CONCEPTION_ASSERT(!debug_caps[method].is_enabled);
      return -1;
    }

  debug_caps[method].is_enabled = FALSE;

  return 0;
}


static int _init_debug()
{
  if (inb(0xE9) == 0xE9)
    debug_caps[DEBUG_BOCHS].is_capable = TRUE;
  else
    debug_caps[DEBUG_BOCHS].is_capable = FALSE;
    
  debug_caps[DEBUG_CONSOLE].is_capable = TRUE;
  
  return 0;
}


__init_text static int init_module_level0(kernel_parameter_t *kp) 
{
  kernel_params = kp;

  if(_init_debug() < 0)
    return -1;

  if (_init_backtracing_stuff())
    return -1;

  __dbg_enable_output(DEBUG_BOCHS);
  
  return 0;
}


EXPORT_FUNCTION(__dbg_dumpmem);

EXPORT_FUNCTION(addr2sym);
EXPORT_FUNCTION(dump_symbol_map);
EXPORT_FUNCTION(__dbg_halt_handler);
EXPORT_FUNCTION(__dbg_lookup_sym);
EXPORT_FUNCTION(__dbg_printk);
EXPORT_FUNCTION(__dbg_enable_output);
EXPORT_FUNCTION(__dbg_disable_output);

EXPORT_FUNCTION(disasm);

EXPORT_FUNCTION(bt_init);
EXPORT_FUNCTION(bt_next);
EXPORT_FUNCTION(bt_is_enabled);
EXPORT_FUNCTION(__dbg_backtrace);
EXPORT_FUNCTION(__dbg_disasm);

DECLARE_INIT_SYMBOL(init_module_level0,  INIT_LEVEL0);
