/*
 * Copyright (C) 2000, David Decotigny
 * http://kos.enix.org
 *
 * (ultra simple) debug messages sent to a serial port
 *
 * @(#) $Id: serial.c,v 1.1.1.1 2003/04/27 12:49:48 thomas Exp $
 */

#include "serial.h"

#include <kos/asm.h>
#include <lib/stdio.h>

#include <loader/mod.h>

static struct 
{
  unsigned short iobase;
  unsigned short is_enabled;
} _serial_config [] = { { 0x3f8, FALSE }, 
			{ 0x2f8, FALSE },
			{ 0x3e8, FALSE },
			{ 0x2e8, FALSE } };

#define SERIAL_PORT_MAX 4

#define serial_inb inb
#define serial_outb(port,val) outb(val,port)

inline static int
serial_isready (unsigned short port)
{
  unsigned char status;

  status = serial_inb (port + UART_LSR);
  return (status & UART_DATA_READY) ? : -1;
}


char
serial_getchar (unsigned short unit)
{
  if(unit >= SERIAL_PORT_MAX 
     || _serial_config[unit].is_enabled == FALSE)
    return -1;

  /* Wait until data is ready.  */
  while ((serial_inb (_serial_config[unit].iobase + UART_LSR) 
	  & UART_DATA_READY) == 0)
    ;

  /* Read and return the data.  */
  return serial_inb (_serial_config[unit].iobase + UART_RX);
}


/* 0 on success */
static int
serial_putchar (unsigned short port, char c)
{
  /* Perhaps a timeout is necessary.  */
  int timeout = 10000;

  /* Wait until the transmitter holding register is empty.  */
  while ((serial_inb (port + UART_LSR) & UART_EMPTY_TRANSMITTER) == 0)
    if (--timeout == 0)
      /* There is something wrong. But what can I do?  */
      return -1;

  serial_outb (port + UART_TX, c);
  return 0;
}

static char serial_printk_buf[1024];

inline int serial_prints(unsigned short unit, const char *str)
{
  const char *c;
  unsigned short port;

  if(unit >= SERIAL_PORT_MAX 
     || _serial_config[unit].is_enabled == FALSE)
    return -1;
  
  port = _serial_config[unit].iobase;
  
  for (c = str; *c != '\0'; c++)
    serial_putchar(port, *c);
  

  return (int) (c-str);
}

int serial_printk(unsigned short unit, const char *format, ...) 
{
  va_list args;
  int len;

  va_start(args, format);
  len=vsprintf(serial_printk_buf,format,args);
  
  return serial_prints(unit, serial_printk_buf);
}

/* 0 On success */
int serial_init (unsigned short unit, unsigned int speed,
		 int word_len, int parity, int stop_bit_len)
{
  unsigned short div = 0;
  unsigned char status = 0;
  unsigned short serial_port;

  if (unit >= SERIAL_PORT_MAX)
    return -1;

  serial_port = _serial_config[unit].iobase;
  
  /* Turn off the interrupt.  */
  serial_outb (serial_port + UART_IER, 0);

  /* Set DLAB.  */
  serial_outb (serial_port + UART_LCR, UART_DLAB);
  
  /* Set the baud rate.  */
  if (speed > SERIAL_BAUDRATE_MAX)
    return -1;

  div = SERIAL_BAUDRATE_MAX / speed;
  
  serial_outb (serial_port + UART_DLL, div & 0xFF);
  serial_outb (serial_port + UART_DLH, div >> 8);
  
  /* Set the line status.  */
  status |= parity | word_len | stop_bit_len;
  serial_outb (serial_port + UART_LCR, status);

  /* Enable the FIFO.  */
  serial_outb (serial_port + UART_FCR, UART_ENABLE_FIFO);

  /* Turn on DTR, RTS, and OUT2.  */
  serial_outb (serial_port + UART_MCR, UART_ENABLE_MODEM);

  /* Drain the input buffer.  */
  while (serial_isready (serial_port) != -1)
    (void) serial_getchar (unit);
  
  _serial_config[unit].is_enabled = TRUE;
  
  return 0;
}

EXPORT_FUNCTION(serial_printk);
EXPORT_FUNCTION(serial_getchar);
