/* Copyright (C) 2002 Thomas Petazzoni
 * @(#) $Id: _fat.c,v 1.1.1.1 2003/04/27 12:49:48 thomas Exp $
 */

#include <fs/fat/_fat.h>
#include <babel/babel.h>

#include <liblist/liblist.h>
#include <kmem/kmem.h>
#include <lib/stdio.h>
#include <lib/string.h>

#include <part/part.h>

#include <kos/assert.h>
#include <debug/debug.h>

extern struct interface fat_interface;

#define CLUSTER_TO_SECTOR(fat_trans, cluster)  ( ((cluster-2)*(fat_trans)->sector_per_cluster) \
	                                         + (fat_trans)->data_start )
#define IS_LOWER(c)  ((97 <= (c)) && ((c) <= 122))
#define ANY_TO_UPPER(c) (IS_LOWER((c)) ? ((c) - 32) : (c))

#define MAX_NAME_LEN      9
#define MAX_EXTENSION_LEN 4

#define FAT_SR_TO_FAT_TRANS(fat_sr) (((struct fat_translator *) ((struct shadow_resource *) (fat_sr))->translator))
#define FAT_SR_TO_SR(fat_sr) ((struct shadow_resource *) (fat_sr))
#define FAT_SR_TO_TRANS(fat_sr) (((struct shadow_resource *) (fat_sr))->translator)

#define FAT_SR_TO_PART_METHODS(fat_sr) (FAT_SR_TO_FAT_TRANS((fat_sr))->part->translator->interface->kernel_methods)
#define FAT_SR_TO_PART_TRANS(fat_sr) ((struct part_translator *) FAT_SR_TO_FAT_TRANS((fat_sr))->part->translator)

static inline bool_t isfat(struct shadow_resource *part)
{
  struct part_translator *part_trans;
  
  part_trans = (struct part_translator *) part->translator;
  
  if ( part_trans->fs_type == 0x4 
       || part_trans->fs_type == 0x6 
       || part_trans->fs_type == 0xb 
       || part_trans->fs_type == 0xc 
       || part_trans->fs_type == 0xe 
       || part_trans->fs_type == 0xf )
    return TRUE;
  
  return FALSE;
}

/* return cluster from fat partition */
static int 
_fat_read_next_cluster(struct fat_shadow_resource *fat_sr, int cluster)
{
  k_ui16_t fat[512];
  int fat_sect, fat_offset;

  if(FAT_SR_TO_FAT_TRANS(fat_sr)->type == FAT16)
    {
      fat_sect   = (FAT_SR_TO_FAT_TRANS(fat_sr)->fat_start + 
		    (cluster * 2 / FAT_SR_TO_FAT_TRANS(fat_sr)->bytes_per_sector));
      fat_offset = ((cluster * 2) % FAT_SR_TO_FAT_TRANS(fat_sr)->bytes_per_sector) / 2;
    }
  else
    {
      DEBUG_PRINT_VERBOSE("FAT Type not supported\n");
      return -1;
    }

  if (FAT_SR_TO_PART_METHODS(fat_sr)->read((struct shadow_resource *) 
					  FAT_SR_TO_FAT_TRANS(fat_sr)->part, 
					  (char *) fat,
					  fat_sect, 
					  FAT_SR_TO_PART_TRANS(fat_sr)->blk_size) != 0)
    return -1;
   
  return fat[fat_offset];
}


static int 
_fat_find_in_root_dir(struct fat_shadow_resource *fat_sr, char *name, char *ext)
{
  char buf[512];
  struct direntry  *direntry;
  int sector;

  int i, j;

  direntry = (direntry_t *) buf;
  sector = FAT_SR_TO_FAT_TRANS(fat_sr)->root_dir_start;

  //  __dbg_printk("Root dir start : %d\n", sector);

  do
    {
      if (FAT_SR_TO_PART_METHODS(fat_sr)->read((struct shadow_resource *) 
					      FAT_SR_TO_FAT_TRANS(fat_sr)->part,
					      buf, sector, 
					      FAT_SR_TO_PART_TRANS(fat_sr)->blk_size) != 0 )
	{
	  return -1;
	}
      
      for(i = 0; direntry[i].filename[0] != 0; i++)
	{
	  if(direntry[i].attribute == ATTR_LONG_NAME)
	    continue;

	  if(direntry[i].filename[0] == 0xE5)
	    continue;
	  
	  for(j = 7; (j >= 0 && direntry[i].filename[j] == 0x20); j--)
	    direntry[i].filename[j] = 0;
	  for(j = 2; (j >= 0 && direntry[i].extension[j] == 0x20); j--)
	    direntry[i].extension[j] = 0;

	  /*
	  	  __dbg_printk("[find_in_root_dir] is name \'%s\' like \'%s\' and ext \'%s\' like \'%s\' ",
		       direntry[i].filename, name,
		       direntry[i].extension, ext);
	  */
	  
	  if( (!strncmp(direntry[i].filename, name, 8)) &&
	      (!strncmp(direntry[i].extension, ext, 3)))
	    {
	      fat_sr->firstclus = (direntry[i].firstclus_lo | (direntry[i].firstclus_hi << 16));
	      fat_sr->size      = direntry[i].filesize;
	      fat_sr->type      = (direntry[i].attribute & ATTR_DIRECTORY) ? DIR : FILE;
	      //	      DEBUG_PRINT("...Yes it is\n");
	      return fat_sr->firstclus;
	    }
	  //	  DEBUG_PRINT("...Nope\n");
	}
      
      if((direntry[i].filename[0] == 0) || 
	 (sector ==  
	  FAT_SR_TO_FAT_TRANS(fat_sr)->root_dir_start + 
	  FAT_SR_TO_FAT_TRANS(fat_sr)->root_dir_size))
	break;

      sector ++;

    } while(1);
  
  return 0;

}


static int 
_fat_find_in_normal_dir(struct fat_shadow_resource *fat_sr, char *name, char *ext)
{
  k_ui16_t buf[256];
  struct direntry  *direntry;
  int i, j;
  int sector;
  int cluster;
  
  direntry = (direntry_t *) buf;

#define GET_PARENT_SR(sr) ((struct fat_shadow_resource *) ((struct shadow_resource *) sr)->parent_sr)
  cluster = GET_PARENT_SR(fat_sr)->firstclus;

  sector = CLUSTER_TO_SECTOR(FAT_SR_TO_FAT_TRANS(fat_sr), cluster);
  
  while(cluster != 0 && cluster != ~0)
    {

      //      DEBUG_PRINT("cluster is %d and sector is %d\n",cluster,sector);
      
      if (FAT_SR_TO_PART_METHODS(fat_sr)->read((struct shadow_resource *) 
					      FAT_SR_TO_FAT_TRANS(fat_sr)->part, 
					      (char *) buf, sector, 
					      FAT_SR_TO_PART_TRANS(fat_sr)->blk_size) != 0 )
	return -1;
      
      
      for(i = 0; direntry[i].filename[0] != 0; i++)
	{
	  if(direntry[i].attribute == ATTR_LONG_NAME)
	    continue;
	  
	  for(j = 7; (j >= 0 && direntry[i].filename[j] == 0x20); j--)
	    direntry[i].filename[j] = 0;
	  for(j = 2; (j >= 0 && direntry[i].extension[j] == 0x20); j--)
	    direntry[i].extension[j] = 0;

	  /*
	    __dbg_printk("[find_in_normal_dir] is name \'%s\' like \'%s\' and ext \'%s\' like \'%s\' ",
		       direntry[i].filename, name,
		       direntry[i].extension, ext);
	  */
	  
	  
	  if( (!strncmp(direntry[i].filename, name, 8)) &&
	      (!strncmp(direntry[i].extension, ext, 3)))
	    {
	      fat_sr->firstclus = (direntry[i].firstclus_lo | (direntry[i].firstclus_hi << 16));
	      fat_sr->size      = direntry[i].filesize;
	      fat_sr->type      = (direntry[i].attribute & ATTR_DIRECTORY) ? DIR : FILE;
	      //  DEBUG_PRINT("...Yes it is\ncluster is %d and fat_sr->firstclus is %d\n",cluster,fat_sr->firstclus);
	      return fat_sr->firstclus;
	    }
	  //	  DEBUG_PRINT("...Nope\n");
	}

      cluster = _fat_read_next_cluster(fat_sr, cluster);

      if (cluster < 0)
	return -1;
    }
  
  return 0;

}


int fat_init_sr(struct shadow_resource *sr)
{
  unsigned int i, j;
  unsigned int starting_cluster = 0;
  struct fat_shadow_resource *fat_sr;
  
  char name[9], ext[4];

  fat_sr = (struct fat_shadow_resource *) sr;

  /* Since in the root directory of a FAT partition there's no '.'
     entry, we have to simulate it ! */
  if(!strcmp(sr->name, ".") && sr->parent.type == MOUNT_POINT)
    {
      fat_sr->firstclus = 0;
      fat_sr->size      = FAT_SR_TO_FAT_TRANS(fat_sr)->root_dir_start;
      fat_sr->type      = ROOT_DIR;
      return 0;
    }

  for ( i = 0 ; i < strlen(sr->name) && (sr->name[i] != '.'); i++ )
    {
      if (i >= 8)
	return -1;
      name[i] = ANY_TO_UPPER(sr->name[i]);
    }

  name[i] = '\0';
  
  for ( j = 0 ; i < strlen(sr->name) && (sr->name[i] != '.'); j++, i++ )
    {
      if (j >= 3)
	return -1;
      ext[j] = ANY_TO_UPPER(sr->name[i]);
    }
  
  ext[j] = '\0';
  
  if (sr->parent_sr->parent.type == MOUNT_POINT)
    {
      starting_cluster = _fat_find_in_root_dir(fat_sr, name, ext);  
    }
  else
    {
      starting_cluster = _fat_find_in_normal_dir(fat_sr, name, ext);
    }
  
  if (!starting_cluster)
    return -1;
  
  __dbg_printk("File (%s) found @%d (size is %d)\n", fat_sr->parent.name, fat_sr->firstclus, fat_sr->size);

  return 0;
}



int fat_read(struct shadow_resource *sr, char *buf, offset_t offset, size_t len)
{
  struct fat_shadow_resource *fat_sr;
  int cluster;
  int byte_read = 0;
  int i;

  fat_sr = (struct fat_shadow_resource *) sr;

  CONCEPTION_ASSERT(sr && buf && len);

  cluster = fat_sr->firstclus;

  UNUSED(offset);
  DEBUG_PRINT("[fat_read_file] We never use offset arg !\n");

  while(cluster != 0 && cluster != 0xFFFF)
    {

      /* Read the whole first cluster of the file */
      for(i = CLUSTER_TO_SECTOR(FAT_SR_TO_FAT_TRANS(fat_sr), cluster); 
	  (i < (CLUSTER_TO_SECTOR(FAT_SR_TO_FAT_TRANS(fat_sr), cluster) + 
		FAT_SR_TO_FAT_TRANS(fat_sr)->sector_per_cluster)) && ((int)byte_read < (int)len); 
	  i++)
	{

	  DEBUG_PRINT("[fat] loading sector %d @0x%x\n", i, (unsigned) buf);

	  if (FAT_SR_TO_PART_METHODS(fat_sr)->read((struct shadow_resource *) 
						  FAT_SR_TO_FAT_TRANS(fat_sr)->part,
						  buf, i, 
						  FAT_SR_TO_PART_TRANS(fat_sr)->blk_size) != 0 )
	    {
	      DEBUG_PRINT_VERBOSE("Error while reading\n");
	      return -1;
	    }

	  byte_read += FAT_SR_TO_FAT_TRANS(fat_sr)->bytes_per_sector;
	  buf       += 512;
	}

      cluster = _fat_read_next_cluster(fat_sr, cluster);
      if(cluster < 0)
	return -1;
    }

  return byte_read;
}

int fat_check_capability( struct shadow_resource *part)
{
  char buffer[512];
  fat16_bpb_t *bpb16 = (fat16_bpb_t *) buffer;
  fat32_bpb_t *bpb32 = (fat32_bpb_t *) buffer;

  k_ui16_t root_dir_start;    
  k_ui16_t root_dir_size;      
  k_ui16_t data_start;         
  k_ui16_t fat_size;
  k_ui32_t total_sector_nb;
  int countclusters;

  if(!isfat(part))
    return FALSE;

  if(part->translator->interface->kernel_methods->read(part, buffer, 0, 1) != 0)
    {
      return FALSE;
    }
  
  if(!bpb16->bpb_bytspersec||
     !bpb16->bpb_secperclus||
     !bpb16->bpb_rsvdseccnt)
    {
      return FALSE;
    }

  if(bpb16->bpb_fatsz16 != 0)
    fat_size = bpb16->bpb_fatsz16;
  else
    fat_size = bpb32->bpb_fatsz32;

  if(bpb16->bpb_totsec16 != 0)
    total_sector_nb = bpb16->bpb_totsec16;
  else
    total_sector_nb = bpb32->bpb_totsec32;
  
  root_dir_start     = 
    bpb16->bpb_rsvdseccnt + (bpb16->bpb_numfats * fat_size);
  root_dir_size      = 
    ((bpb16->bpb_rootentcnt * 32) + 
     (bpb16->bpb_bytspersec - 1)) / bpb16->bpb_bytspersec;
  data_start         = 
    root_dir_start + root_dir_size;
	  
  countclusters = (total_sector_nb - data_start) 
    / bpb16->bpb_secperclus;
	  
  if(countclusters < 4085)
    /* FAT12 not implemented yet */
    return FALSE; 
  else if(countclusters < 65525)
    return TRUE; /* FAT 16 */
  else
    return FALSE; /* FAT 32 implementation not tested */

  return FALSE;
}

int fat_getsize(struct resource *res)
{
  return ((struct fat_shadow_resource *) res->shadow_resource)->size;
}

int fat_constructor (struct translator *trans, void *rc)
{
  char buffer[512];
  struct fat_translator *fat_trans = (struct fat_translator *) trans;
  fat16_bpb_t *bpb16 = (fat16_bpb_t *) buffer;
  fat32_bpb_t *bpb32 = (fat32_bpb_t *) buffer;
  int countclusters;
  struct fs_constructor_parameters *f_params = (struct fs_constructor_parameters *) rc;

  CONCEPTION_ASSERT(trans && f_params && f_params->part);

  if(f_params->part->translator->interface->kernel_methods->read(f_params->part, buffer, 0, 1) != 0)
    {
      return -1;
    }
  
  /* Correctly initialize values */
  fat_trans->bytes_per_sector   = bpb16->bpb_bytspersec;
  fat_trans->sector_per_cluster = bpb16->bpb_secperclus;
  fat_trans->fat_start          = bpb16->bpb_rsvdseccnt;
  fat_trans->fat_nb             = bpb16->bpb_numfats;

  if(bpb16->bpb_fatsz16 != 0)
    fat_trans->fat_size = bpb16->bpb_fatsz16;
  else
    fat_trans->fat_size = bpb32->bpb_fatsz32;

  if(bpb16->bpb_totsec16 != 0)
    fat_trans->total_sector_nb = bpb16->bpb_totsec16;
  else
    fat_trans->total_sector_nb = bpb32->bpb_totsec32;
  
  fat_trans->root_dir_start     = 
    bpb16->bpb_rsvdseccnt + (fat_trans->fat_nb * fat_trans->fat_size);
  fat_trans->root_dir_size      = 
    ((bpb16->bpb_rootentcnt * 32) + 
     (bpb16->bpb_bytspersec - 1)) / bpb16->bpb_bytspersec;
  fat_trans->data_start         = 
    fat_trans->root_dir_start + fat_trans->root_dir_size;
	  
  countclusters = (fat_trans->total_sector_nb - fat_trans->data_start) 
    / fat_trans->sector_per_cluster;
	  
  if(countclusters < 4085)
    fat_trans->type = FAT12;
  else if(countclusters < 65525)
    fat_trans->type = FAT16;
  else
    fat_trans->type = FAT32;

  __dbg_printk("\nMounted partition (%s)\n with params : bps=%d, spc=%d, fat_start=%d, fat_nb=%d, total_nb=%d, data_start=%d, type=",
	       fat_trans->parent.name, fat_trans->bytes_per_sector, fat_trans->sector_per_cluster, 
	       fat_trans->fat_start, fat_trans->fat_nb,
	       fat_trans->total_sector_nb, fat_trans->data_start);
  switch(fat_trans->type)
    {
    case FAT12:
      __dbg_printk("FAT12\n");
      break;
    case FAT16:
      __dbg_printk("FAT16\n");
      break;
    case FAT32:
      __dbg_printk("FAT32\n");
      break;
    default:
      __dbg_printk("Unknown\n");
    }
  
  fat_trans->part = f_params->part;

  return 0;
}













