/*
 * Copyright (C) 2000, Thomas Petazzoni <thomas.petazzoni@enix.org>
 * http://kos.enix.org
 *
 * ATA driver
 * with the help of Soren Schmidt, <sos@freebsd.dk>
 * @(#) $Id: _ide.c,v 1.1.1.1 2003/04/27 12:49:48 thomas Exp $
 */

#include <kos/asm.h>
#include <lib/stdio.h>
#include <scheduler/scheduler.h>
#include <idt/irq.h>
#include <ide/_ide.h>
#include <liblist/liblist.h>
#include <kos/spinlock.h>
#include <lib/string.h>
#include <kmem/kmem.h>
#include <debug/debug.h>


typedef struct ide_op
{
  int type;
  enum { IDE_OP_WAITING, IDE_OP_IN_PROGRESS } status;
  int cyl;
  int sector;
  int head;
  char *buffer;
  struct ide_op *next, *prev;
} ide_op_t;

ide_op_t *ide_op_list;

controller_t controllers[NB_CONTROLLERS] =
{
  { 0x1F0, 14, ATA_NOT_PRESENT, PRIMARY_CONTROLLER, 0 },
  { 0x170, 15, ATA_NOT_PRESENT, SECONDARY_CONTROLLER, 0 }
};

static int _ide_init_controller(controller_t *ctrl)
{
  int status0, status1;
  int mask = 0;
  int timeout;

  outb(ATA_D_IBM | ATA_D_MASTER, ctrl->ioaddr + ATA_DRIVE);
  udelay(1);
  status0 = inb(ctrl->ioaddr + ATA_STATUS);

  outb(ATA_D_IBM | ATA_D_SLAVE, ctrl->ioaddr + ATA_DRIVE);
  udelay(1);
  status1 = inb(ctrl->ioaddr + ATA_STATUS);
  
  if((status0 & 0xf8) != 0xf8)
    mask |= 0x01;
  if((status1 & 0xf8) != 0xf8)
    mask |= 0x02;

  /* If both BSY bit are set => leave */
  if((status0 & ATA_S_BSY) && (status1 & ATA_S_BSY))
    return 0;

  /* no device present */
  if(!mask)
    return 0;
  
  /* select the master */
  outb(ATA_D_IBM | ATA_D_MASTER, ctrl->ioaddr + ATA_DRIVE);
  usleep(1);
  /* send reset */
  outb(ATA_A_nIEN | ATA_A_RESET, ctrl->ioaddr + ATA_DEVICE_CONTROL);
  udelay(1);
  outb(ATA_A_nIEN, ctrl->ioaddr + ATA_DEVICE_CONTROL);
  udelay(1);
  inb(ctrl->ioaddr + ATA_ERROR);
  outb(ATA_A_4BIT, ctrl->ioaddr + ATA_DEVICE_CONTROL);
  udelay(1);

  /* wait busy */
  for(timeout = 0; timeout < 300000; timeout++)
    {
      /* select master and get status register */
      outb(ATA_D_IBM | ATA_D_MASTER, ctrl->ioaddr + ATA_DRIVE);
      udelay(1);
      status0 = inb(ctrl->ioaddr + ATA_STATUS);
      
      /* select slave and get status register */
      outb(ATA_D_IBM | ATA_D_SLAVE, ctrl->ioaddr + ATA_DRIVE);
      udelay(1);
      status1 = inb(ctrl->ioaddr + ATA_STATUS);

      if(mask == 0x01)
	if(!(status0 & ATA_S_BSY))
	  break;
      if(mask == 0x02)
	if(!(status1 & ATA_S_BSY))
	  break;
      if(mask == 0x03)
	if(!(status0 & ATA_S_BSY) && !(status1 & ATA_S_BSY))
	  break;
      udelay(1);
    }

  if(status0 & ATA_S_BSY)
    mask &= ~0x01;
  if(status1 & ATA_S_BSY)
    mask &= ~0x02;
  
  /* no device */
  if(!mask)
    return 0;

  /* there's at least one device, initialize state */
  ctrl->state = ATA_IDLE;

  /* check what is the type of the device */
  outb(ATA_D_IBM | ATA_D_MASTER, ctrl->ioaddr + ATA_DRIVE);
  udelay(1);
  
  if(inb(ctrl->ioaddr + ATA_CYL_LSB) == ATAPI_MAGIC_LSB &&
     inb(ctrl->ioaddr + ATA_CYL_MSB) == ATAPI_MAGIC_MSB)
    {
      ctrl->devices |= ATA_ATAPI_MASTER;
    }
  
  outb(ATA_D_IBM | ATA_D_SLAVE, ctrl->ioaddr + ATA_DRIVE);
  udelay(1);
  
  if(inb(ctrl->ioaddr + ATA_CYL_LSB) == ATAPI_MAGIC_LSB &&
     inb(ctrl->ioaddr + ATA_CYL_MSB) == ATAPI_MAGIC_MSB)
    {
      ctrl->devices |= ATA_ATAPI_SLAVE;
    }

  if(status0 != 0x00 && !(ctrl->devices & ATA_ATAPI_MASTER))
    {
      outb(ATA_D_IBM | ATA_D_MASTER, ctrl->ioaddr + ATA_DRIVE);
      usleep(1);
      outb(0x58, ctrl->ioaddr + ATA_ERROR);
      outb(0xa5, ctrl->ioaddr + ATA_CYL_LSB);
      if(inb(ctrl->ioaddr + ATA_ERROR) != 0x58 &&
	 inb(ctrl->ioaddr + ATA_CYL_LSB) == 0xa5)
	{
	  ctrl->devices |= ATA_ATA_MASTER;
	}
    }
  
  if(status1 != 0x00 && !(ctrl->devices & ATA_ATAPI_SLAVE))
    {
      outb(ATA_D_IBM | ATA_D_SLAVE, ctrl->ioaddr + ATA_DRIVE);
      usleep(1);
      outb(0x58, ctrl->ioaddr + ATA_ERROR);
      outb(0xa5, ctrl->ioaddr + ATA_CYL_LSB);
      if(inb(ctrl->ioaddr + ATA_ERROR) != 0x58 &&
	 inb(ctrl->ioaddr + ATA_CYL_LSB) == 0xa5)
	{
	  ctrl->devices |= ATA_ATA_SLAVE;
	}
    }

  return 0;
}

static int _ide_convert_to_ascii(char *src, char *dst, int len)
{
  int i;
  
  for(i = 0; i < len; i=i+2)
    {
      dst[i] = (src[i+1] & 0xff);
      dst[i+1] = (src[i] & 0xff);
    }

  dst[len] = '\0';
  for(i = len; i > 0; i--)
    {
      if(dst[i] == 0)
	continue;
      else if(dst[i] == ' ')
	dst[i] = '\0';
      else
	break;
    }
  
  return 0;
}

static void _ide_handler(int nb, cpu_state_t *state)
{
  UNUSED(state);

  if(controllers[PRIMARY_CONTROLLER].state == ATA_WAIT_INTR)
    {
      __dbg_printk("Got an IRQ %d !\n", nb);
      __dbg_printk("The job was of type : %d\n", ide_op_list->type);
      controllers[PRIMARY_CONTROLLER].state = ATA_IDLE;
    }
}

static int _ide_init_devices(int controller)
{
  char c;
  char name[5];	  
  controller_t *ctrl;
  struct ide_dev dev;

  switch(controller)
    {
    case PRIMARY_CONTROLLER:
      c = 'a';
      ctrl = &controllers[PRIMARY_CONTROLLER];
      break;
    
    case SECONDARY_CONTROLLER:
      c = 'c';
      ctrl = &controllers[SECONDARY_CONTROLLER];
      break;
  
    default:
      DEBUG_PRINT_VERBOSE("Unknown controller !\n");
      return -1;
    }
  
  dev.controller = ctrl;
    
  if(ctrl->state == ATA_NOT_PRESENT)
    return 0;
  
  
  if(ctrl->devices & ATA_ATA_MASTER)
    {
      snprintf(name, sizeof(name), "hd%c", c);
      dev.device = MASTER;
      if (__ide_add_hd(name,&dev) != 0)
	return -1; 
    }
  
  if(ctrl->devices & ATA_ATA_SLAVE)
    {
      snprintf(name, sizeof(name), "hd%c", c+1);
      dev.device = SLAVE;

      if (__ide_add_hd(name,&dev) != 0)
	return -1; 
    }
  
  if(ctrl->devices & ATA_ATAPI_MASTER)
    {
      DEBUG_PRINT_VERBOSE("hd%c (CD-ROM) ", c);
    }

  if(ctrl->devices & ATA_ATAPI_SLAVE)
    {
      DEBUG_PRINT_VERBOSE("hd%c (CD-ROM) ", c+1);
    }

  return 0;
}


int __init_ide(void)
{

  _ide_init_controller(&controllers[PRIMARY_CONTROLLER]);
  _ide_init_devices(PRIMARY_CONTROLLER);

  _ide_init_controller(&controllers[SECONDARY_CONTROLLER]);
  _ide_init_devices(SECONDARY_CONTROLLER);

  register_irq_handler(14, _ide_handler);
  
  return 0;
}


int _ide_add_op_polled_mode(struct shadow_resource *sr, int type, 
			    k_ui32_t sector_num, k_ui16_t *buffer,
			    size_t len)
{
  int timeout, status = 0;
  int i;
  k_ui8_t cyl_lo, cyl_hi, sect, head;
  controller_t *ctrl;
  
  struct ide_translator *ide_trans = (struct ide_translator *)sr->translator;

  CONCEPTION_ASSERT(len == ide_trans->blk_size);

  ctrl = ide_trans->ctrl;

  if(type != READ && type != WRITE)
    {
      DEBUG_PRINT("type invalid\n");
      return -1;
    }
  
  ctrl->state = ATA_IDLE;

  /* Does harddisk support LBA ? */
  if(ide_trans->flags & HARDDISK_LBA_CAPABLE)
    {
      sect   = (sector_num & 0xff);
      cyl_lo = (sector_num >> 8) & 0xff;
      cyl_hi = (sector_num >> 16) & 0xff;
      head   = ((sector_num >> 24) & 0x7) | 0x40;
    }
  else
    {
      int cylinder = sector_num / 
	(ide_trans->logical_head_nb * ide_trans->logical_sector_per_track);
      int temp = sector_num %
	(ide_trans->logical_head_nb * ide_trans->logical_sector_per_track);
      cyl_lo = cylinder & 0xff;
      cyl_hi = (cylinder >> 8) & 0xff;
      head   = temp / ide_trans->logical_sector_per_track;
      sect   = (temp % ide_trans->logical_sector_per_track) + 1;
    } 

  /* I don't know if we really need this, but without this, the Error
     bit of the status register is set, even before sending the
     command ! */
  outb(ATA_D_IBM | ATA_D_MASTER, ctrl->ioaddr + ATA_DRIVE);
  udelay(100);
  outb(ATA_A_nIEN | ATA_A_RESET, ctrl->ioaddr + ATA_DEVICE_CONTROL);
  udelay(100);
  outb(ATA_A_nIEN, ctrl->ioaddr + ATA_DEVICE_CONTROL);
  inb(ctrl->ioaddr + ATA_ERROR);

  /* Waiting for controller to be ready (BSY=0) */
  for(timeout = 0; timeout < 30000; timeout++)
    {
      status = inb(ctrl->ioaddr + ATA_STATUS);
      if(!(status & ATA_S_BSY))
	break;
      
      udelay(100);
    }
  if(timeout == 30000)
    {
      DEBUG_PRINT("[ide] timeout 1\n");
      return -1;
    }

  /* Check for ERR bit to be cleared */
  if(inb(ctrl->ioaddr + ATA_STATUS) & ATA_S_ERROR)
    {
      DEBUG_PRINT("err bit set\n");
      return -1;
    }

  /* Write to registers */
  outb(ATA_A_nIEN | ATA_A_4BIT, ctrl->ioaddr + ATA_DEVICE_CONTROL);
  outb(1, ctrl->ioaddr + 1);
  outb(0, ctrl->ioaddr + ATA_PRECOMP);
  outb(1, ctrl->ioaddr + ATA_SECTOR_COUNT);
  outb(sect, ctrl->ioaddr + ATA_SECTOR_NUMBER);
  outb(cyl_lo, ctrl->ioaddr + ATA_CYL_LSB);
  outb(cyl_hi, ctrl->ioaddr + ATA_CYL_MSB);
  outb(ATA_D_IBM
       | ((ide_trans->device == MASTER) ? ATA_D_MASTER : ATA_D_SLAVE)
       | head, 
       ctrl->ioaddr + ATA_DRIVE);

  /* ATA-4 says to wait 400nsec here, but it doesn't change anything ! */
  udelay(100);

  /* For the moment, we don't care about the operation type => always
     read */
  if(type == READ)
    outb(ATA_C_READ, ctrl->ioaddr + ATA_CMD);
  else if(type == WRITE)
    outb(ATA_C_WRITE, ctrl->ioaddr + ATA_CMD);

  /* Wait for controller to be ready (BSY=0) */
  for(timeout = 0; timeout < 30000; timeout++)
    {
      status = inb(ctrl->ioaddr + ATA_STATUS);
      if(!(status & ATA_S_BSY))
	break;
      
      udelay(100);
    }

  if(timeout >= 30000)
    {
      DEBUG_PRINT("timeout 2\n");
      return -1;
    }
  
  /* Check for ERR to be cleared */
  if(inb(ctrl->ioaddr + ATA_STATUS) & ATA_S_ERROR)
    {
      DEBUG_PRINT("err bit set\n");
      return -1;
    }

  /* Check for DRQ to be set */
  if(!(inb(ctrl->ioaddr + ATA_STATUS) & ATA_S_DRQ))
    {
      DEBUG_PRINT("drq not set\n");
      return -1;
    }

  /* Read the data */
  if(type == READ)
    {
      for(i = 0; i < 256; i++)
	buffer[i] = inw(ctrl->ioaddr);
    }
  else if(type == WRITE)
    {
      for (i = 0; i < 256; i++)
	outw(buffer[i], ctrl->ioaddr);
    }

  return 0;
}

/* this function set the command "Identify Drive". For this command,
   we don't wait for an IRQ, we just wait for BSY bit to be
   cleared. (polled mode) */

int _ide_get_drv_infos(struct ide_translator *trans)
{ 
  int timeout, status=0;
  k_ui16_t buffer[256];
  int ioaddr = trans->ctrl->ioaddr;
  identify_device_info_t *dinfo;
  int i;


  /* Disable interrupts, set device, and send IDENTIFY command */
  outb(ATA_A_nIEN | ATA_A_4BIT, ioaddr + ATA_DEVICE_CONTROL);
  udelay(1);
  outb(ATA_D_IBM | (trans->device == 0) ? ATA_D_MASTER : ATA_D_SLAVE, 
       ioaddr + ATA_DRIVE);
  outb(ATA_C_ATA_IDENTIFY, ioaddr + ATA_CMD);

  /* Wait for command completion */
  for(timeout = 0; timeout < 30000; timeout++)
    {
      status = inb(ioaddr + ATA_STATUS);
      if(!(status & ATA_S_BSY))
	break;
      
      udelay(1);
    }

  /* Check for DRQ to be set */
  if(!(status & ATA_S_DRQ))
    {
      __dbg_printk("[get_drv_infos] DRQ wasn't set after IDENTIFY\n");
      return -1;
    }

  /* Read data */
  for(i = 0; i < 256; i++)
    buffer[i] = inw(ioaddr);

  dinfo = (identify_device_info_t *) buffer;

  trans->logical_cylinder_nb      = dinfo->nb_logical_cylinders;
  trans->logical_head_nb          = dinfo->nb_logical_heads;
  trans->logical_sector_per_track = dinfo->nb_logical_sectors;
  trans->blk_size                 = 512;

  if(trans->logical_sector_per_track == 0 ||
     trans->logical_head_nb == 0 ||
     trans->logical_cylinder_nb == 0)
    {
      return -1;
    }
  
  trans->lba_capacity = dinfo->lba_capacity;
  trans->flags = 0;

  /* Does this trans really support LBA */
  if(dinfo->capabilities & (1<<1)     /* LBA bit must be set */
     && dinfo->major_version          /* major version must be non 0 */
     && (dinfo->fields_valid & 1)     /* lba fields must be valid */
     && (dinfo->lba_capacity))        /* lba capacity must be non 0 */
    trans->flags |= HARDDISK_LBA_CAPABLE;

  /* Find real capacity */
  if(trans->logical_head_nb == 16 &&
     trans->logical_sector_per_track == 63 &&
     trans->logical_cylinder_nb == 16383)
    {
      if(trans->flags & HARDDISK_LBA_CAPABLE)
	trans->size = trans->lba_capacity;
      else
	FAILED_VERBOSE("Large hd without LBA ...\n");
    }
  else
    trans->size = 
      trans->logical_cylinder_nb * 
      trans->logical_sector_per_track *
      trans->logical_head_nb;

  /* Convert different strings */
  _ide_convert_to_ascii(dinfo->model_number, trans->model_number,40);
  _ide_convert_to_ascii(dinfo->serial_number, trans->serial_number,20);  
  _ide_convert_to_ascii(dinfo->firmware_revision, 
			trans->firmware_revision,8);

  return 0;
}



