/*
 * Copyright (C) 2000, Thomas PETAZZONI
 * http://kos.enix.org
 *
 * Source file for DSR (deferred service routines) management
 *
 * @(#) $Id: _dsr.c,v 1.1.1.1 2003/04/27 12:49:48 thomas Exp $
 */

#include <kos/spinlock.h>
#include <idt/_dsr.h>
#include <kmem/kmem.h>
#include <kos/assert.h>
#include <liblist/liblist.h>
#include <lib/string.h>

/* One queue per CPU */
dsr_queue_t dsr_queue[NB_CPU];

#define RETURN(val) \
        do { restore_flags(flags); return (val); } while(0)

#define QUEUE dsr_queue[current_processor()]

/*
 * 0 when Ok; -1 on error
 */
int register_dsr(int dsr_level, dsr_handler_t *handler, void *data)
{
  k_ui32_t flags;

  ASSERT_RETURN_VAL(handler != NULL, -1);  
  ASSERT_RETURN_VAL((dsr_level >= 0) && (dsr_level < DSR_LEVELS), -1);
  
  disable_local_interrupts(flags);


  if(QUEUE.level[dsr_level].state != UNREGISTERED_DSR)
    RETURN(-1);

  CONCEPTION_ASSERT(QUEUE.level[dsr_level].handler == NULL);
  
  QUEUE.level[dsr_level].handler = handler;
  QUEUE.level[dsr_level].data = data;
  QUEUE.level[dsr_level].state = INACTIVE_DSR;

  RETURN(0);
}

e_dsr_t unregister_dsr(int dsr_level)
{
  k_ui32_t flags;
  
  ASSERT_RETURN_VAL((dsr_level >= 0) && (dsr_level < DSR_LEVELS), E_DSR_ERROR);
  disable_local_interrupts(flags);

  switch(QUEUE.level[dsr_level].state)
    {
    case UNREGISTERED_DSR:
      WARNING("Trying to unregister an unregistered DSR\n");
      RETURN(E_DSR_ERROR);
      break;
      
    case INACTIVE_DSR:
      /* Happy :) */
      break;

    case PENDING_DSR:
      // WARNING("Trying to unregister a pending DSR\n");
    case EXECUTING_DSR:
      // WARNING("Trying to unregister an executing DSR\n");
      RETURN(E_DSR_TRY_AGAIN);
      break;
      
    default:
      FAILED_VERBOSE("Unknown DSR state\n");
    }
  
  CONCEPTION_ASSERT(QUEUE.level[dsr_level].handler);
  CONCEPTION_ASSERT(!get_bit_in_bitmap((addr_t) & QUEUE.bitmap, dsr_level));
    
  QUEUE.level[dsr_level].handler = NULL;
  QUEUE.level[dsr_level].data    = NULL;
  QUEUE.level[dsr_level].state   = UNREGISTERED_DSR;
  
  RETURN(E_DSR_OK);
}


int mark_dsr(int dsr_level)
{
  k_ui32_t flags;
  
  ASSERT_RETURN_VAL((dsr_level >= 0) && (dsr_level < DSR_LEVELS), -1);
  disable_local_interrupts(flags);
  
  if(QUEUE.level[dsr_level].state == UNREGISTERED_DSR)
    RETURN(-1);

  QUEUE.level[dsr_level].state = PENDING_DSR;
  
  set_bit_in_bitmap((addr_t) & QUEUE.bitmap,
		    dsr_level,
		    1);

  RETURN(0);
}


/* Assumes interrupts are disabled. Returns with interrupts disabled */
void spawn_dsr(void)
{
  while(1)
    {
      dsr_t *dsr;
      int toplevel = bsf_bitmap((addr_t) & QUEUE.bitmap, 
				sizeof(QUEUE.bitmap));

      if(toplevel >= DSR_LEVELS)
	break; /* No more DSRs remaining */

      dsr = QUEUE.level + toplevel;

      ASSERT_FATAL(dsr->state == PENDING_DSR);      
      ASSERT_FATAL(dsr && dsr->handler);
      
      /*
       * Clear the bit in order to allow the very same DSR to be
       * marked while it is being executed : It prevents from
       * "forgetting" some data between the end of the DSR and the
       * "cli".
       */
      dsr->state = EXECUTING_DSR;
      set_bit_in_bitmap((addr_t) & QUEUE.bitmap,
			toplevel,
			0);

      /* Interrupts are REALLY disabled (we're during an IRQ handler),
         but we want then enabled during execution of the handlers */
      sti();
      dsr->handler(toplevel, dsr->data);
      cli();
    }
}


int init_dsr(void)
{
  int i;
  
  for(i = 0; i < DSR_LEVELS; i++)
    {
      QUEUE.level[i].data = NULL;
      QUEUE.level[i].handler = NULL;
      QUEUE.level[i].state = UNREGISTERED_DSR;
    }
  
  memset((char *) QUEUE.bitmap,
	 0,
	 sizeof(QUEUE.bitmap));
  
  return 0;
}




