#include <kos/assert.h>
#include <kos/macros.h>
#include <liblist/liblist.h>

#include <idt/_dst.h>
#include <kmem/kmem.h>

dst_t _dst[DST_LEVELS];

struct dst_job *new_job(dst_handler_t *handler, void *data)
{
  dst_job_t *job;

  if (!handler)
    return NULL;
  
  job = (dst_job_t*) kmalloc(sizeof(dst_job_t));
  if (! job)
    return NULL;
  
  job->handler = handler;
  job->data    = data;
  job->state   = INACTIVE_DST;

  return job;
}

int mark_job(struct dst_job *job, int dst_level)
{
  if (! job)
    return -1;

  ASSERT_RETURN_VAL((dst_level >= 0) && (dst_level < DST_LEVELS), -1);

  switch(job->state)
    {
      k_ui32_t flags;

    case PENDING_DST:
      /* Job is already is the queue, and is going to be executed
         later. So we don't add it again */
      break;
    case EXECUTING_DST:
      /* Job is currently executing, so he won't take account of this
         new job. Let's add our job */
    case INACTIVE_DST:
      /* Job is inactive, let's activate it ! */
      
      write_spin_lock(_dst[dst_level].lock, flags);

      job->state = PENDING_DST;
      
      write_spin_unlock(_dst[dst_level].lock, flags);

      if(_dst[dst_level].msg_port == NULL)
	FAILED_VERBOSE("Port does not exist (NULL) for this level !\n");
	  
      if(kmsg_send(_dst[dst_level].msg_port, (void *) job) < 0)
	return -1;
      
      break;
      
    default:
      FAILED_VERBOSE("Unknown DST state\n");
    }
  
  return 0;
}

e_dst_t del_job(struct dst_job *job)
{
  k_ui32_t flags;
  dst_t    *dst;

  if ( (! job) && (! job->owner) )
    return E_DST_ERROR;

  dst = job->owner;

  write_spin_lock(dst->lock, flags);

  switch (job->state) {
  case INACTIVE_DST:
    break; /* Happy :) */
  case PENDING_DST:
  case EXECUTING_DST:
    write_spin_unlock(dst->lock, flags);
    return E_DST_TRY_AGAIN;
  default:
    write_spin_unlock(dst->lock, flags);
    FAILED_VERBOSE("Unauthorized DST state\n");
  }

  job->state = 0xdeadcafe; /* ... waiting for kfree below */
  write_spin_unlock(dst->lock, flags);

  kfree_size_known(job, sizeof(dst_job_t));

  return E_DST_OK;
}

static void __deferred_service_thread(void *data)
{
  k_ui32_t flags;
  dst_t *myself = (dst_t*) data;

  CONCEPTION_ASSERT(myself);

  while(1) {
    dst_job_t *job;

    /* Wait for next job in queue */
    job = (dst_job_t *) kmsg_receive(_dst[myself->dst_level].msg_port);
    
    exclusive_spin_lock(myself->lock, flags);
  
    ASSERT_FATAL(job != NULL);

    CONCEPTION_ASSERT(job->state == PENDING_DST);
    job->state = EXECUTING_DST;

    CONCEPTION_ASSERT(job->handler);

    exclusive_spin_unlock(myself->lock, flags);

    job->handler(myself->dst_level, job->data);
    job->state = INACTIVE_DST;
  }

  FAILED_VERBOSE("Should REALLY do something useful !");
  
}


void init_dst()
{
  int i;

  for (i = 0 ; i < DST_LEVELS ; i++)
    {
      spinlock_init(_dst[i].lock);
      _dst[i].dst_level = i;
      _dst[i].msg_port = kmsg_create_port();
      ASSERT_FATAL(_dst[i].msg_port);
      _dst[i].thread = create_kernel_thread(get_kernel_team(),
					    __deferred_service_thread,
					    (void*) &_dst[i]);
      ASSERT_FATAL(_dst[i].thread);
    }
}

