#include <karm/karm.h>
#include <karm/devfs.h>
#include <karm/view.h>

#include <karm/fs_support.h>
#include <karm/interface.h>
#include <karm/interface/dir.h>
#include <karm/devfs.h>
#include <karm/mount.h>

#include <lib/string.h>
#include <kmem/kmem.h>

#include <kos/assert.h>
#include <kos/macros.h>
#include <kos/errno.h>
#include <liblist/liblist.h>


/** devfs can only be mounted once ! */
static unsigned devfs_mounted = 0;

/** private list of device */
static struct devfs_device *devices_list;

struct devfs_kres_data {
  /* The type of the kres :
   * ROOT_KRES : the type of the / kres
   * SUB_KRES  : the type of the /mem, /disk, etc... kres
   */
  enum { ROOT_KRES, SUB_KRES } type;

  /* In case of a SUB_KRES, group points to the name of the group */
  char *group;
};

/** Registers a device in the device filesystem. 
 *
 *  @param group     The directory inside the device filesystem where the
 *                   device will be registered (for example : 'disk',
 *                   'part', 'sound', etc..).
 *
 *  @param dev_name  The name of the device in the device filesystem.
 *
 *  @param nb_views  The number of interfaces supported by the device.
 *
 *  @param view      The implementation of the supported interfaces.
 *
 *  @param device_id Where to return the ID of the device in the 
 *                   device filesystem.
 *
 *  @return 0 on success, an error code on failure
 */
result_t devfs_register(const char *group, const char *dev_name, 
			unsigned nb_views, struct view *views,
			dev_t *device_id)
{
  struct devfs_device *ndev;

  __dbg_printk("* devfs_register group=%s, name=%s, nb_views=%d, views=%p\n",
	       group, dev_name, nb_views, views);

  /* Sanity checks */
  ASSERT(group != NULL);
  ASSERT(dev_name != NULL);
  ASSERT(views != NULL);
  ASSERT(device_id != NULL);

  ndev = kmalloc(sizeof(struct devfs_device));

  if(ndev == NULL)
    {
      return -ENOMEM;
    }

  ndev->group = strdup(group);
  ndev->dev   = strdup(dev_name);
  ndev->nb_views = nb_views;
  ndev->views    = views;
   
  list_add_tail(devices_list, ndev);

  *device_id = (dev_t) ndev;

  __dbg_printk("* devfs_register successful\n");

  return 0;
}

/** Unregisters a device from the device filesystem
 *
 *  @param device_id The ID of the device to unregister. This ID is
 *                   the one that is returned (in the device_id
 *                   parameter) by the devfs_register() function.
 *
 *  @return 0 on success, an error code on failure
 */
result_t devfs_unregister(dev_t dev_id)
{
  struct devfs_device *dev;
  
  if (dev_id == 0)
    return -1;

  dev = (struct devfs_device *)dev_id;
  
  /** remove the device from the directory */
  list_delete(devices_list, dev);
  kfree(dev->group);
  kfree(dev->dev);
  kfree(dev->views);
  kfree(dev);

  return 0;
}

/** The devfs is mounted. The root point and group directories must
 *  support the directory interface.
 *  
 */

static result_t devfs_lookup(struct ures *ures, const char *name, 
			     struct kres **kr);
static result_t devfs_mkfile(struct ures *ures, const char *name);
static result_t devfs_mkdir(struct ures *ures, const char *name);

struct dir devfs_dir = 
  {
    NULL,
    NULL,
    devfs_lookup,
    devfs_mkfile,
    devfs_mkdir
  };

/** devfs must register in the fs_support 
 */

static result_t devfs_attach(struct ures *block_device_sr, struct kres **kres, 
			     void **fs_infos, void *conf);
static result_t devfs_detach(void *fs_infos);

struct filesystem devfs_fs = 
  {
    .attach = devfs_attach,
    .detach = devfs_detach
  };

struct view view_dir = 
  { .iid=     INTERFACE_DIR_ID,
    .nb_ops=  INTERFACE_DIR_NB_OPS,
    .ops=     &devfs_dir
  };

struct devfs_kres_data devfs_root_kres_data =
  {
    .type  = ROOT_KRES,
    .group = NULL
  };

struct kres devfs_kr = {
  .nb_views=  1,
  .views=     &view_dir,
  .ext=       & devfs_root_kres_data,
  .ures_list= NULL,
  .prev=      NULL,
  .next=      NULL
};

static result_t devfs_attach(struct ures *device, struct kres **kres, 
			     void **fs_infos, void *conf)
{
  UNUSED(conf);
  
  /* In the case of devfs, there is NO underlying device, so this
     pointer should be NULL */
  ASSERT(device == NULL);
  ASSERT(kres != NULL);
  ASSERT(fs_infos != NULL);
  
  /* Devfs is now mounted, and cannot be mounted more than once */
  devfs_mounted = 1;

  /* Since only one mount of devfs is possible, we don't need to
     return any filesystem specific informations */
  *fs_infos     = NULL;

  /* The root kres */
  *kres         = & devfs_kr;

  return 0;
}

static result_t devfs_detach(void *fs_infos)
{
  UNUSED(fs_infos);
  
  return -ENOSYS;
}


static result_t devfs_lookup(struct ures *ures, const char *name, struct kres **kr)
{
  struct devfs_kres_data *data;

  ASSERT_FATAL(ures != NULL);
  ASSERT_FATAL(name != NULL);
  ASSERT_FATAL(kr != NULL);

  data = (struct devfs_kres_data *) ures->kres->ext;

  ASSERT_FATAL(data != NULL);

  if(data->type == ROOT_KRES) 
    {
      /* We are at the root of the filesystem, looking for the correct
	 group. Let's see whether such a group already exists. */
      struct devfs_device *device;
      int i;

      list_foreach(devices_list, device, i)
	{
	  if(!strcmp(device->group, name))
	    {
	      /* Ok, at least one device in this group */
	      struct kres *kres;
	      struct devfs_kres_data *kres_data;

	      kres = kmalloc(sizeof(struct kres));

	      if(kres == NULL) 
		{
		  return -ENOMEM;
		}

	      kres_data = kmalloc(sizeof(struct devfs_kres_data));

	      if(kres_data == NULL)
		{
		  kfree(kres);
		  return -ENOMEM;
		}

	      kres_data->type  = SUB_KRES;
	      kres_data->group = strdup(name);

	      kres->nb_views  = 1;
	      kres->views     = & view_dir;
	      kres->ext       = kres_data;
	      kres->ures_list = NULL;
	      kres->prev = kres->next = NULL;

	      *kr = kres;
	      
	      return 0;
	    }
	}

      /* Group wasn't found */
      return -ENOENT;
    }
  else if(data->type == SUB_KRES)
    {
      /* We are in a group of device, and look for a device inside
	 this group. */
      struct devfs_device *device;
      int i;

      list_foreach(devices_list, device, i)
	{
	  /* We have to find a decide whose group in the good one (the
	     same as the name of the current kres), and whose name
	     matches the given name. */
	  if(!strcmp(device->group, data->group) &&
	     !strcmp(device->dev, name))
	    {
	      /* Ok, at least one device in this group */
	      struct kres *kres;

	      kres = kmalloc(sizeof(struct kres));

	      if(kres == NULL) 
		{
		  return -ENOMEM;
		}

	      kres->nb_views = device->nb_views;
	      kres->views    = device->views;

	      /* TODO : Who initializes this field ?? (maybe nobody
		 because these kres are not filesystem kres ???) */
	      kres->ext               = NULL;
	      kres->ures_list         = NULL;
	      kres->prev = kres->next = NULL;

	      *kr = kres;
	      
	      return 0;
	    }
	}
      
      return -ENOENT;
    }

  return -ENOENT;
}

static result_t devfs_mkfile(struct ures *ures, const char *name)
{
  UNUSED(ures);
  UNUSED(name);
  
  return -EPERM;
}

static result_t devfs_mkdir(struct ures *ures, const char *name)
{
  UNUSED(ures);
  UNUSED(name);
  
  /* You cannot currently create a dir in the devfs */
  return -EPERM;
}

result_t devfs_init(void)
{
  result_t result;
  struct filesystem *fs;

  result = register_supported_fs("devfs", & devfs_fs);

  if(result < 0)
    {
      return result;
    }

  fs = lookup_supported_fs("devfs");

  if(fs == NULL)
    {
      return -ENOENT;
    }

  result = mount(NULL, "/dev", fs, NULL);

  if(result < 0)
    {
      return result;
    }

  return 0;
}
