#include <karm/karm.h>

#include <karm/fs_support.h>
#include <karm/interface.h>
#include <karm/interface/dir.h>
#include <karm/mount.h>

#include <lib/string.h>
#include <kmem/kmem.h>

#include <kos/assert.h>
#include <kos/macros.h>
#include <kos/errno.h>

/* Explanation :
 *
 * The fakefs is a very simple filesystem, completely hard-coded. It
 * only provide a root ('/') and a 'dev' directory, so that we can
 * mount the device filesystem. It will allow us to access real block
 * devices, and so to mount a real filesystem as the root filesystem.
 */

static result_t fakefs_lookup(struct ures *ures, const char *name, 
			      struct kres **kres);

/* DIR interface implementation. We don't implement all functions,
   because we only need some of them */
struct dir fakefs_dir =
  {
    NULL, /* opendir */
    NULL, /* closedir */
    fakefs_lookup,
    NULL, /* mkfile */
    NULL, /* mkdir */
  };

struct view fakefs_view_dir = 
  {
    .iid     = INTERFACE_DIR_ID,
    .nb_ops  = INTERFACE_DIR_NB_OPS,
    .ops     = & fakefs_dir
  };

enum { FAKEFS_ROOT_KRES, FAKEFS_DEV_KRES } fakefs_kres_type;

/* This is the root kres */
struct kres fakefs_root_kres =
  {
    .nb_views  = 1,
    .views     = & fakefs_view_dir,
    .ext       = (void*) FAKEFS_ROOT_KRES,
    .ures_list = NULL,
    .prev      = NULL,
    .next      = NULL
  };

/* This is the dev kres */
struct kres fakefs_dev_kres =
  {
    .nb_views  = 1,
    .views     = & fakefs_view_dir,
    .ext       = (void*) FAKEFS_DEV_KRES,
    .ures_list = NULL,
    .prev      = NULL,
    .next      = NULL    
  };

static result_t fakefs_lookup(struct ures *ures, const char *name, 
			      struct kres **kres)
{
  int type;
  ASSERT(ures != NULL);
  ASSERT(name != NULL);
  ASSERT(kres != NULL);

  if(ures->view->iid != INTERFACE_DIR_ID)
    {
      return -EINVAL;
    }

  type = (int) ures->ext;

  /* We are at the root point of the file system */
  if(type == FAKEFS_ROOT_KRES)
    {
      ASSERT(ures->kres == & fakefs_root_kres);

      /* We only have an hard coded "dev" directory */
      if(! strcmp(name, "dev"))
	{
	  /* Return the hard coded dev kres */
	  *kres = & fakefs_dev_kres;
	  return 0;
	}
    }

  /* In any other cases, nothing found */
  return -ENOENT;
}

static result_t fakefs_attach(struct ures *device, struct kres **kres, 
			      void **fs_infos, void *conf);
static result_t fakefs_detach(void *fs_infos);

struct filesystem fakefs_fs = 
  {
    .attach = fakefs_attach,
    .detach = fakefs_detach
  };

static result_t fakefs_attach(struct ures *device, struct kres **kres, 
			     void **fs_infos, void *conf)
{
  UNUSED(conf);
  
  /* In the case of fakefs, there is NO underlying device, so this
     pointer should be NULL */
  ASSERT(device == NULL);
  ASSERT(kres != NULL);
  ASSERT(fs_infos != NULL);
  
  /* Since only one mount of devfs is possible, we don't need to
     return any filesystem specific informations */
  *fs_infos     = NULL;

  /* The root kres */
  *kres         = & fakefs_root_kres;

  return 0;
}

static result_t fakefs_detach(void *fs_infos)
{
  ASSERT(fs_infos != NULL);
  
  return -ENOSYS;
}

result_t fakefs_init(void)
{
  result_t result;
  struct filesystem *fs;

  result = register_supported_fs("fakefs", & fakefs_fs);

  if(result < 0)
    {
      return result;
    }


  fs = lookup_supported_fs("fakefs");

  if(fs == NULL)
    {
      return -ENOENT;
    }

  result = mount(NULL, "/", fs, NULL);

  if(result < 0)
    {
      return result;
    }

  return 0;
}
