#include <karm/mount.h>
#include <karm/ures.h>
#include <karm/fs_support.h>
#include <karm/nscache.h>
#include <karm/interface.h>

#include <lib/string.h>
#include <kos/errno.h>
#include <kos/assert.h>

/** Mount a device with the given filesystem at the given
 *  mountpoint. 
 *
 *   @param device     the device to be mounted. This function assumes
 *                     that this device is opened using the BLOCK 
 *                     interface.
 *   @param mountpoint the string that specifies where the device
 *                     should be monted in the global namespace.
 *   @param fs         the filesystem driver to use to mount the given 
 *                     device.
 *   @param conf       some configuration for the filesystem. It is just
 *                     passed to the attach() function of the filesystem 
 *                     driver for specific configuration.
 *
 *   @return 0 on success, an error code on failure
 */

result_t mount(struct ures *device, char *mountpoint, struct filesystem *fs, 
	       void *conf)
{
  struct kres       *root_kr;
  void              *fs_infos;
  struct nsnode     *root_node;
  int result;

  ASSERT_RETURN_VAL(mountpoint && fs, -1);

  /* Attach the filesystem */
  if (fs->attach(device, &root_kr, &fs_infos, conf) != 0)
    {
      return -1;
    }

  /* Mount the root filesystem */
  if(!strcmp(mountpoint, "/")) 
    {
      result = nscache_init(root_kr, fs_infos);

      if(result < 0)
	{
	  fs->detach(fs_infos);
	  return result;
	}
    }
  else
    {
      struct nsnode *node;
      struct ures *ures;

      /* Make a open so that the nsnode will get generated */
      result = open(mountpoint, INTERFACE_DIR_ID, & ures);

      if(result < 0)
	{
	  return result;
	}

      /* SYNCHRONIZATION SUBTELTY : Nobody should be able to touch the
	 nscache between the open() aboce and the lookup() below. The
	 open() is here to generate the nsnode in the tree, and the
	 nsnode_lookup will find it.
	 
	 Two solutions : 

          1. The classical synchronization solution : lock these two
             functions, so that nobody can access the nscache in
             between. BUT, if we take the nscache semaphore, it will
             result in a deadlock, because the nscache function will
             try to hold it. This can be solved by using a mutex, but
             it is not very clean.

	  2. Have a 'nsnode' pointer in each kres. Then, just do an
             open, and get the nsnode by doing
             ures->kres->nsnode. That's cleaner, but I am not sure
             that there is a 1:1 mapping between nsnode and kres
             because of symbolic links ?
      */

      result = nscache_lookup(NULL, NULL, mountpoint, NULL, & node);

      if(result < 0) 
	{
	  __dbg_printk("mount(): Lookup of %s failed with %d\n",
		       mountpoint, result);
	  fs->detach(fs_infos);
	  return result;
	}

      __dbg_printk("mount(): Lookup of %s OK ! node=%p\n",
		   mountpoint, node);

      result = nsnode_root_alloc(& root_node);

      if (result < 0)
	{
	  __dbg_printk("mount(): Nsnode_alloc failed with %d\n", 
		       result);
	  fs->detach(fs_infos);
	  return result;
	}

      __dbg_printk("mount(): Nsnode_alloc OK ! root_node=%p\n",
		   root_node);

      root_node->kres = root_kr;
      root_node->mounted_root.fs_infos = (void *)fs_infos;
      root_node->local_root = root_node;

      result = nscache_mount(node, root_node);

      if (result < 0)
	{
	  __dbg_printk("mount(): Nscache_mount failed with %d\n", 
		       result);
	  fs->detach(fs_infos);
	  nsnode_free(root_node);
	  return result;
	}
    }

  __dbg_printk("Mount() successfull!\n");
  
  return 0;
}
