#include <karm/nscache.h>
#include <liblist/liblist.h>
#include <lib/string.h>
#include <kmem/kmem.h>
#include <kos/macros.h>
#include <kos/assert.h>
#include <kos/errno.h>

static struct nsnode *root_nsnode = NULL; // static !!!
static struct kslab_cache *nsnode_cache;

#define NSCACHE_DEBUGGING

#ifdef NSCACHE_DEBUGGING
 #define NSCACHE_DEBUG(fmt,args...) __dbg_printk(fmt, ## args)
#else
 #define NSCACHE_DEBUG(fmt,args...)
#endif

bool_t dirname_eq(char* dir1, char* dir2)
{
  for ( ; *dir1 == *dir2 ; dir1++, dir2++)
    if (*dir1 == 0)
      return TRUE;
#define EODirname(s) ((*(s) == '\0') || (*(s) == '/'))
  return (EODirname(dir1) && EODirname(dir2));
}


size_t dirname_len(const char *str, char ** begin_dirname)
{
  char  *cptr;
  size_t len = 0;

  /* Skip any '/////' before dir name */
  for (*begin_dirname = (char *) str ; 
       **begin_dirname == '/' ; 
       (*begin_dirname)++)
    continue;

  /* Lookup end of current dir name */
  for (cptr = *begin_dirname ;
       (*cptr != '/') && (*cptr != '\0') ;
       cptr++, len++)
    continue;

  CONCEPTION_ASSERT(*begin_dirname <= cptr);

  __dbg_printk("dirname_len: Length returned : %d\n", len);
  return len;
}

/**
 * This function initializes the nscache, by creating the root node of
 * the filesystem hierarchy, with the given informations.
 *
 * kres     : the kernel resource which is related to the root of the
 *            filesystem hierarchy.
 * fs_infos : the mounted filesystem informations
 *
 * Returns 0 on success, an error code on failure
 */
result_t nscache_init (struct kres *kres, void *fs_infos)
{
  nsnode_cache = kslab_cache_create("nsnode_cache", sizeof(struct nsnode), 0, 0, 0);
  if (nsnode_cache == NULL)
    FAILED_VERBOSE("(kares) nscache could not create the nsnode_cache.\n");

  root_nsnode = (struct nsnode *)kslab_cache_alloc(nsnode_cache);
  if (root_nsnode == NULL)
    FAILED_VERBOSE("nscache couldn't allocate the fake root !\n");

  memset(root_nsnode, 0x0 , sizeof(struct nsnode));

  root_nsnode->name = (char *)kmalloc(sizeof(char) * 2);
  if (root_nsnode->name == NULL)
    FAILED_VERBOSE("nscache couldn't allocate the fake root !\n");

  strzcpy(root_nsnode->name,"/", 2);

  root_nsnode->kres                    = kres;
  root_nsnode->parent                  = root_nsnode;
  root_nsnode->children                = NULL;
  root_nsnode->local_root              = root_nsnode;
  root_nsnode->mountpoint.mounted_root = NULL;
  root_nsnode->mounted_root.fs_infos   = fs_infos;
  root_nsnode->mounted_root.mountpoint = NULL;
  
  return 0;
}

/* lookup an entry belonging to current, having the given name, which
   may be a usual AZT string, or a string ended by '/'. Return the
   nsnode found, or NULL;. Always update end_dirname with the address
   of the remaining string */
result_t nscache_lookup_entry(struct nsnode *process_root,
			      struct nsnode *current,
			      const char *name,
			      char **end_dirname,
			      struct nsnode **out_node)
{
  struct nsnode *tmp_nsnode;
  unsigned nb;
  char *begin_dirname;

  NSCACHE_DEBUG("nscache_lookup_entry called with process_root=%p, current=%p, "
	       "name=%s, end_dirname=%p, out_node=%p\n",
	       process_root, current, name, end_dirname, out_node);

  NSCACHE_DEBUG("root_nsnode->name = %s, root_nsnode->children=%p\n",
		root_nsnode->name, root_nsnode->children);

  ASSERT_FATAL(end_dirname != NULL);

  nb = dirname_len(name, & begin_dirname);
  *end_dirname = begin_dirname + nb;

  if (nb <= 0)
    {
      *out_node = current;
      return -ENOENT;
    }

  NSCACHE_DEBUG("Comparing with '.'\n");

  if (dirname_eq(begin_dirname, "."))
    {
      *out_node = current;
      return 0;
    }

  NSCACHE_DEBUG("Comparing with '..'\n");
      
  if (dirname_eq(begin_dirname, ".."))
    {
      /* We don't want to get out of the chroot jail */
      if(process_root == current)
	{
	  *out_node = current;
	  return 0;
	}

      /* Because we can mount many times at the same place, there
         could be a chain of mounted_root/mountpoint. We go through
         this chain up to the first one to get the *real* parent */
      while (NSNODE_IS_MOUNTED_ROOT(current))
	current = current->mounted_root.mountpoint;

      CONCEPTION_ASSERT(current->parent);
      
      *out_node = current->parent;

      return 0;
    }

  NSCACHE_DEBUG("Looking for the entry in the list (current=%p)\n", current);

  /* Lookup the current dirname in the existing children nsnodes */
  list_foreach(current->children, tmp_nsnode, nb)
    {
      NSCACHE_DEBUG(" Visiting node %s (mountpoint.mounted_root=%p)\n", 
		    tmp_nsnode->name, tmp_nsnode->mountpoint.mounted_root);

      if(dirname_eq(begin_dirname, tmp_nsnode->name))
	{
	  /* Because we can mount many times at the same place, there
	     could be a chain of mounted_root/mountpoint. We go through
	     this chain down to the last one to get the *real*
	     mounted_root */
	  while (NSNODE_IS_MOUNTPOINT(tmp_nsnode))
	    tmp_nsnode = tmp_nsnode->mountpoint.mounted_root;
	  
	  *out_node = tmp_nsnode;

	  NSCACHE_DEBUG(" --> Returning node %s @ %p\n", tmp_nsnode->name,
			tmp_nsnode);

	  return 0;
	}
    }

  NSCACHE_DEBUG("nscache_lookup_entry exit with -ENOENT\n");

  *out_node = current;  
  return -ENOENT;
}


result_t nscache_lookup(struct nsnode *process_root,
			struct nsnode *cwd,
			const char *path,
			char **out_end_dirname,
			struct nsnode **out_node)
{
  char          *begin_dirname;
  char          *end_dirname;
  struct nsnode *cur_nsnode;

  NSCACHE_DEBUG("nscache_lookup started with process_root=%p, cwd=%p, path=%s, "
	       "end_dirname=%p, outnode=%p\n",
	       process_root, cwd, path, out_end_dirname, out_node);

  CONCEPTION_ASSERT(root_nsnode);

  if (process_root == NULL)
    process_root = root_nsnode;

  if(cwd == NULL)
    cur_nsnode = process_root;
  else
    cur_nsnode = cwd;

  begin_dirname = (char *) path;
  end_dirname  = (char *) path;

  while(1)
    {
      char * tmp_end = NULL;
      result_t result;
      struct nsnode * tmp_nsnode;

      NSCACHE_DEBUG("Calling nscache_lookup_entry with %s (tmp_end=%s)\n", 
		    begin_dirname, tmp_end);

      result = nscache_lookup_entry(process_root,
				    cur_nsnode,
				    begin_dirname,
				    & tmp_end,
				    & tmp_nsnode);

      NSCACHE_DEBUG("End of lookup_entry with tmp_end=%s\n", tmp_end);

      /* The next step wasn't found, just return the last node found */
      if (result < 0 || dirname_len(tmp_end, & tmp_end) == 0)
	{
	  NSCACHE_DEBUG("On va sortir bientot\n");
	  if(out_end_dirname)
	    *out_end_dirname = end_dirname;
	  *out_node        = tmp_nsnode;
	  NSCACHE_DEBUG("nscache_lookup Ending with end_dirname=%s and node=%p\n",
			end_dirname, tmp_nsnode);
	  return 0;
	}

      cur_nsnode = tmp_nsnode;
      begin_dirname = tmp_end;
      end_dirname  = tmp_end;
    }

  /* We should never get there */
  CONCEPTION_ASSERT(FALSE);

  /* Make Gcc happy */
  return -1;
}

result_t nsnode_root_alloc(struct nsnode **out_node)
{
  struct nsnode *node;

  node = (struct nsnode *) kslab_cache_alloc (nsnode_cache);

  if (node == NULL)
    return -ENOMEM;

  memset(node, 0x0, sizeof(struct nsnode));

  node->name    = kmalloc(strlen("/") + 1);

  if (node->name == NULL)
    {
      kslab_cache_free (nsnode_cache, node);
      return -ENOMEM;
    }

  node->name[0] = '/';
  node->name[1] = 0;

  *out_node = node;

  return 0;
}

result_t nsnode_alloc(const char * name, struct nsnode **out_node)
{
  char   *begin_dirname;
  size_t len;
  struct nsnode * node;

  NSCACHE_DEBUG("nsnode_alloc called with name=%s, out_node=%p\n",
		name, out_node);

  len = dirname_len(name, & begin_dirname);
  if (len <= 0)
    {
      return -EBADNAME;
    }

  NSCACHE_DEBUG("nsnode_alloc name length=%d\n", len);

  node = (struct nsnode *) kslab_cache_alloc (nsnode_cache);

  if (node == NULL)
    return -ENOMEM;

  memset(node, 0x0, sizeof(struct nsnode));

  NSCACHE_DEBUG("nsnode_alloc node=%p\n", node);

  node->name = kmalloc((len + 1) * sizeof(char));

  if (node->name == NULL)
    {
      kslab_cache_free (nsnode_cache, node);
      return -ENOMEM;
    }

  NSCACHE_DEBUG("nsnode_alloc node=%p, node->name=%p\n", 
		node, node->name);

  strzcpy(node->name, begin_dirname, len + 1);

  NSCACHE_DEBUG("nsnode_alloc node=%p, node->name=%s\n", 
		node, node->name);

  *out_node = node;

  return 0;
}


result_t nsnode_free(struct nsnode *n)
{
  kfree(n->name);
  return kslab_cache_free(nsnode_cache, n);
}


result_t nscache_attach(struct nsnode *parent,
			struct nsnode *tree)
{
  NSCACHE_DEBUG("nscache_attach called with parent=%p, tree=%p\n",
		parent, tree);

  /* If the following tests fail, this is because you should call
     nsnode_mount() instead */
  ASSERT_RETURN_VAL(! NSNODE_IS_MOUNTPOINT(parent), -1);
  ASSERT_RETURN_VAL(! NSNODE_IS_MOUNTED_ROOT(tree), -1);

  /* Make sure we move the tree inside the address space of the very
     same driver (aka same file system) */
  ASSERT_RETURN_VAL(tree->local_root == parent->local_root, -1);

  list_add_tail(parent->children, tree);
  tree->parent = parent;

  NSCACHE_DEBUG("nscache_attach done\n");

  return 0;
}


result_t nscache_detach(struct nsnode *tree)
{
  struct nsnode *parent = tree->parent;

  /* If the following tests fail, this is because you should call
     nsnode_umount() instead */
  ASSERT_RETURN_VAL(! NSNODE_IS_MOUNTPOINT(parent), -1);
  ASSERT_RETURN_VAL(! NSNODE_IS_MOUNTED_ROOT(tree), -1);

  list_delete(parent->children, tree);
  tree->parent = NULL;

  return 0;
}

result_t nscache_mount(struct nsnode *mountpoint,
		       struct nsnode *mounted_root)
{
  NSCACHE_DEBUG("nscache_mount called with mounpoint=%p (name=%s), mounted_root=%p (name=%s)",
		mountpoint, mountpoint->name, 
		mounted_root, mounted_root->name);
  
  /* Make sure mountpoint is not yet considered as a mountpoint */
  ASSERT_RETURN_VAL(! NSNODE_IS_MOUNTPOINT(mountpoint), -1);
  /* Make sure the tree is not already mounted somewhere */
  ASSERT_RETURN_VAL(! NSNODE_IS_MOUNTED_ROOT(mounted_root), -1);
  /* Make sure the mounted root is the root of the name space of its
     driver */
  ASSERT_RETURN_VAL(mounted_root->local_root == mounted_root, -1);

  mounted_root->mounted_root.mountpoint = mountpoint;
  mountpoint->mountpoint.mounted_root = mounted_root;

  return 0;
}


result_t nscache_umount(struct nsnode *mounted_root)
{
  struct nsnode *mountpoint = mounted_root->mounted_root.mountpoint;

  /* Make sure mounted_root is a mounted root */
  ASSERT_RETURN_VAL(NSNODE_IS_MOUNTED_ROOT(mounted_root), -1);
  CONCEPTION_ASSERT(mountpoint != NULL);

  /* Make sure mountpoint is a mount point */
  ASSERT_RETURN_VAL(NSNODE_IS_MOUNTPOINT(mountpoint), -1);

  ASSERT_RETURN_VAL(mountpoint->mountpoint.mounted_root == mounted_root, -1);

  /* Make sure the mounted root is the root of the name space of its
     driver */
  ASSERT_RETURN_VAL(mounted_root->local_root == mounted_root, -1);

  mounted_root->mounted_root.mountpoint = NULL;
  mountpoint->mountpoint.mounted_root   = NULL;

  return 0;
}

result_t nsnode_rename(struct nsnode *node, char *new_name)
{
  char *begin_dirname;
  char *str;
  size_t len;

  CONCEPTION_ASSERT(node && node->name && new_name);

  len = dirname_len(new_name, & begin_dirname);

  if (len <= 0)
    {
      return -EBADNAME;
    }

  str = kmalloc((len + 1) * sizeof(char));

  if (str == NULL)
    return -1;

  kfree(node->name);
  node->name = str;

  strzcpy(node->name, begin_dirname, len + 1);

  return 0;  
}

bool_t nsnode_can_be_deleted(struct nsnode *node)
{
  CONCEPTION_ASSERT(node);

  if(NSNODE_IS_MOUNTPOINT(node) || NSNODE_IS_MOUNTED_ROOT(node))
    return FALSE;

  if(node->chroot_cnt != 0 || node->cwd_cnt != 0)
    return FALSE;

  if(node->children != NULL)
    return FALSE;

  return TRUE;
}

