#ifndef __NSNODE_H__
#define __NSNODE_H__

#include <kos/types.h>
#include <karm/karm.h>


/** Namespace node : canonical item for the kernel representation of
    the namespace tree */
struct nsnode
{
  char *name; /**< Kmallocated */

  struct kres *kres; /**< Must never be NULL */

  /** Root of the current name space (NOT the real root of the whole
      system !) */
  struct nsnode *local_root;

  /** Parent of the current node. For the root of the whole system, it
      points to the node itself */
  struct nsnode *parent;
  
  /** Children of the current node */
  struct nsnode *children;
  /** Sibling nodes */
  struct nsnode *next, *prev;

#define NSNODE_IS_MOUNTPOINT(node)   ((node)->mountpoint.mounted_root != NULL)
#define NSNODE_IS_MOUNTED_ROOT(node) ((node)->mounted_root.mountpoint != NULL)

  /** Number of processes that has this nsnode as chroot */
  unsigned chroot_cnt;

  /** Number of processes that has this nsnode as current working
      directory */
  unsigned cwd_cnt;

  /** When NSNODE_IS_MOUNTPOINT() == true */
  struct
  {
    struct nsnode          *mounted_root;
  } mountpoint;
  
  /** When NSNODE_IS_ROOT() == true */
  struct
  {
    struct nsnode          *mountpoint; 
    void                   *fs_infos;    /**< filesystem private informations needed by the driver */
  } mounted_root;

};


/**
 * This function initializes the nscache, by creating the root node of
 * the filesystem hierarchy, with the given informations.
 *
 * kres     : the kernel resource which is related to the root of the
 *            filesystem hierarchy.
 * fs_infos : the mounted filesystem informations
 *
 * Returns 0 on success, an error code on failure
 */
result_t nscache_init (struct kres *kres, void *fs_infos);

/** Looks up the entry 'name' as a child of 'current' if
    available. Returns the corresponding node and update end_dirname
    which contains the remaining path of the object. If the entry does
    not exist, it's up to the driver of the current namespace to
    create it (if possible) or not. */
result_t nscache_lookup_entry(struct nsnode *process_root,
			      struct nsnode *current, 
			      const char *name, 
			      char **end_dirname,
			      struct nsnode **out_node);

/** Looks up as much entries in the path as possible, given the
    process_root and cwd. Returns the most further node up to which we
    were able to resolve the path, and update end_dirname which
    contains the remaining path of the object. If part of the path
    does not exist, it's up to the driver of the current namespace to
    create them (if possible) or not. */
result_t nscache_lookup(struct nsnode *process_root,
			struct nsnode *cwd,
			const char *path,
			char **remaining_path,
			struct nsnode **out_node);

/** Insert 'tree' into the hierarchy below parent. Parent shouldn't be
    a mountpoint, and tree shouldn't be a mounted_root (use
    mount/umount for that). */
result_t nscache_attach(struct nsnode *parent,
			struct nsnode *tree);

/** Remove 'tree' from the hierarchy. Tree shouldn't be a mounted_root
    (use mount/umount for that). */
result_t nscache_detach(struct nsnode *tree);

/** Insert the namespace beginning at 'mounted_root' at 'mountpoint'
    in the global namespace (mountpoint can in turn be a mounted_root
    itself). */
result_t nscache_mount(struct nsnode *mountpoint,
		       struct nsnode *mounted_root);

/** Remove the namespace beginning at 'mounted_root' from the global
    namespace. */
result_t nscache_umount(struct nsnode *mounted_root);

/** Check if dir1 is equivalent to dir2, according to namespace
    standards : we check until we reach the end of both strings (ie
    '\0' or '/') */
bool_t dirname_eq(char* dir1, char* dir2);

/** Returns the length of 'str', according to namespace standards : we
    skip the leading slashes, and we stop at the end of the string (ie
    '/' or '\0'). 'begin_dirname' contains the beginning of the real
    name (after the leading slashes). */
size_t dirname_len(const char *str, char ** begin_dirname);

/** Tells whether the 'node' can be deleted or not (checks whether it
    is a mount point, a mounted root, a node that is used as current
    working directory or as chroot, or a node that has children) */
bool_t nsnode_can_be_deleted(struct nsnode *node);

/** Change the name of 'node' by 'new_name', without any check (assume
    conflicts are managed by caller). */
result_t nsnode_rename(struct nsnode *node, char *new_name);

/** Allocate a new node with the given name */
result_t nsnode_alloc(const char * name, struct nsnode **out_node);

/** Allocate a new node for the root node of a filesystem. There is a
    special function, because the normal nsnode_alloc function doesn't
    work when generating node with "/" name. It is much cleaner to
    have something different, I think. (thomas)
*/
result_t nsnode_root_alloc(struct nsnode **out_node);

/** Free the node and its name. No checks. */
result_t nsnode_free(struct nsnode *n);


#endif /* __NSNODE_H__ */






