#include <karm/open.h>
#include <karm/ures.h>

#include <karm/interface.h>
#include <karm/interface/dir.h>

#include <karm/nscache.h>
#include <karm/kres.h>

#include <lib/string.h>
#include <kos/assert.h>
#include <kos/macros.h>

/** Open an object with a specific interface. The object is specified
    using a location in the global namespace (e.g "/dev/disk/hda"),
    and the interface is specified using an unique interface ID.
    
    @param path      The location of the object to open in the global
                     namespace.

    @param interface The ID of the interface with which we want to
                     open the object.

    @param ures      The location where open() should return the pointer
                     to the new user ressource.

    @return 0 on success, an error code on failure
*/
result_t open(const char *path, unsigned interface, struct ures **ures)
{
  char *remaining_path;
  char *begin_dirname;
  result_t result;
  int stop = FALSE;

  struct kres *kres;
  struct ures *tmp_ures;
  struct nsnode *cnode;
  struct nsnode *tnode;

  /** Sanity checks */
  ASSERT(path != NULL);
  ASSERT(ures != NULL);

  __dbg_printk("* open: trying to open %s with interface %d (ures=%p)\n",
	       path, interface, ures);

  /** Find the deepest node already opened in the cache */
  result = nscache_lookup(NULL,NULL,path, &remaining_path, & cnode);

  __dbg_printk("open: nscache_lookup returned %d\n", result);

  /** Loop to open all the directories until we reach the leaf
      resource to open */
  while (! stop)
    {
      size_t len;

      __dbg_printk("Remaining_path=%s\n", remaining_path);
      
      len = dirname_len(remaining_path, &begin_dirname);

      /* Did we reach the leaf ressource to open ? */
      if(begin_dirname[len] == '\0')
	{
	  stop = TRUE;
	}
      else 
	{
	  begin_dirname[len] = '\0';  /** < gorish hack to build a zascii string */
	}
      
      __dbg_printk("open: Look for %s\n", begin_dirname);
      
      /** if the current item is '.' or '..' ask the nscache to move
	  in the node tree.
      */
      if ((strcmp(begin_dirname, "." ) == 0) 
	  || (strcmp(begin_dirname, "..") == 0))
	{
	  begin_dirname[len] = '/';     /** < restore the path */
	  result = nscache_lookup(NULL, cnode, begin_dirname, &remaining_path,
				  & cnode);
	  
	  /* The above call should always return success */
	  ASSERT_FATAL(result >= 0); 
	  
	  break;
	}

      __dbg_printk("Open the current node as a directory : cnode=%p\n", 
		   cnode);

      /** open the current node as a directory */
      result = ures_open(cnode->kres, INTERFACE_DIR_ID, & tmp_ures);
      
      if(result < 0) 
	{
	  /* TODO : undo everything that was done */
	  return result;
	}

      __dbg_printk("Allocate a new node and open the next temporary node\n");

      /** allocate a new node and open the next temporary node */
      result = nsnode_alloc(begin_dirname, & tnode);
      
      if(result < 0)
	{
	  /* TODO : undo what was done */
	  return result;
	}

      __dbg_printk("Lookup the directory\n");
      result = DIR_OPS(tmp_ures->view->ops)->lookup(tmp_ures, begin_dirname, &kres);
      
      /** free the useless directory node */
      ures_close(tmp_ures);
      
      if(result < 0)
	{
	  /* TODO : undo what was done */
	  return result;
	}
      
      /** setup the new temporary node */
      tnode->kres = kres;
      tnode->local_root = cnode->local_root;
      
      begin_dirname[len] = '/';     /** < restore the path */
      
      /** insert the new node as a current node leaf */
      nscache_attach(cnode, tnode);
      
      cnode = tnode; 
      remaining_path = begin_dirname + len;
    }

  __dbg_printk("current node = %s\n", cnode->name);
  
  /** open the leaf resource with the proper interface */
  result = ures_open(cnode->kres, interface, & tmp_ures);
  
  if(result < 0)
    {
      /* TODO : undo what was done */
      return result;
    }

  tnode->kres       = kres;
  tnode->local_root = cnode->local_root;
  
  /** insert the ures the kres list of sres */
  kres_add_ures(kres, tmp_ures);

  /** insert the new node in the cache */
  nscache_attach(cnode, tnode);
  
  *ures = tmp_ures;

  __dbg_printk("OPEN RETURNED SUCCESFULLY WITH %p\n", tmp_ures);

  return 0;
}
