#ifndef __URES_H__
#define __URES_H__

#include <kos/errno.h>

/** An user ressource is closely related to a kernel ressource. An
    user ressource correspond to a kernel ressource in use by a
    specific task. In KOS, you open a kernel ressource (= an object in
    the system), using a specific interface. When this open is done,
    an user ressource is created.

    So, an user ressource is mainly an association of :
    * a kernel ressource
    * a point of view for this kernel ressource (= interface)
    * private data

    Private data in the kernel ressource are global to the system,
    while the private data in the user ressource are local to the
    open() of a file. */

typedef struct ures {

  /** The associated kernel ressource */
  struct kres *kres;

  /** The point of view for the kernel ressource : it is a pointer to
      the interface implementation that was choosen during the
      open() */
  struct view *view;

  /** Private data, specific to this user ressource.  For example, in
      the case of a 'file', we may have the file pointer (current
      position in the file). It contains the information specific to
      this open() of the kernel ressource. */
  void *ext;

  /** Pointers to user ressources that allows to create a chained list
      of user ressources per kernel ressource. */
  struct ures *prev, *next;
} ures_t;

/** Open an user ressource, corresponding to the given kernel
 *  ressource with the given interface.
 *
 *  @param kres      The kernel ressource for which we want to generate a
 *                   user ressource.
 *
 *  @param interface The ID of the interface with which we want to
 *                   open the kernel ressource.
 *
 *  @param ures      Where to return the pointer of the new user ressource
 *
 *  @result 0 on success, an error code on failure
*/
result_t ures_open(struct kres *kres, unsigned interface, struct ures **ures);

/** Close a user ressource
 *
 *  @param ures The user ressource to close
 *
 *  @return 0 on success, an error code on failure
 */
result_t ures_close(struct ures *ures);

#endif /* __URES_H__ */
