/*
 * Copyright (C) 2001, Thomas Petazzoni
 * http://kos.enix.org
 *
 * Kitc : kmsg part
 * 
 * @(#) $Id: _kmsg.c,v 1.1.1.1 2003/04/27 12:49:49 thomas Exp $
 */

#include <kitc/_kmsg.h>
#include <kos/types.h>
#include <kmem/kmem.h>
#include <liblist/liblist.h>
#include <kos/assert.h>

kmsg_port_t *kmsg_create_port(void)
{
  kmsg_port_t *new_port;
  
  new_port = kmalloc(sizeof(kmsg_port_t));
  
  if(! new_port)
    return NULL;
  
  new_port->msg_list = NULL;
  new_port->receive_state = UNBLOCKED;
  spinlock_init(new_port->lock);
  ksignal_init(&new_port->ksig);

  return new_port;
}

int kmsg_delete_port(kmsg_port_t *kmsg_port)
{
  ASSERT_FATAL(kmsg_port != NULL);
  
  if (kmsg_port->msg_list != NULL)
    return -1;

  ksignal_destroy(&kmsg_port->ksig);
  
  kfree_size_known(kmsg_port, sizeof(kmsg_port));

  return 0;
}

int kmsg_send(kmsg_port_t *port, void *msg)
{
  kmsg_t *new_msg;
  k_ui32_t flags;

  ASSERT_FATAL(port != NULL);
  ASSERT_FATAL(msg != NULL);
  
  new_msg = kmalloc(sizeof(kmsg_t));
  if(!new_msg)
    return -1;
  
  new_msg->msg = msg;
  
  write_spin_lock(port->lock, flags);
  list_add_tail(port->msg_list, new_msg);

  /* If anyone is blocked while receiving on this port, signal it ! */
  if(port->receive_state == BLOCKED)
    {
      port->receive_state = UNBLOCKED;
      write_spin_unlock(port->lock, flags);
      ksignal_wakeup(&port->ksig);
    }
  else
    write_spin_unlock(port->lock, flags);

  return 0;
}

void *kmsg_receive(kmsg_port_t *port)
{
  kmsg_t *msg;
  k_ui32_t flags;
  void *message;
  
  ASSERT_FATAL(port != NULL);

  /* Only one thread can be blocked on receiving at a time */
  ASSERT_RETURN_VAL(port->receive_state != BLOCKED, NULL);

  write_spin_lock(port->lock, flags);  

  if(port->msg_list)
    {
      msg = port->msg_list;
      message = msg->msg;
      list_delete(port->msg_list, msg);
    }
  else 
    {
      port->receive_state = BLOCKED;
      write_spin_unlock(port->lock, flags);

      ksignal_wait(&port->ksig);

      write_spin_lock(port->lock, flags);
      /* When returning from ksignal_wait, list should be not empty */
      ASSERT_FATAL(port->msg_list != NULL);
      msg = port->msg_list;
      message = msg->msg;
      list_delete(port->msg_list, msg);
      kfree(msg);
    }
  
  write_spin_unlock(port->lock, flags);
  return message;
}

void *kmsg_poll(kmsg_port_t *port)
{
  kmsg_t *msg;
  k_ui32_t flags;
  void *message;

  ASSERT_FATAL(port != NULL);

  write_spin_lock(port->lock, flags);

  if(port->msg_list)
    {
      msg = port->msg_list;
      message = msg->msg;
      list_delete(port->msg_list, msg);
      write_spin_unlock(port->lock, flags);
      kfree(msg);

      return message;
    }
  else
    write_spin_unlock(port->lock, flags);

  return NULL;
}

