#include <kitc/_ksem.h>
#include <liblist/liblist.h>
#include <task/task.h>

#include <kos/types.h>
#include <kos/assert.h>
#include <debug/debug.h>

spinlock_t global_semaphore_lock;

int ksem_create(struct ksem *sem, int pshared, unsigned int value)
{
  k_ui32_t flags;

  UNUSED(pshared);
  CONCEPTION_ASSERT(sem != NULL);

  write_spin_lock(global_semaphore_lock, flags);
  
  if (sem->state == SEM_READY)
    {
      write_spin_unlock(global_semaphore_lock, flags);
      return -1;
    }

  sem->type = (pshared == 0 ? P_LOCAL : P_SHARED);
  sem->mvalue = value;
  sem->value  = value;
  sem->waiting_thread_list = NULL;
  spinlock_init(sem->lock);
  sem->state = SEM_READY;

  write_spin_unlock(global_semaphore_lock, flags);

  return 0;
}


int ksem_destroy(struct ksem *sem)
{
  k_ui32_t flags;
  CONCEPTION_ASSERT(sem != NULL);

  write_spin_lock(sem->lock, flags);

  if ((sem->state != SEM_READY) 
      || (sem->value != 0) 
      || (sem->waiting_thread_list != NULL))
    {
      write_spin_unlock(sem->lock, flags);
      return -1;
    }

  sem->state = SEM_ZOMBIE;
  sem->mvalue = 0;
  write_spin_unlock(sem->lock, flags);
  
  return 0;
}

int ksem_post(struct ksem *sem)
{
  struct thread *pop_thread;
  k_ui32_t flags;
  CONCEPTION_ASSERT(sem != NULL);

  write_spin_lock(sem->lock, flags);
  
  if (sem->state != SEM_READY)
    {
      write_spin_unlock(sem->lock, flags);
      return -1;
    }
  
  if (sem->waiting_thread_list != NULL)
    {
      pop_thread = sem->waiting_thread_list;
      list_delete_named(sem->waiting_thread_list,
			pop_thread,
			u.waiting_sem_thread.prev,
			u.waiting_sem_thread.next);

      pop_thread->u.waiting_sem_thread.prev = NULL;
      pop_thread->u.waiting_sem_thread.next = NULL;
      pop_thread->u.waiting_sem_thread.sem  = NULL;
      pop_thread->state = INTERRUPTED;
    }
  else
    {
      if ((sem->value + 1) > sem->mvalue)
	{
	  write_spin_unlock(sem->lock, flags);
	  return -1;
	}
      sem->value++;
    }
  
  write_spin_unlock(sem->lock, flags);
  
  return 0;
}


int ksem_wait(struct ksem *sem)
{
  k_ui32_t flags;
  struct thread *cthread;
  CONCEPTION_ASSERT(sem != NULL);

  write_spin_lock(sem->lock, flags);
  
  if (sem->state != SEM_READY)
    {
      write_spin_unlock(sem->lock, flags);
      return -1;
    }
  
  if (sem->value == 0)
    {
      cthread = get_current_thread();
      CONCEPTION_ASSERT(cthread);

      cthread->state = WAITING_SEM;
      cthread->u.waiting_sem_thread.sem = sem;
      list_add_tail_named(sem->waiting_thread_list,
			  cthread,
			  u.waiting_sem_thread.prev,
			  u.waiting_sem_thread.next);
      
      write_spin_unlock(sem->lock, flags);
      reschedule_run_next();
    }
  else
    {
      sem->value--;
      write_spin_unlock(sem->lock, flags);
    }
  
  return 0;
}


int ksem_trywait(struct ksem *sem)
{
  k_ui32_t flags;
  CONCEPTION_ASSERT(sem != NULL);

  write_spin_lock(sem->lock, flags);
  
  if (sem->state != SEM_READY)
    {
      write_spin_unlock(sem->lock, flags);
      return -1;
    }
  
  if (sem->value == 0)
    {
      write_spin_unlock(sem->lock, flags);
      return EAGAIN;
    }
  else
    {
      sem->value--;
    }

  write_spin_unlock(sem->lock, flags);
  
  return 0;
}


int ksem_getvalue(struct ksem *sem, int *sval)
{
  CONCEPTION_ASSERT(sem != NULL);

  *sval = (int)sem->value;
  return 0;
}

__init_text int _ksem_init(void)
{
  spinlock_init(global_semaphore_lock);
  return 0;
}
