/*
 * Copyright (C) 2000, Thomas Petazzoni
 * http://kos.enix.org
 *
 * Allocating space in a cache, Bonwick implementation
 *
 * @(#) $Id: _kslab_cache_alloc.c,v 1.2 2003/04/27 20:55:55 thomas Exp $
 */

#include <kmem/_kslab.h>
#include <kos/assert.h>

#define UNLOCK_CACHE(cache) \
        do { write_spin_unlock((cache)->lock, lock_flags);  } while(0)
#define LOCK_CACHE(cache) \
        do { write_spin_lock((cache)->lock, lock_flags); } while(0)
  
void *kslab_cache_alloc(kslab_cache_t *cache)
{
  k_ui32_t lock_flags;
  bool_t   need_grow = FALSE;
  void *block;

  /*
  DEBUG_PRINT("Cache alloc %s avail=%d virt=0x%x phys=0x%x\n", 
  	      cache->name, cache->nb_available, (unsigned)cache, 
  	      (unsigned)arch_get_paddr_at_vaddr((vaddr_t)cache));
  */

  LOCK_CACHE(cache);

  /* See if we should program a slab enlargement */
  if ((cache->grow_threshold > 0) && !cache->growth_in_progress)
      cache->growth_in_progress
	= need_grow
	= (cache->nb_available-1 <= cache->grow_threshold);

  /* Make sure there's something in the full free list, either by
     adding a slab to the cache, or by moving slab from the free list
     to the full free list */
  if(!(cache->semi_full))
  {
    if(!(cache->free))
      {
	addr_t new_space;

	/* BE CAREFULL: WE SHOULD NEVER GET THERE WHILE ALLOCATING AN
           rmap STRUCT !!! */

	/* Allocate space for the new slab */
	new_space = kvalloc(cache->nb_pages_per_slab, 
			    (cache->flags & SLAB_IS_SWAPPABLE),
			    TRUE);
	if(__kslab_cache_grow(cache, new_space) < 0)
	  {
	    UNLOCK_CACHE(cache);
	    return NULL;
	  }

      }
    __kslab_slab_move_head(cache->free,
			   cache->semi_full);
  }
  CONCEPTION_ASSERT(cache->semi_full != NULL);

#define CACHE_FREE_LIST cache->semi_full->free

  CONCEPTION_ASSERT(CACHE_FREE_LIST != NULL);

  /* We're sure that cache->semi_full is a slab with free blocks. The
     block CACHE_FREE_LIST is the first free block. */
  block = CACHE_FREE_LIST;

  CACHE_FREE_LIST = CACHE_FREE_LIST->next;

  cache->semi_full->nb_free--;

  cache->nb_available--;

  /* If slab is full, move it to the correct list */
  if(cache->semi_full->nb_free == 0)
    {
      __kslab_slab_move_head(cache->semi_full,
			     cache->full);
    }

  UNLOCK_CACHE(cache);

  /* Do we need to allocate some slabs in advance ? */
  if (need_grow)
    {
      addr_t new_space;

      /* Allocate space for the new slab */
      new_space = kvalloc(cache->nb_pages_per_slab, 
			  (cache->flags & SLAB_IS_SWAPPABLE),
			  TRUE);
      LOCK_CACHE(cache);
      __kslab_cache_grow(cache, new_space);
      cache->growth_in_progress = FALSE;
      UNLOCK_CACHE(cache);
    }

  return block;
}


