/*
 * Copyright (C) 2000, Thomas PETAZZONI
 * http://kos.enix.org
 *
 * Create a cache in the kslab allocator
 * Cache line optimization, and different computations inspired by Linux 
 * source code
 *
 * @(#) $Id: _kslab_cache_create.c,v 1.1.1.1 2003/04/27 12:49:49 thomas Exp $
 */

#include <_kslab.h>
#include <liblist/liblist.h>
#include <kos/spinlock.h>
#include <lib/string.h>

kslab_cache_t *kslab_cache_create(char *name,
				  size_t size,
				  int align,
				  int flags,
				  k_ui32_t grow_threshold)
{
  kslab_cache_t *new_cache;
  unsigned int waste = 0;
  int original_size = size;

  /* Align size on 4 bytes */
  size = ALIGN_SUP(size, BYTES_PER_WORD);

  /* Take 'align' param into account */
  if(align != 0)
    size = ALIGN_SUP(size, align);
  
  new_cache = kslab_cache_alloc(&cache_of_kslab_cache);

  ASSERT_FATAL(new_cache != NULL);

  memset(new_cache, 0, sizeof(kslab_cache_t));

  strncpy(new_cache->name, name, KSLAB_CACHE_NAME_LENGTH);
  new_cache->flags = flags & KSLAB_FLAG_MASK;
  spinlock_init(new_cache->lock);
  new_cache->original_size = original_size;
  
  /* Small size objets => using on slab */
  if(size < (PAGE_SIZE>>7))
      new_cache->flags |= ON_SLAB;
  else
    new_cache->flags |= OFF_SLAB;

  /* Try to get the best alignement with cache */
  if(flags & KSLAB_FLAG_CACHE_ALIGN)
    {
      /* Two objets cannot fit into cache line */
      if(size > (L1_CACHE_SIZE/2))
	size = ALIGN_SUP(size, L1_CACHE_SIZE);

      /* We can put at least two elements per cache line.. great :) */
      else
	{
	  int nb_elts_per_line = L1_CACHE_SIZE/size;
	  waste = L1_CACHE_SIZE - (nb_elts_per_line*size);

	  if(waste)
	    {
	      /* Elts does not fit exactly into cache, so ajust size */
	      if(waste%nb_elts_per_line)
		{
		  nb_elts_per_line--;
		  waste += size;
		}
	      size += (waste/nb_elts_per_line);
	    }
	}
    }

  if(flags & KSLAB_FLAG_IS_SWAPPABLE)
    new_cache->flags |= SLAB_IS_SWAPPABLE;

  new_cache->nb_pages_per_slab = 0;

  /* Trying to determine what is the best number of pages per slab */
  do
    {
      new_cache->nb_pages_per_slab += 1;
      
      if(new_cache->nb_pages_per_slab >= MAX_PAGES_PER_SLAB)
	break;

      waste = new_cache->nb_pages_per_slab*PAGE_SIZE;
      if(new_cache->flags & ON_SLAB)
	waste -= sizeof(kslab_slab_t);
      
      new_cache->nb_elts_per_slab = waste / size;
      waste -= new_cache->nb_elts_per_slab*size;

      if(new_cache->nb_elts_per_slab < MIN_ELTS_PER_SLAB)
	continue;
      
      if(waste < ALIGN_SUP(sizeof(kslab_slab_t), 
			   L1_CACHE_SIZE))
	break;

      /* Allowing 3% of lost space */
      if((waste*10000/(new_cache->nb_pages_per_slab*PAGE_SIZE)) < 300)
	break;
    } while(1);

  new_cache->size = size;

  if(new_cache->nb_pages_per_slab*PAGE_SIZE - 
     new_cache->nb_elts_per_slab*new_cache->size >= sizeof(kslab_slab_t))
    flags |= ON_SLAB;
  
#ifdef KSLAB_DEBUG
  __dbg_printk(_B_BLUE "[kslab] cache create size %d, nb_pages_per_slab %d, nb_elts_per_slab %d, lost space %d/10000, waste %d \n" _B_NORM,
	       new_cache->size,
	       new_cache->nb_pages_per_slab,
	       new_cache->nb_elts_per_slab,
	       (waste*10000)/(new_cache->nb_pages_per_slab*PAGE_SIZE),
	       waste);
#endif

  new_cache->full        = NULL;
  new_cache->free        = NULL;
  new_cache->semi_full   = NULL;

  /* If the threshold is bigger than the nb_elements per page,
     indicate error */
  if (grow_threshold >= new_cache->nb_elts_per_slab)
    {
      kslab_cache_free(&cache_of_kslab_cache, new_cache);
      return NULL;
    }
  new_cache->nb_available   = 0;
  new_cache->grow_threshold = grow_threshold;
  new_cache->growth_in_progress = FALSE;

  /* Pre-grow the new cache in order to avoid lazy allocation, in
     order not to delay the first allocations in the new cache */
  if (__kslab_cache_grow(new_cache,   /* Allocate space for the new slab */
			 kvalloc(new_cache->nb_pages_per_slab, 
				 (new_cache->flags & SLAB_IS_SWAPPABLE),
				 TRUE)) < 0) {
    kslab_cache_free(&cache_of_kslab_cache, new_cache);
    return NULL;
  }

  write_spin_lock(kernel_kslab_lock, flags);
  list_add_tail(kslab_cache_list, new_cache);
  write_spin_unlock(kernel_kslab_lock, flags);
  
  return new_cache;
}
