/*
 * Copyright (C) 2000, Thomas PETAZZONI, David DECOTIGNY
 * http://kos.enix.org
 *
 * _kvmem_alloc.c
 * Kernel virtual memory range allocation
 *
 * @(#) $Id: _kvmem_alloc.c,v 1.1.1.1 2003/04/27 12:49:49 thomas Exp $
 */

#include <kos/types.h>
#include <kmem/_kvmem.h>
#include <kos/spinlock.h>
#include <pmm/pmm.h>
#include <kos/assert.h>
#include <arch/mm/mm.h>

#define RETURN(ret) \
        do { write_spin_unlock(kernel_kvalloc_lock, flags); \
             return ret; } while(0)


addr_t kvalloc(size_t nb_pages, bool_t is_swappable, bool_t needs_mapping)
{
  chained_page_range_t *new_range;
  chained_page_range_t *current_range;
  addr_t virt_base_addr = 0;
  k_ui32_t flags;

  if(nb_pages <= 0)
    return 0;

  write_spin_lock(kernel_kvalloc_lock, flags);

  current_range = __kvmem_find_suitable_range_in_free_list(nb_pages);
  if (! current_range) {
    DEBUG_PRINT("[kvalloc] No VM left for %d pages\n", nb_pages);
    RETURN(0);
  }
  /* If range has exactly the same size, just move it to used list */
  if(current_range->nb_pages == nb_pages) 
    {
      __kvmem_remove_range_from_free_list(current_range);
      __kvmem_add_range_to_used_list(current_range);
      virt_base_addr = current_range->start;
    }

  /* otherwise the range is bigger than the size requested, split it !
     reduce its size, and allocate a new range, which is going to be
     added to used list */
  else
    {
      ASSERT_FATAL(current_range->nb_pages > nb_pages);

      /* __dbg_printk("Splitting range with start 0x%x\n",
	 current_range->start); */

      new_range = __kvmem_allocate_new_unused_range();
      if (! new_range)
	RETURN(0);
      
      new_range->start    = current_range->start;
      new_range->nb_pages = nb_pages;
      current_range->start    += nb_pages * PAGE_SIZE;
      current_range->nb_pages -= nb_pages;
      
      __kvmem_add_range_to_used_list(new_range);
      virt_base_addr =  new_range->start;
    }
  
  /* If we're close to the MINIMAL_SURVIVAL_RANGE_NB, allocate
             one more page, for more ranges */
  if(total_range_nb == MINIMAL_SURVIVAL_RANGE_NB) {
    write_spin_unlock(kernel_kvalloc_lock, flags);
    if(__kvmem_add_new_page_of_ranges() < 0)
      return 0;
  }
  else
    write_spin_unlock(kernel_kvalloc_lock, flags);

  /* Map virtual pages to physical pages if needed */
  if(needs_mapping)
    {
      if(__kvmem_map_range(virt_base_addr, nb_pages, is_swappable) != 0)
	{
	  DEBUG_PRINT("No more physical memory to map pages");
	  RETURN(0);
	}
    }

  return virt_base_addr;
}
