/*
 * Copyright (C) 2000, Thomas PETAZZONI, David DECOTIGNY
 * http://kos.enix.org
 *
 * _kvmem_free.c
 * Kernel virtual memory range liberation
 *
 * @(#) $Id: _kvmem_free.c,v 1.1.1.1 2003/04/27 12:49:49 thomas Exp $
 */

#include <kos/types.h>
#include <kmem/_kvmem.h>
#include <kos/spinlock.h>
#include <kos/macros.h>
#include <pmm/pmm.h>
#include <debug/debug.h>
#include <kos/assert.h>
#include <arch/mm/mm.h>
#include <liblist/liblist.h>

inline int __kvfree_unsafe(addr_t range_start, bool_t needs_unmapping)
{
  chained_page_range_t *current_range;
  chained_page_range_t *this_range = NULL;
  int nb_elements;

  /* Find range in used_page_range_list list */

  list_foreach(used_page_range_list, current_range, nb_elements)
    if(current_range->start == range_start)
      {
	this_range = current_range;
	__dbg_printk(_B_BLUE "[kvfree] Range found with start 0x%x, size %d\n" _B_NORM,
		       this_range->start, this_range->nb_pages);
	break;
      }

  /* Check whether a range with start range_start has been found */

  if(this_range == NULL)
    return -1;

  /* Unmap this range if needed */
  if(needs_unmapping)
    __kvmem_unmap_range(this_range);

  /* Remove this range from the used_page_range_list list */
  list_delete(used_page_range_list, this_range);

  if (! free_page_range_list) {
    list_add_head(free_page_range_list, this_range);
    return 0;
  }

  /* Look for preceding range in free_page_range_list list, for insertion */
  if (free_page_range_list->start > this_range->start)
    current_range = NULL;
  else 
    {
      list_foreach(free_page_range_list, current_range, nb_elements)
	{
	  if ((current_range->start < this_range->start)
	      &&
	      ((! current_range->next) || (current_range->next->start > this_range->start)))
	    break;
	}
        if (! list_foreach_early_break(free_page_range_list, current_range,
                                       nb_elements))
          current_range = NULL;
    }


  /* Check whether the range preceding does not overlap on the range
     we want to insert */
  ASSERT_FATAL((! current_range)
	       ||
	       (RANGE_END(current_range) <= this_range->start));
  
  /* Must insert before the head AND this range is next to the current
     head => expand the head range */
  if((! current_range)
     &&
     (RANGE_END(this_range) == free_page_range_list->start))
    {
      __dbg_printk("Range is the first and is close to its next\n");
      free_page_range_list->start     = this_range->start;
      free_page_range_list->nb_pages += this_range->nb_pages;
      __kvmem_reclaim_free_range(this_range);
      __dbg_printk("After reclaiming this_range\n");
    }

  /* Or this range is to be inserted before the head, but not next to
     this head */
  else if(! current_range)
    {
      __dbg_printk("Range is the first, and not close to its first\n");
      list_add_head(free_page_range_list, this_range);
    }

  /* If there is a range after the one we want to insert, and that
     this range is close to its next, and that the preceding range is
     close to the new range, merge those three ranges */

  else  if (current_range->next && 
	    (RANGE_END(this_range) == current_range->next->start) && 
	    (RANGE_END(current_range) == this_range->start))
    {
      chained_page_range_t *tmp_range;
      
      __dbg_printk("Range is near its next and its prev\n");

      current_range->nb_pages += current_range->next->nb_pages
	                         + this_range->nb_pages;

      tmp_range  = current_range->next;

      /* Strange thing : if we pass current_range->next as argument
         for remove_range_from_free_list, it seems to become NULL
         after the call, thing which does not occur when using a
         temporary tmp_range */

      __kvmem_remove_range_from_free_list(tmp_range);
      __kvmem_reclaim_free_range(tmp_range);
      __kvmem_reclaim_free_range(this_range);
    }

  /* The current range is close to the next, merge them */
  else if (current_range->next &&
	   (RANGE_END(this_range) == current_range->next->start))
    {
      __dbg_printk("Range is near its next\n");
      current_range->next->start     = this_range->start;
      current_range->next->nb_pages += this_range->nb_pages;
      __kvmem_reclaim_free_range(this_range);
    }

  else if (RANGE_END(current_range) == this_range->start)
    {
      //      __dbg_printk("Range is near its prev\n");
      current_range->nb_pages += this_range->nb_pages;
      __kvmem_reclaim_free_range(this_range);
    }
  else
    {
      // __dbg_printk("Range is near nothing\n");
      list_insert_after(current_range, this_range);
    }

  return 0;
}


int kvfree(addr_t range_start, bool_t needs_unmapping)
{
  k_ui32_t flags;
  int ret;

  write_spin_lock(kernel_kvalloc_lock, flags);
  ret = __kvfree_unsafe(range_start, needs_unmapping);
  write_spin_unlock(kernel_kvalloc_lock, flags);
  
  return ret;
}

