#include <kos/types.h>
#include <kmem/_kvmem.h>
#include <kos/spinlock.h>
#include <loader/mod.h>
#include <pmm/pmm.h>
#include <arch/mm/mm.h>
#include <kos/assert.h>
#include <kos/system.h>
#include <liblist/liblist.h>
#include <lib/string.h>

chained_page_range_t *free_page_range_list;
chained_page_range_t *used_page_range_list;
page_of_range_t *page_of_range_list;
k_ui32_t total_range_nb;

struct chained_page_range * __kvmem_get_used_page_range_list(void)
{
  return used_page_range_list;
}

/* Function called during init_module's, so there's nothing to lock :
   interrupts are disabled */

__init_text int __kvmem_init(kernel_parameter_t *kp)
{
  page_of_range_t *first_page_of_range;
  chained_page_range_t *new_range;
  kernel_module_t *current_module;
  paddr_t paddr;

  total_range_nb = 0;
  page_of_range_list = NULL;
  used_page_range_list = NULL;
  free_page_range_list = NULL;

  first_page_of_range = 
    (page_of_range_t *) PAGE_ALIGN_SUP(kp->allocated_memory_top_virt_addr);
  paddr = get_physical_page(PHYS_PAGE_KERNEL, PHYS_PAGE_NON_SWAPPABLE);
  RETURN_VAL_IF_FAIL(paddr, -1);
  
  map_virtual_page(NULL,
		   (addr_t)first_page_of_range,
		   paddr,
		   VM_ACCESS_WRITE | VM_ACCESS_READ);

  /* Reset the page of range */
  memset(first_page_of_range, 0, PAGE_SIZE);
  first_page_of_range->nb_ranges = RANGES_PER_PAGE;
  total_range_nb = RANGES_PER_PAGE;
  list_add_head(page_of_range_list, first_page_of_range);

  /* mark the range after page_of_range_list as free */
  new_range = __kvmem_allocate_new_unused_range();
  CONCEPTION_ASSERT(new_range != NULL);
  new_range->start    = (addr_t)page_of_range_list + PAGE_SIZE;
  new_range->nb_pages =
    PAGE_ALIGN_SUP(CORE_KERNEL_VIRTUAL_ADDR + 
		   CORE_KERNEL_VIRTUAL_SIZE - 
		   new_range->start) / PAGE_SIZE;
  list_add_head(free_page_range_list, new_range);

  /* mark range for gpfm + kernel_params + modules_params as used */
  new_range = __kvmem_allocate_new_unused_range();
  CONCEPTION_ASSERT(new_range != NULL);
  ASSERT_FATAL(CORE_KERNEL_VIRTUAL_ADDR > _get_gpfm_ram_map_size());
  new_range->start    = CORE_KERNEL_VIRTUAL_ADDR - _get_gpfm_ram_map_size();
  new_range->nb_pages =
    PAGE_ALIGN_SUP(((vaddr_t)kp->modules_head) - new_range->start)
                    / PAGE_SIZE;
  __kvmem_add_range_to_used_list(new_range);

  /* mark the ranges associated with each module as used */
  for(current_module = kp->modules_head;
      current_module != NULL;
      current_module = current_module->next) {
    new_range = __kvmem_allocate_new_unused_range();
    CONCEPTION_ASSERT(new_range != NULL);
    new_range->start    = (addr_t) current_module;
    new_range->nb_pages = PAGE_ALIGN_SUP(current_module->size) / PAGE_SIZE;
    __kvmem_add_range_to_used_list(new_range);
  }

  /* Mark the range assocated with the bootstrap stack as used */
  new_range           = __kvmem_allocate_new_unused_range();
  CONCEPTION_ASSERT(new_range != NULL);
  new_range->start    = kp->bootstrap_stack_base_addr;
  new_range->nb_pages = PAGE_ALIGN_SUP(BOOTSTRAP_STACK_SIZE) / PAGE_SIZE;
  __kvmem_add_range_to_used_list(new_range);

  /* Mark the range assocated with the video mapping as used */
  new_range           = __kvmem_allocate_new_unused_range();
  CONCEPTION_ASSERT(new_range != NULL);
  new_range->start    = kp->video_memory;
  new_range->nb_pages = VIDEO_MEM_MAP_SIZE / PAGE_SIZE;
  __kvmem_add_range_to_used_list(new_range);

  /* Mark the range assocated with the .init section as used */
  new_range           = __kvmem_allocate_new_unused_range();
  CONCEPTION_ASSERT(new_range != NULL);
  new_range->start    = kp->init_heap_vaddr;
  new_range->nb_pages = PAGE_ALIGN_SUP(kp->init_heap_size) / PAGE_SIZE;
  __kvmem_add_range_to_used_list(new_range);

  /* mark the page associated with the page_of range as used */
  new_range = __kvmem_allocate_new_unused_range();
  CONCEPTION_ASSERT(new_range != NULL);
  new_range->start = (addr_t) page_of_range_list;
  new_range->nb_pages = 1;
  __kvmem_add_range_to_used_list(new_range);


#ifdef KVMEM_DEBUG
  __kvmem_show_all_ranges();
#endif  

  return 0;
}
