/*
 * Copyright (C) 2000, Petazzoni Thomas
 * http://kos.enix.org
 *
 * Utility functions for the paged-base kernel allocator.
 *
 * @(#) $Id: _kvmem_utils.c,v 1.1.1.1 2003/04/27 12:49:49 thomas Exp $
 */

#include <kos/types.h>
#include <kmem/_kvmem.h>
#include <kmem/kmem.h>
#include <kos/spinlock.h>
#include <debug/debug.h>
#include <pmm/pmm.h>
#include <arch/mm/mm.h>
#include <kos/assert.h>
#include <liblist/liblist.h>
#include <lib/string.h>

int __kvmem_add_new_page_of_ranges(void)
{
  page_of_range_t *new_range_page;

  /* We can use kvalloc because we are sure that they're still at
     least MINIMAL_SURVIVAL_RANGE_NB ranges in order to call this
     function */

  new_range_page = (page_of_range_t *) kvalloc(1, FALSE, TRUE);
  if(! new_range_page)
    return -1;
      
  memset((addr_t *)new_range_page, 0, PAGE_SIZE);

  new_range_page->nb_ranges = RANGES_PER_PAGE;
  total_range_nb += RANGES_PER_PAGE;

  /* Add page to the queue */
  list_add_tail(page_of_range_list, new_range_page);
  
  return 0;
}

int __kvmem_reclaim_page_of_ranges(page_of_range_t *range_page)
{

  /* First, check that number of ranges is really 255 */
  if(range_page->nb_ranges != RANGES_PER_PAGE)
    return -1;

  total_range_nb -= RANGES_PER_PAGE;

  /* Then, remove page from the list */
  list_delete(page_of_range_list, range_page);
  
  /* Then, free the page */
  if(__kvfree_unsafe((addr_t) range_page, TRUE) < 0)
    {
      DEBUG_PRINT("[Reclaim_page_of_ranges] : Failure while freeing\n");
      return -1;
    }
  return 0;
}


void __kvmem_unmap_range(chained_page_range_t *cp_range)
{
  range_t range;
  range.start = cp_range->start;
  range.end   = range.start + cp_range->nb_pages*PAGE_SIZE;
  
  unmap_virtual_range(NULL, & range);
}


int __kvmem_map_range(addr_t virt_base_addr, int nb_pages, 
		      bool_t is_swappable)
{
  int i;
  for (i = 0 ; i < nb_pages ; i++, virt_base_addr += PAGE_SIZE) {
    addr_t page_addr = get_physical_page(PHYS_PAGE_KERNEL,
					 (is_swappable)?PHYS_PAGE_SWAPPABLE:PHYS_PAGE_NON_SWAPPABLE);

    if (! page_addr)
      return -1;
    
    map_virtual_page(NULL, virt_base_addr, page_addr,
		     VM_ACCESS_WRITE | VM_ACCESS_READ);
  }

  return 0;
}


chained_page_range_t *__kvmem_allocate_new_unused_range(void)
{
  int i;
  page_of_range_t *current;
  unsigned int nb_elts;

  list_foreach(page_of_range_list, current, nb_elts)
    {
      if(current->nb_ranges <= 0)
	continue;

      for(i = 0 ; i < RANGES_PER_PAGE ; i++) {
	if (IS_RANGE_FREE(current->ranges + i)) {

	  current->nb_ranges--;
	  total_range_nb--;

	  ASSERT_FATAL(current->nb_ranges < RANGES_PER_PAGE);
	  ASSERT_FATAL(IS_RANGE_FREE(current->ranges + i));

	  return current->ranges + i;
	}
      }

    }
  return NULL;
}


int __kvmem_add_range_to_used_list(chained_page_range_t *range)
{
  chained_page_range_t *current_range;
  unsigned int nb_elements;

  ASSERT_FATAL(range->start != 0);

  /* when used_list is empty */
  if (! used_page_range_list) {
    list_add_head(used_page_range_list, range);
    return 0;
  }

  /* otherwise, does the range have to be inserted at the head of
     used_page_range_list ? */
  if (used_page_range_list->start > range->start) {
    /* yes: range is at the head of used_page_range_list */
    list_add_head(used_page_range_list, range);

    /* Check for non-overlapping */
    ASSERT_FATAL(RANGE_END(used_page_range_list) <= range->start);
    
    return 0;
  }

  /* Otherwise look for a right place where to insert range */
  list_foreach(used_page_range_list, current_range, nb_elements)
    {
      if((current_range->start < range->start)
	 && 
	 ((! current_range->next) || 
	  (current_range->next->start > range->start)))
	break;
    }

  /* Found ? Not found ? */
  if (! list_foreach_early_break(used_page_range_list, current_range,
                                 nb_elements))
    current_range = NULL;

  /* Check for non-overlapping */
  ASSERT_FATAL((! current_range)
	       ||
	       (RANGE_END(current_range) <= range->start));

  /* Insert after current_range or at end of list */
  if (current_range)
    list_insert_after(current_range, range);
  else
    list_add_tail(used_page_range_list, range);
       
  return 0;
}


void __kvmem_reclaim_free_range(chained_page_range_t *range) 
{
  page_of_range_t *range_page;

  ASSERT_FATAL(range != NULL);

  /* mark this range as free for future use */
  SET_RANGE_FREE(range);

  range_page = (page_of_range_t*) PAGE_ALIGN_INF((addr_t)range);
  range_page->nb_ranges++;
  total_range_nb++;

  __dbg_printk("[Reclaim_free_range]:Nb_range for this page (0x%x) is %d\n", 
		 (int) range_page, range_page->nb_ranges);

  ASSERT_FATAL((range_page->nb_ranges > 0) && 
	       (range_page->nb_ranges <= RANGES_PER_PAGE));


  if(range_page->nb_ranges == RANGES_PER_PAGE)
    __kvmem_reclaim_page_of_ranges(range_page);
}

/* Remove range from free_list, but does not free the range ! */

int __kvmem_remove_range_from_free_list(chained_page_range_t *range)
{
  ASSERT_FATAL(range != NULL);

  list_delete(free_page_range_list, range);

  return 0;
}

/* The allocation startegy is "first fit" by default */
inline chained_page_range_t * __kvmem_find_suitable_range_in_free_list(unsigned int nb_pages) 
{
  chained_page_range_t *range;
  unsigned int nb_elements;
  
  list_foreach(free_page_range_list, range, nb_elements)
    if(range->nb_pages >= nb_pages)
      return range;

  return NULL;
}

void __kvmem_show_all_ranges(void)
{
  chained_page_range_t *current_range;
  unsigned int nb_elements;
  k_ui32_t flags;

  read_spin_lock(kernel_kvalloc_lock, flags);

  __dbg_printk(_B_GREEN "==> Used ranges\n" _B_NORM);

  list_foreach(used_page_range_list, current_range, nb_elements)
    __dbg_printk(_B_GREEN "    Range (0x%x) start : 0x%x, size = %d, end = 0x%x\n" _B_NORM,
		   (int) current_range,
		   current_range->start, 
		   current_range->nb_pages,
		   RANGE_END(current_range));

  __dbg_printk(_B_GREEN "==> Free ranges\n" _B_NORM);
  list_foreach(free_page_range_list, current_range, nb_elements)
    __dbg_printk(_B_GREEN "    Range (0x%x) start : 0x%x, size = %d, end = 0x%x\n" _B_NORM,
		   (int)current_range,
		   current_range->start, 
		   current_range->nb_pages,
		   RANGE_END(current_range));

  read_spin_unlock(kernel_kvalloc_lock, flags);
}



