/*
 * Copyright (C) 2000, Thomas PETAZZONI & David DECOTIGNY
 * http://kos.enix.org
 *
 * Library for creating and accessing splay trees
 *
 * @(#) $Id: _splay.c,v 1.1.1.1 2003/04/27 12:49:49 thomas Exp $
 */

#include <loader/mod.h>
#include <libbst/libbst.h>
#include <libbst/_splay.h>
#include <kos/macros.h>


/*
 * This function splays a node to the root, even if this node does not
 * exist in the tree. In this case, the neighboor of the node is
 * splayed at the root.
 */
static tree_node_t *splay_node (unsigned int key, tree_node_t *tree) 
{
  tree_node_t n;
  tree_node_t *l; // 'l' stands for left
  tree_node_t *r; // 'r' stands for right
  tree_node_t *y;
  
  if(tree == NULL) return tree;
  
  n.up = n.left = n.right = NULL;
  l = r = &n;
  
  for(;;)
    {
      // key < tree->key, we go left
      if(key < tree->key)
	{
	  // no left child ?
	  if(tree->left == NULL) break;
	  // is key < to the key of the left child of the 'tree' node ?
	  // rotate right
	  if(key < tree->left->key) 
	    {
	      y = tree->left;

	      if (y->right) y->right->up = tree;
	      tree->left = y->right;

	      tree->up = y;
	      y->right = tree;

	      tree = y;
	      if(tree->left == NULL) break;
	    }
	  // link right
	  tree->up = r;
	  r->left = tree;
	  r = tree;
	  tree = tree->left;
	}
      // key > tree->key, we go right
      else if (key > tree->key) 
	{
	  // no right child ?
	  if(tree->right == NULL) break;
	  // is key > to the key of the right child of the 'tree' node ?
	  // rotate left
	  if(key > tree->right->key)
	    {
	      y = tree->right;

	      if (y->left) y->left->up = tree;
	      tree->right = y->left;

	      tree->up = y;
	      y->left = tree;

	      tree = y;
	      if(tree->right == NULL) break;
	    }
	  // link left
	  tree->up = l;
	  l->right = tree;
	  l = tree;
	  tree = tree->right;
	}
      else 
	break;
    }

  if (tree->left) tree->left->up = l;
  l->right = tree->left;

  if (tree->right) tree->right->up = r;
  r->left = tree->right;

  if (n.right) n.right->up = tree;
  tree->left = n.right;

  if (n.left) n.left->up = tree;
  tree->right = n.left;

  tree->up = NULL;
  
  return tree;  
}

/* This function insert a node into a tree. If the node already
 *  exists, it remains the tree unchanged. If the tree is empty, it
 *  creates the tree with the new node. 
 */

tree_node_t *_splay_insert_node (tree_node_t *new_node, tree_node_t *tree)
{

  unsigned int key = new_node->key;
  
  // if the tree is empty, current node is the beginning of a new tree... a 
  // new story :)
  if(tree == NULL) {
    new_node->left = new_node->right = new_node->up = NULL;
    return new_node;
  }
  // we splay the futur neighboor of the node which has to be inserted, 
  // so it comes to the root !
  tree = splay_node(key, tree);
  
  // current node is < than his neighboor
  if(key < tree->key)
    {
      if (tree->left) tree->left->up = new_node;
      new_node->left = tree->left;

      tree->up = new_node;
      new_node->right = tree;

      tree->left = NULL;
      return new_node;
    }
  // current node is > than his neighboor
  else if (key > tree->key)
    {
      if (tree->right) tree->right->up = new_node;
      new_node->right = tree->right;

      tree->up = new_node;
      new_node->left = tree;

      tree->right = NULL;
      return new_node;
    }
  else
    {
      return tree;
    }
  
  // we should never get there
  return tree;
  
  
}

/* This function delete a node into a tree. If the node does not
 *  exists, the tree remains unchanged 
 */

tree_node_t *_splay_delete_node (tree_node_t *remove_node, tree_node_t *tree)
{
  unsigned int key = remove_node->key;
  
  tree_node_t *x;
  
  if(tree == NULL) return tree;
  tree = splay_node(key, tree);
  if(key == tree->key) 
    {
    if(tree->left == NULL) 
      {
	x = tree->right;
      }
    else 
      {
	x = splay_node(key, tree->left);
	if (tree->right) tree->right->up = x;
	x->right = tree->right;
      }
    return x;
    }
  return tree;
    
}


/* Walk through node calling visitor for each node. We stop the
   walk-through as soon as visitor returns TRUE.
   It isn't allowed to uncallocate node nor change left, right, up.
*/
tree_node_t* _splay_visit(tree_node_t *node, 
			  visitor_t visitor, 
			  void *client_data)
{
  enum { LEFT, FIND_RIGHT } go = LEFT;
  tree_node_t *n, *r;

  n = node;
  r = NULL;
  while (n) {

    if (go == LEFT) {

      if (visitor(n, client_data)) return n;

      if (n->left)
	n = n->left;

      else if (n->right)
	n = n->right;

      else {
	r = n;
	n = n->up;
	go = FIND_RIGHT;
      }

      continue;

    } else {

      if ((n->right) && (n->right != r)) {
	n = n->right;
	r = NULL;
	go = LEFT;

      } else {

	r = n;
	n = n->up;
      }

    }
  }
      
  return NULL;
}

/* This visitor should be used to unallocate each node of the given tree */
int _splay_reset (tree_node_t *tree,
		  visitor_t cleaner,
		  void *user_data)
{
  tree_node_t *old = tree;

  while(tree)
    {
      tree = delete_node (tree, tree);
      cleaner(old, user_data);
      old = tree;
    }

  return 0;
}



struct tree_node *_splay_lookup (tree_node_t *tree,
				 compare_fct_t compare_fct,
				 void *client_data)
{
  tree_node_t *node;
  int res;

  if (! compare_fct || ! client_data)
    return NULL;

  for (node = tree;
       node && ((res = compare_fct(node, client_data)) != 0);
       node = (res < 0) ? node->left : node->right)
    continue;

 
  return splay_node (node->key, tree);
}
			    

EXPORT_FUNCTION(_splay_insert_node);
EXPORT_FUNCTION(_splay_delete_node);
EXPORT_FUNCTION(_splay_visit);
EXPORT_FUNCTION(_splay_reset);

