#include <libhash/_libhash.h>
#include <liblist/liblist.h>

#include <loader/mod.h>
#include <kos/macros.h>

#include <lib/string.h>
#include <kmem/kmem.h>
#include <kos/assert.h>
#include <debug/debug.h>

extern unsigned int hash_sizes[];
extern unsigned int hash_nsizes;

struct kslab_cache *hash_item_cache;

#define MAX_HASH_TABLE_SIZE 4294967296 /* 2^32 */

int
hash_init(void)
{
  hash_item_cache = kslab_cache_create("hash item cache",
				       sizeof(struct hash_item),
				       0, 0, 0);
  if (!hash_item_cache)
    return -1;
  
  return 0;
}

struct hash *
hash_create_table(unsigned int nitems, compare_fn cmpf, destruct_fn destructf)
{
  struct hash *h;
  unsigned int i;
  unsigned int prime = 0;

  if (nitems > MAX_HASH_TABLE_SIZE)
    return NULL;
  
  for (i = 0 ; i < hash_nsizes ; i++)
    {
      if (hash_sizes[i] > nitems)
	{
	  prime = hash_sizes[i];
	  break;
	}
    }
  
  if (prime == 0)
    return NULL;
  
  h = kmalloc(sizeof(struct hash));
  if (!h)
    return NULL;

  h->prime = prime;
  h->cmp   = cmpf ? cmpf : NULL ;
  h->destruct = destructf?destructf:NULL;

  for ( i = 0 ; i < prime ; i++)
    {
      h->table[i] = (struct hash_item *)kmalloc(sizeof(struct hash_item *));
    }
  
  return h;
}

int hash_delete_table(struct hash *h)
{
  struct hash_item *tmp;
  unsigned int i;

  ASSERT_RETURN_VAL(h,-1);
  
  for ( i = 0  ; i < h->prime ; i++ )
    {
      list_collapse(h->table[i],tmp)
	{
	  if (h->destruct)
	    h->destruct(tmp->private);
	  else
	    kfree(tmp->private);
	  kslab_cache_free(hash_item_cache,tmp);
	}
    }
  
  kfree(h);

  return 0;
}

int
hash_set_cmp_fn(struct hash *h, compare_fn cmpf)
{
  ASSERT_RETURN_VAL(cmpf && h, -1);
  h->cmp = cmpf;
  
  return 0;
}


unsigned int
hash_std_str_to_int(struct hash *h, const char *str)
{
  const char *p;
  unsigned i = 0, j;

  ASSERT(h && str);
  
  for ( p = str ; *p ; p++ )
    {
      i = (i<<4) + (*p);
      if ((j = (i & 0xf0000000)))
	{
	  i = i ^ (j >> 24);
	  i = i ^ j;
	}
    }

  return i;
}


int 
hash_insert(struct hash *h, unsigned int key, void *private)
{
  struct hash_item *hash_item;

  ASSERT_RETURN_VAL(h && private, -1);

  if (key > ( h->prime - 1))
    return -1;

  hash_item = kslab_cache_alloc(hash_item_cache);
  if (!hash_item)
    return -1;

  hash_item->private = private;
  
  list_add_head(h->table[key],hash_item);

  DEBUG_PRINT_VERBOSE("hash item inserted\n");

  return 0;
}

int
hash_remove(struct hash *h, unsigned int key, void *cmpd)
{
  struct hash_item *tmp;
  unsigned int n;
  int found = !1;
  
  ASSERT_RETURN_VAL(h && h->cmp, -1);
  
  if ((key > (h->prime - 1)))
    return -1;
  
  list_foreach((struct hash_item *)h->table[key],tmp,n)
    {
      found = h->cmp(cmpd,tmp->private);
      if (found)
	break;
    }
  
  if (!found)
    return -1;
  
  list_delete((struct hash_item *)h->table[key],tmp);
  if (h->destruct)
    h->destruct(tmp->private);
  else
    kfree(tmp->private);
  kslab_cache_free(hash_item_cache,tmp);

  DEBUG_PRINT_VERBOSE("hash item deleted\n");
  return 0;
}

struct hash_item *
hash_find(struct hash *h, unsigned int key, void *cmpd)
{
  struct hash_item *tmp = NULL;
  unsigned int n;
  unsigned int found = !1;

  ASSERT_RETURN_VAL(h && cmpd && h->cmp, NULL);
  
  if (h->table[key])
    { 
      list_foreach(h->table[key],tmp,n)
	{
	  found = h->cmp(cmpd,tmp->private);
	  if (found)
	    break;
	}
    }
  
  if (!found)
    return NULL;
  
  return tmp;
}

int
hash_visit(struct hash *h,
	   visit_fn vfn)
{
  struct hash_item *tmp;
  unsigned int k, n;
  
  ASSERT_RETURN_VAL(h && vfn, -1);
  
  for (k = 0; k < h->prime ; k++)
    {
      if (h->table[k])
	list_foreach((struct hash_item *)h->table[k],tmp,n)
	{
	  if (vfn(tmp) != 0)
	    return -1;
	}
    }
  return 0;
}
	       

EXPORT_FUNCTION(hash_std_str_to_int);
EXPORT_FUNCTION(hash_create_table);
EXPORT_FUNCTION(hash_delete_table);
EXPORT_FUNCTION(hash_set_cmp_fn);
EXPORT_FUNCTION(hash_insert);
EXPORT_FUNCTION(hash_remove);
EXPORT_FUNCTION(hash_visit);
EXPORT_FUNCTION(hash_find);
