#include <fs/filesystem.h>
#include <part/_part.h>
#include <liblist/liblist.h>
#include <babel/tower/tower.h>
#include <kmem/kmem.h>
#include <lib/string.h>

#include <debug/debug.h>

static int part_write_res(struct resource *res, char *buf, offset_t offset, size_t len);
static int part_read_res(struct resource *res, char *buf, offset_t offset, size_t len);
static int part_mount_res(struct resource *res, char *mount_point, char *fstype, int flags);
static int part_destructor(struct translator *trans, void *rc);
static int part_constructor(struct translator *trans, void *rc);
static int part_mount(struct shadow_resource *sr, char *mount_point, char *fstype, int flags);
static int part_read(struct shadow_resource *sr, char *buf, 
		     offset_t offset, size_t len);
static int part_write(struct shadow_resource *sr, char *buf, 
		     offset_t offset, size_t len);

static struct hd_part *hd_part_list;

struct part_mt part_methods_table = {
  .mount= part_mount_res,
  .read=  part_read_res,
  .write= part_write_res
};

struct part_kernel_methods_table part_kmt = {
  .parent= {
    .read=        part_read,
    .write=       part_write,
    .init_sr=     NULL,
    .init_res=    NULL,
    .cleanup_sr=  NULL,
    .cleanup_res= NULL },
  .mount=       part_mount
};

struct interface part_interface =  {
  .name= "part",
  .family= "part",
  
  .description= "partitions management tool",
  .authors= "thomas petazzoni",
  .depends= NULL,
  
  .version= { 0 , 1 , 0 },
  
  .prev= NULL, 
  .next= NULL,
  .trans_list= NULL,
  
  .max_trans_instance= 256,
  .trans_size=         sizeof(struct part_translator),
  .res_size=           0,
  .sr_size=            0,
  .trans_cnt=          0,
  .trans_greatest_id=  0,

  .constructor=   part_constructor,
  .destructor=   part_destructor,
  
  .kernel_methods=  (struct kernel_methods_table *)&part_kmt,
  
  .methods_table= (struct interface_mt *)&part_methods_table,
  NULL
};


static int part_read(struct shadow_resource *sr, char *buf, 
		     offset_t offset, size_t len)
{
  CONCEPTION_ASSERT(sr && buf && len);

#define part_trans ((struct part_translator *) sr->translator)

  return 
  part_trans->hd->translator->interface->kernel_methods->read(part_trans->hd, 
							     buf, 
							     part_trans->start 
							     + offset, 
							     ALIGN_SUP(len, 
								       part_trans->blk_size));
}

static int part_write(struct shadow_resource *sr, char *buf, 
		      offset_t offset, size_t len)
{
  CONCEPTION_ASSERT(sr && buf && len);

  return 
  part_trans->hd->translator->interface->kernel_methods->write(part_trans->hd, 
							      buf, 
							      part_trans->start 
							      + offset, 
							      ALIGN_SUP(len, 
									part_trans->blk_size));
}

static int part_constructor(struct translator *trans, void *rc)
{
  UNUSED(trans);
  UNUSED(rc);
  
  return 0;
}

static int part_destructor(struct translator *trans, void *rc)
{
  UNUSED(trans);
  UNUSED(rc);
  
  return 0;
}

static bool_t check_fs(struct interface *interface, void *rc)
{
  struct shadow_resource *part = (struct shadow_resource *)rc;
  CONCEPTION_ASSERT(part);

  return ((struct fs_methods_table *)(interface->methods_table))->check_capability(part);
}

static int part_mount_res(struct resource *res, char *mount_point, char *fstype, int flags)
{
  return part_mount(res->shadow_resource, mount_point, fstype, flags);
}

static int part_mount(struct shadow_resource *sr, char *mount_point, char *fstype, int flags)
{
  struct interface *interface;
  struct translator *trans;
  struct fs_constructor_parameters f_params;

  ASSERT_RETURN_VAL(sr && mount_point, -1);
  
  interface = bbl_interface_visit("fs",fstype,check_fs,(void *)sr);
  if (!interface)
    return -1;

  f_params.part  = sr;
  f_params.flags = flags;

  trans = bbl_add_translator(interface, NULL, &f_params);
  if (!trans)
    return -1;
  
  if (flags & FS_REMOUNT)
    {
      if (bbl_remount(trans, mount_point) != 0)
	{
	  bbl_del_translator(trans, NULL);
	  return -1;
	}
    }
  else
    {
      if (bbl_mount(trans, mount_point) != 0)
	{
	  bbl_del_translator(trans, NULL);
	  return -1;
	}
    }

  return 0;
}

static int part_read_res(struct resource *res, char *buf, offset_t offset, size_t len)
{
  return part_read(res->shadow_resource, buf, offset, len);
}

static int part_write_res(struct resource *res, char *buf, offset_t offset, size_t len)
{
  return part_write(res->shadow_resource, buf, offset, len);
}


int __init_part_babel(void)
{
  DEBUG_PRINT("registering interface babel...");
  if (bbl_add_interface(&part_interface) != 0)
    {
      DEBUG_PRINT("Failed)\n");
      return -1;
    }
  DEBUG_PRINT("Ok\n");
  return 0;
}


int __init_part(void)
{
  struct hd_part *tmp;
  int n,i;

  struct shadow_resource *sr;
  struct part_translator *trans;

  char name[5];
  int extstart = 0;  
  int extsup = 0;
  int partnum = 1;  

  struct partition_entry *pe;
  char buf[512];
  
  pe = (struct partition_entry *) (buf+446);

  list_foreach(hd_part_list,tmp,n)
    {
      sr = bbl_open_sr_by_trans(tmp->hd);
      
      if (!sr)
	{
	  DEBUG_PRINT("open sr by trans returned a null sr\n");
	  return -1;
	}

      if (sr->translator->interface->kernel_methods->read(sr,buf,0,1) != 0)
	{
	  DEBUG_PRINT("sr->trans (read) failed !\n");
	  return -1;
	}
      
      /* Manage primary partitions */
      for (i = 0; i < 4; i++)
	{
	  if (pe[i].size == 0)
	    continue;

	  snprintf(name,sizeof(name),"%s%d",tmp->hd->name,i);
	  trans = (struct part_translator *)bbl_add_translator(&part_interface,
							       name,NULL);
	  if (!trans)
	    {
	      DEBUG_PRINT("adding trans %s failed for prim part\n",name); 
	      return -1;
	    }

	  trans->is_active = (pe[i].active == 0) ? 0 : 1;
	  trans->fs_type   = pe[i].type;
	  trans->start     = pe[i].lba;
	  trans->size      = pe[i].size;
	  trans->number    = partnum++;
	  trans->hd        = sr;
	  trans->blk_size  = tmp->blk_size;
	  __dbg_printk("[part] registering %s sz=%d\n", name, trans->size);


	  if(pe[i].type == 0x5)
	    extstart = pe[i].lba;
	}

      /* Manage extended partitions */
      while (extstart != 0)
	{
	  memset(buf, 0, 512);
	  
	  if (sr->translator->interface->kernel_methods->read(sr,buf,extstart+extsup,1) 
	      != 0)
	    {
	      DEBUG_PRINT("kernelmethods read failed for extended part\n");
	      return -1;
	    }

	  snprintf(name,sizeof(name),"%s%d",tmp->hd->name,partnum);
	  trans = (struct part_translator *)bbl_add_translator(&part_interface,
							       name,NULL);
	  
	  if (!trans)
	    {
	      DEBUG_PRINT("adding trans %s failed for ext part\n",name);
	      return -1;
	    }
	  
	  trans->is_active = (pe[0].active == 0) ? 0 : 1;
	  trans->fs_type   = pe[0].type;
	  trans->start     = extstart + extsup + pe[0].lba;
	  trans->size      = pe[0].size;
	  trans->number    = partnum++;
	  trans->blk_size  = tmp->blk_size;
	  
	  extsup = pe[1].lba;
	  if(extsup == 0)
	    break;
	}
      
      /* We don't close the shadow resource used to read hd, because
         read/write functions of partition will need this shadow
         resource ! */
    }

  return 0;
}



int _part_request_for_registering(struct translator *trans, size_t blk_size)
{
  struct hd_part *hdp;
  
  hdp = kmalloc(sizeof(struct hd_part));
  if (!hdp)
    return -1;

  hdp->hd       = trans;
  hdp->blk_size = blk_size;
  list_add_head(hd_part_list,hdp);
  
  return 0;
}


