
#include <kos/macros.h>
#include <debug/debug.h>
#include <arch/mm/mm.h>
#include <lib/string.h>

#include "_pmm.h"

/* Tiny GPFM for the pages [0-12k] in RAM, in order to allow the
   allocation of the first GPFM page (and eventually its PT). Goal:
   avoid the eager & egg problem (in order to get the physical pages
   for the gpfm, we need to have a valid GPFM !) */
__init_data static gpfme_t sieve_gpfm[3];
/*
 * [0] = 0k: NULL page (=> to generate page faults)
 * [1] = 4k: First PT for first GPFM page (if needed) /OR/ first GPFM page
 * [2] = 8k: first GPFM page /OR/ still free
 */


/* Initial physical memory map:
 *
 *   0 ---------------------------------------------------------
 *   |                                               (discarded)
 *   4k --------------------------------------------------------
 *   |                                                      free
 *   0xA0000 ---------------------------------------------------
 *   |                                  hw_mapping (reclaimable)
 *   0x100000 --------------------------------------------------
 *   |                                                      free
 *   allocated_memory_base_phys_addr ---------------------------
 *   |                                             non_swappable
 *   loader_allocatable_mem_size -------------------------------
 *   |                                                      free
 *   total_mem_size --------------------------------------------
 */
#define BIOS_N_VIDEO_START 0xA0000
#define BIOS_N_VIDEO_END   0x100000

/************************* BUG loader **************************/
__init_text static int init_ram_gpfme(kernel_parameter_t *kp, int page_index)
{
  paddr_t paddr = page_index << PAGE_SIZE_SHIFT;

  INIT_GPFME(gpfm.ram_map + page_index);
  gpfm.ram_map[page_index].address = paddr;

#define BETWEEN(lower,x,upper) (((x) >= (lower)) && ((x) < (upper)))
  if (page_index == 0)
    {
      gpfm.ram_map[page_index].ref_cnt           = ~0;
      gpfm.ram_map[page_index].flags.page_type   = PHYS_PAGE_KERNEL;
      gpfm.ram_map[page_index].flags.swap_status = PHYS_PAGE_NON_SWAPPABLE;
      /* Definitely lost */
    }
  else if (BETWEEN(PAGE_SIZE, paddr, BIOS_N_VIDEO_START))
    {
      gpfm.ram_map[page_index].flags.page_type   = PHYS_PAGE_FREE;
      gpfm.ram_map[page_index].flags.swap_status = PHYS_PAGE_NON_SWAPPABLE;
      GPFM_LIST_ADD(free, gpfm.ram_map + page_index);
    }
  else if (BETWEEN(BIOS_N_VIDEO_START, paddr, BIOS_N_VIDEO_END))
    {
      gpfm.ram_map[page_index].flags.page_type   = PHYS_PAGE_HW_MAPPING;
      gpfm.ram_map[page_index].flags.hw_mapping_reclaiming_status
	= PHYS_PAGE_HW_MAPPING_RECLAIMABLE;
      GPFM_LIST_ADD(hw_mapping, gpfm.ram_map + page_index);
    }
  else if (BETWEEN(BIOS_N_VIDEO_END, paddr, kp->allocated_memory_base_phys_addr))
    {
      gpfm.ram_map[page_index].flags.page_type   = PHYS_PAGE_FREE;
      gpfm.ram_map[page_index].flags.swap_status = PHYS_PAGE_NON_SWAPPABLE;
      GPFM_LIST_ADD(free, gpfm.ram_map + page_index);
    }
  else if (BETWEEN(kp->allocated_memory_base_phys_addr,
		   paddr,
		   kp->loader_allocatable_mem_size))
    {
      gpfm.ram_map[page_index].ref_cnt           = 1;
      gpfm.ram_map[page_index].flags.page_type   = PHYS_PAGE_KERNEL;
      gpfm.ram_map[page_index].flags.swap_status = PHYS_PAGE_NON_SWAPPABLE;
      GPFM_LIST_ADD(non_swappable, gpfm.ram_map + page_index);
    }
  else if (BETWEEN(kp->loader_allocatable_mem_size,
		   paddr,
		   kp->total_mem_size))
    {
      gpfm.ram_map[page_index].flags.page_type   = PHYS_PAGE_FREE;
      gpfm.ram_map[page_index].flags.swap_status = PHYS_PAGE_NON_SWAPPABLE;
      GPFM_LIST_ADD(free, gpfm.ram_map + page_index);
    }
  else
    FAILED_VERBOSE("Invalid paddr !\n");

  return 0;
}


__init_text int init_gpfm(kernel_parameter_t *kp)
{
  unsigned int i;
  paddr_t      ppage;
  vaddr_t      gpfme_page;

  main_memory_size = kp->total_mem_size;

  /* Evaluate size of GPFM to allocate: */
  gpfm.ram_map_size = PAGE_ALIGN_SUP((kp->total_mem_size >> PAGE_SIZE_SHIFT) * sizeof(gpfme_t));
  if (CORE_KERNEL_VIRTUAL_ADDR - PAGE_SIZE < gpfm.ram_map_size)
    FAILED_VERBOSE("Could not find enough virtual space to allocate GPFM\n");
  /* Note: with CORE_KERNEL_VADDR=768M and RAM=64G, max gpfme size =
     47.9998 Bytes */

  /* Init gpfm lists */
  gpfm.free          = NULL;
  gpfm.swappable     = NULL;
  gpfm.non_swappable = NULL;
  gpfm.hw_mapping    = NULL;
  spinlock_init(gpfm.lock);

  /* Init sieve GPFM: allows to allow the first PT and 1st GPFM page */
  gpfm.ram_map = sieve_gpfm;
  init_ram_gpfme(kp, 0);
  init_ram_gpfme(kp, 1);
  init_ram_gpfme(kp, 2);

  /* Allocate first GPFM page */
  gpfme_page = CORE_KERNEL_VIRTUAL_ADDR - gpfm.ram_map_size;
  ppage      = get_physical_page(PHYS_PAGE_KERNEL, PHYS_PAGE_NON_SWAPPABLE);
  ASSERT_FATAL(ppage != 0);
  ASSERT_FATAL(arch_map_virtual_page(NULL,
				     gpfme_page, ppage,
				     VM_ACCESS_WRITE | VM_ACCESS_READ)
	       == 0);
  memset((void*)gpfme_page, 0x0, PAGE_SIZE);

  /* Ok, got our first page => let's init its first 3 entries */
  gpfm.free          = NULL;
  gpfm.swappable     = NULL;
  gpfm.non_swappable = NULL;
  gpfm.hw_mapping    = NULL;
  gpfm.ram_map       = (gpfme_t*)gpfme_page;
  init_ram_gpfme(kp, 0);
  for (i = 1 ; i < 3 ; i++)
    {
      if (sieve_gpfm[i].flags.page_type != PHYS_PAGE_FREE)
	{
	  INIT_GPFME(gpfm.ram_map + i);
	  gpfm.ram_map[i].address = i << PAGE_SIZE_SHIFT;
	  gpfm.ram_map[i].flags.page_type   = PHYS_PAGE_KERNEL;
	  gpfm.ram_map[i].flags.swap_status = PHYS_PAGE_NON_SWAPPABLE;
	  GPFM_LIST_ADD(non_swappable, gpfm.ram_map + i);
	}
      else
	init_ram_gpfme(kp, i);
    }

  /* Allocate and init the other GPFME */
  for (i = 3 ; i < (kp->total_mem_size >> PAGE_SIZE_SHIFT) ; i++)
    {
      vaddr_t cur_gpfme_end_page
	= PAGE_ALIGN_INF(((vaddr_t)(gpfm.ram_map + (i + 1)))-1);

      /* Considering a new page for the GPFM ? */
      if (cur_gpfme_end_page != gpfme_page)
	{
	  /* Map another page for the GPFM */
	  ppage = get_physical_page(PHYS_PAGE_KERNEL, PHYS_PAGE_NON_SWAPPABLE);
	  ASSERT_FATAL(ppage != 0);
	  ASSERT_FATAL(arch_map_virtual_page(NULL, cur_gpfme_end_page,
					     ppage, VM_ACCESS_WRITE)
		       == 0);
	  memset((void*)cur_gpfme_end_page, 0x0, PAGE_SIZE);
	  gpfme_page = cur_gpfme_end_page;
	}

      /* Actually init the GPFME */
      init_ram_gpfme(kp, i);
    }

  return 0;  
}
