#ifndef __PMM_H__
#define __PMM_H__

#include <kos/system.h>
#include <arch/mm/mm.h>

struct kslab_slab;

// GPFM (Global Page Frame Map) Entry = GPFME
// One GPFME correspond to one physical page
typedef struct gpfme gpfme_t;
struct gpfme 
{
  paddr_t address;              // 4 bytes (address of the page)

  struct gpfme_flags_s {
#define PHYS_PAGE_FREE       1
#define PHYS_PAGE_KERNEL     2
#define PHYS_PAGE_USER       3
#define PHYS_PAGE_HW_MAPPING 4
    k_ui32_t page_type :3;

    /* For PHYS_PAGE_{KERNEL,USER} pages ONLY */
#define PHYS_PAGE_NON_SWAPPABLE 0
#define PHYS_PAGE_SWAPPABLE     1
    k_ui32_t swap_status :1;
    
    /* For PHYS_PAGE_HW_MAPPING pages ONLY */
#define PHYS_PAGE_HW_MAPPING_NON_RECLAIMABLE 0
#define PHYS_PAGE_HW_MAPPING_RECLAIMABLE     1
    /** Reclaimable: upon cancellation of the hw_mapping, the page is
     * inserted into the free list for further get_physical_page().\ 
     * Unreclaimable: upon hw_mapping cancellation, the gpfme is
     * destroyed.
     */
    k_ui32_t hw_mapping_reclaiming_status :1;

  } flags; // 4 bytes

  union {
    /* Pointers used for linking in the free pages list */
    struct {
      gpfme_t *next;               // 4 bytes
      gpfme_t *prev;               // 4 bytes      
    } free;

    /* Pointers used for linking in the swappable pages lists */
    struct {
      gpfme_t *next;
      gpfme_t *prev;
    } swappable;

    /* Pointers used for linking in the non swappable pages lists */
    struct {
      gpfme_t *next;
      gpfme_t *prev;
    } non_swappable;

    /* Pointers used for linking in the hardware mapping pages list */
    struct {
      gpfme_t *next;               // 4 bytes
      gpfme_t *prev;               // 4 bytes
    } hw_mapping;
  } u; // 8 bytes

  // Bitmap for kernel memory allocation
  struct kslab_slab *slab; //4B

  struct mapping_s  *mapping_list;
  k_ui32_t ref_cnt; //4B -- Number of virtual mappings
};

/** Get the gpfme at a physical address.
 * @return gpfme or NULL (+locked)
 * @note Always call gpfme_unlock, even if gpfme is NULL !!!
 */
gpfme_t *get_gpfme_at_phys_addr(paddr_t phys, k_ui32_t* gpfm_lock_flags);

/** Get the gpfme at a virtual address (in current address space).
 * @return gpfme or NULL (+locked)
 * @note Always call gpfme_unlock, even if gpfme is NULL !!!
 */
gpfme_t *get_gpfme_at_virt_addr(vaddr_t virt, k_ui32_t* gpfm_lock_flags);

/**
 * Unlock the gpfm
 * @param gpfme may be NULL...
 * @param gpfm_lock_flags never NULL !
 * @return 0 (always)
 */
int gpfme_unlock(gpfme_t* gpfme, k_ui32_t* gpfm_lock_flags);

/** For any User or Kernel page, change its swap status.
 * @note assumes gpfm is locked. The gpfm is kept locked.
 * @return -1 on error, old swap status when Ok
 */
int change_gpfme_swap_status(/*in*/gpfme_t *gpfme,
			     int new_swap_status);

/** Get a new free physical page.
 * @param page_type Either PHYS_PAGE_KERNEL, PHYS_PAGE_USER or
 *                  PHYS_PAGE_HW_MAPPING
 * @param swap_status Either PHYS_PAGE_SWAPPABLE or PHYS_PAGE_NON_SWAPPABLE
 * @return The physical address of the new page, or 0 on error.
 * @note BEWARE: ref_cnt of the new page is set to 0 !
 * @note SAFE (lock Ok).
 * @see put_physical_page()
 */
paddr_t get_physical_page(int page_type, int swap_status);

/** Release a previously allocated or hardware mapped page. If the
 * page is a hardware mapping and if it has status
 * PHYS_PAGE_HW_MAPPING_RECLAIMABLE, then the corresponding page is
 * inserted into the free list for later get_physical_page().
 *
 * @param paddr the physical address of the page to release.
 * @return 0 when Ok. -1 when page not physically present and used.
 * @see get_physical_page() and @see get_hw_mapping_page()
 * @note SAFE (lock Ok).
 */
int put_physical_page(paddr_t paddr);

/** Add a new gpfme to the free pages list.
 * @param paddr Is checked agains existing physical addresses already
 * mapped.
 * @see remove_additional_allocatable_page()
 * @note SAFE (lock Ok).
 * @return 0 when Ok. -1 when already mapped/kmalloc(new_gpfme) failed.
 */
int new_additional_allocatable_page(paddr_t paddr);

/** Remove a page previously marked as allocateble with
 * new_additional_allocatable_page().
 * @see new_additional_allocatable_page()
 * @note SAFE (lock Ok).
 * @return -1 on error (page still used, or not an 'additional memory'
 *         area, or not mapped)
 */
int remove_additional_allocatable_page(paddr_t paddr);

/** Map an hardware device at page paddr.
 * @param paddr if corresponding to a free allocatable page, then the
 *              hw_mapping is initialized with status
 *              PHYS_PAGE_HW_MAPPING_RECLAIMABLE. Otherwise, a new
 *              gpfme is allocated, and hw_mapping is initialized
 *              with status PHYS_PAGE_HW_MAPPING_NON_RECLAIMABLE.
 * @return The new gpfme, or NULL if paddr overlaps an already
 *         allocated page.
 * @see put_physical_page()
 * @see gpfme_t::flags.u.hw_mapping_reclaiming_status
 * @note SAFE (lock Ok).
 */
int declare_hw_mapping_page(paddr_t paddr);


/**
 * For kvmem_init only, in order to declare the kernel range for the GPFM
 */
size_t _get_gpfm_ram_map_size(void);

/**
 * Callback called on each gpfme. Should return 0 if want to see next
 * gpfme, or != if must stop.
 * @note locks held
 * @see visit_gpfm_list()
 */
typedef int (*pmm_list_visitor_t)(gpfme_t* gpfme, void* custom_param);
/**
 * Visit the lists according to page type
 * @param page_type either PHYS_PAGE_FREE, PHYS_PAGE_KERNEL, PHYS_PAGE_USER, PHYS_PAGE_HW_MAPPING
 * @param custom_param Passed to the visitor
 * @note DOES NOT acquire ANY lock !
 * @return 0 (even when error), or the return value of the first visitor which returns != 0.
 * @see pmm_list_visitor_t
 */
int _gpfm_visit_list_unsafe(int page_type, pmm_list_visitor_t visitor, void* custom_param);

#endif
