#include <scheduler/_scheduler.h>
#include <kos/macros.h>
#include <task/task.h>
#include <arch/mm/mm.h>
#include <lib/stdio.h>
#include <kos/spinlock.h>
#include <idt/irq.h>

struct cpu_state *current_cpu_state[NB_CPU] = { NULL, };

#define _ELLIGIBLE(thr) ((thr->state == READY) || (thr->state == INTERRUPTED))

thread_t *reschedule_unsafe (thread_t *current_thread)
{
  thread_t *next_thread;

  if (!current_thread)
    FAILED_VERBOSE("(scheduler) No current thread\n");

  next_thread = current_thread->next_global;

  /* Finds next elligible thread */  
  while (TRUE) {
    if (!next_thread)
      next_thread = get_global_thread_list_head();
    if (_ELLIGIBLE(next_thread))
      break;
    next_thread = next_thread->next_global;
  }

  /* Update the system TSS, because a User thread has been
     elected. This thread will require kernel stack for its next
     interrupts, and he will find the address of the stack in the
     system TSS */
  if(next_thread->type == USER_THREAD)
    update_system_tss(((unsigned)next_thread->cpu_context.cpl0) +
		      sizeof(cpu_state_t));
  
  next_thread->state = RUNNING;

  return next_thread;
}

/* Contrary to first impression, the thread lock is released
   immediatly when switching to the next_thread, precisely by this
   next_thread */
void reschedule_run_next(void)
{
  thread_t *next_thread;
  thread_t *current_thread;
  k_ui32_t flags;

  ASSERT_FATAL(get_hw_isr_nested_level() == 0);
  
  write_spin_lock(kernel_thread_list_lock, flags);
  current_thread = get_current_thread();

  next_thread = reschedule_unsafe(current_thread);

  set_current_thread(next_thread);
  
  write_spin_unlock_remote(kernel_thread_list_lock, flags);
  
  cpl0_switch_with_return(&(current_thread->cpu_context.cpl0),
			  next_thread->cpu_context.cpl0);

  write_spin_unlock_local(kernel_thread_list_lock, flags);
}

struct cpu_state *
reschedule_after_interrupt (struct cpu_state *cpu_state)
{
  struct cpu_state *next_cpu_state;
  thread_t *next_thread;
  thread_t *current_thread;
  k_ui32_t flags;
  
  exclusive_spin_lock(kernel_thread_list_lock, flags);
  current_thread = get_current_thread();

  if (current_thread->state != RUNNING){
    char * state_str;
    switch (current_thread->state) {
#define CASE_STR(val) \
      case val: state_str=# val; break
      CASE_STR(READY);
      CASE_STR(RUNNING);
      CASE_STR(INTERRUPTED);
      CASE_STR(STOPPED);
      CASE_STR(DELETED);
      CASE_STR(SLEEPING);
      CASE_STR(WAITING);
    default:
      state_str = "!!!unknown!!!";
    }
    FAILED_VERBOSE("Current thread is in the %s state : NOT RUNNING !\n",
		   state_str);
  }

  /* Interrupting a RUNNING thread. */
  current_thread->state = INTERRUPTED;
  current_thread->cpu_context.cpl0 = (cpu_state_t*) cpu_state;
  
  /* FIXME FOR SMP */
  next_thread = reschedule_unsafe(current_thread);
  next_cpu_state 
    = current_cpu_state[current_processor()] 
    = (struct cpu_state*) next_thread->cpu_context.cpl0;

  set_current_thread(next_thread);

  exclusive_spin_unlock(kernel_thread_list_lock, flags);

  return next_cpu_state;
}

