/*
 * Copyright (C) 2000, Thomas PETAZZONI
 * http://kos.enix.org
 *
 * Timer management
 *
 * @(#) $Id: _sleep.c,v 1.1.1.1 2003/04/27 12:49:49 thomas Exp $
 */

#include <task/task.h>
#include <kos/types.h>
#include <kos/system.h>
#include <kos/spinlock.h>
#include <arch/task/task.h>
#include <scheduler/_scheduler.h>
#include <arch/lib/delay.h>

thread_t *sleeping_thread_list = NULL;
SPINLOCK(sleeping_thread_list_lock);

/* No need to spinlock because it has already done by the caller */
static void add_thread_to_sleeping_list(thread_t *thread)
{
  if(!sleeping_thread_list) 
    {
      sleeping_thread_list = thread;
      sleeping_thread_list->u.sleeping_thread.prev = NULL;
      sleeping_thread_list->u.sleeping_thread.next = NULL;
    }
  else 
    {
      thread->u.sleeping_thread.next = sleeping_thread_list;
      thread->u.sleeping_thread.prev = NULL;
      sleeping_thread_list->u.sleeping_thread.prev = thread;
      sleeping_thread_list = thread;
    }
}

/* No need to spinlock because it has already done by the caller */
static void remove_thread_from_sleeping_list(thread_t *thread)
{
  if(thread->u.sleeping_thread.prev)
    thread->u.sleeping_thread.prev->u.sleeping_thread.next = 
      thread->u.sleeping_thread.next;
  else
    sleeping_thread_list = thread->u.sleeping_thread.next;

  if(thread->u.sleeping_thread.next)
    thread->u.sleeping_thread.next->u.sleeping_thread.prev = 
      thread->u.sleeping_thread.prev;
}


void check_timers(jiffy_t current_jiffies)
{
  thread_t *current_thread;
  k_ui32_t flags;

  for(current_thread = sleeping_thread_list;
      current_thread != NULL;
      current_thread = current_thread->u.sleeping_thread.next)
    {
      if(current_thread->u.sleeping_thread.expiration_time < current_jiffies) {
	write_spin_lock(sleeping_thread_list_lock, flags);
	remove_thread_from_sleeping_list(current_thread);
	current_thread->state = INTERRUPTED;
	write_spin_unlock(sleeping_thread_list_lock, flags);
      }
    }

}


void usleep(udelay_t delay)
{
  thread_t *current_thread = get_current_thread();
  k_ui32_t flags_sched;
  k_ui32_t flags_sleep;

  if(delay <= 0)
    return;

  write_spin_lock(kernel_thread_list_lock, flags_sched);

  current_thread->state       = SLEEPING;
  current_thread->u.sleeping_thread.expiration_time = 
    jiffies + ((delay*HZ)/1000000);


  write_spin_lock(sleeping_thread_list_lock, flags_sleep);
  add_thread_to_sleeping_list(current_thread);
  write_spin_unlock(sleeping_thread_list_lock, flags_sleep);

  write_spin_unlock_remote(kernel_thread_list_lock, flags_sched);

  reschedule_run_next();

  write_spin_unlock_local(kernel_thread_list_lock, flags_sched);
}

static unsigned long loops_per_sec = 1 << 12;

void udelay(unsigned long usecs)
{
  usecs = (usecs * loops_per_sec) / 1000000;
  arch_loop_delay(usecs);
}

void init_delay_loop(void)
{
  unsigned long ticks, loopbit;
  unsigned long lps_precision = 8;


  while(loops_per_sec <<= 1)
    {
      ticks = jiffies;

      while(ticks == jiffies)
	;

      ticks = jiffies;
      arch_loop_delay(loops_per_sec);
      ticks = jiffies - ticks;
      if(ticks)
	break;
    }
  loops_per_sec >>= 1;
  loopbit = loops_per_sec;
  
  while(lps_precision-- && (loopbit >>=1))
    {
      loops_per_sec |= loopbit;
      ticks = jiffies;

      while(ticks == jiffies)
	;

      ticks = jiffies;
      arch_loop_delay(loops_per_sec);
      if(jiffies != ticks)
	loops_per_sec &= ~loopbit;
    }
  
  loops_per_sec *= HZ;
    
  printk("%lu.%02lu BogoMIPS",
	 (loops_per_sec+2500)/500000,
	 ((loops_per_sec+2500)/5000) % 100);
}
