/*
 * Copyright (C) 2000, Thomas PETAZZONI
 * http://kos.enix.org
 *
 * Functions needed to manage kernel stacks.
 *
 * @(#) $Id: _task_kstack.c,v 1.1.1.1 2003/04/27 12:49:49 thomas Exp $
 */


#include <task/_task.h>
#include <kos/system.h>
#include <kos/assert.h>
#include <arch/lib/lib.h>
#include <vmm/vmm.h>
#include <pmm/pmm.h>
#include <kmem/kmem.h>
#include <lib/string.h>

struct bitmap {
  addr_t data;
  size_t size;
};

/* Private variables */

struct bitmap free_cpl0_kernel_stack_bitmap;
struct bitmap free_cpl0_user_stack_bitmap;

void unallocate_cpl0_stack (thread_t *thread)
{
  range_t stack;  
  
  CONCEPTION_ASSERT(thread->mother_team);

  switch (thread->type) {
    size_t stack_index;
    
  case KERNEL_THREAD:
    stack_index = 
      (thread->cpl0_stack_base_addr - CPL0_KERNEL_STACK_VIRTUAL_ADDR) 
      / CPL0_KERNEL_STACK_VIRTUAL_SIZE;
    
    set_bit_in_bitmap ((addr_t) free_cpl0_kernel_stack_bitmap.data,
		       stack_index, 1);		       

    stack.start = thread->cpl0_stack_base_addr;
    stack.end   = stack.start + CPL0_KERNEL_STACK_VIRTUAL_SIZE;
    unmap_virtual_range(thread->mother_team, &stack);

    break;
    
  case USER_THREAD:
    stack_index = 
      (thread->cpl0_stack_base_addr - CPL0_USER_STACK_VIRTUAL_ADDR) 
      / CPL0_USER_STACK_VIRTUAL_SIZE;
    set_bit_in_bitmap ((addr_t) free_cpl0_user_stack_bitmap.data,
		       stack_index, 1);		       

    stack.start = thread->cpl0_stack_base_addr;
    stack.end   = stack.start + CPL0_USER_STACK_VIRTUAL_SIZE;
    unmap_virtual_range(thread->mother_team, &stack);

    break;
   
  default:
    FAILED_VERBOSE("Unknown thread type\n");
  }
  
}


int allocate_cpl0_stack (thread_t *thread)
{
  addr_t ppage;
  vaddr_t vpage;

  CONCEPTION_ASSERT(thread->mother_team);

  switch (thread->type) {
    size_t stack_index;
    
  case KERNEL_THREAD:
    stack_index = bsf_bitmap ((addr_t) free_cpl0_kernel_stack_bitmap.data,
			      free_cpl0_kernel_stack_bitmap.size);
    if (stack_index > free_cpl0_kernel_stack_bitmap.size * 8)
      return -1;
    thread->cpl0_stack_base_addr = CPL0_KERNEL_STACK_VIRTUAL_ADDR 
      + stack_index * CPL0_KERNEL_STACK_VIRTUAL_SIZE;
    
    set_bit_in_bitmap ((addr_t) free_cpl0_kernel_stack_bitmap.data,
		       stack_index, 0);		       
    for(vpage = (thread->cpl0_stack_base_addr +
		 CPL0_KERNEL_STACK_VIRTUAL_SIZE -
		 CPL0_KERNEL_STACK_PHYSICAL_SIZE);
	vpage < (thread->cpl0_stack_base_addr +
		 CPL0_KERNEL_STACK_VIRTUAL_SIZE);
	vpage += PAGE_SIZE)
      {
	ppage = get_physical_page(PHYS_PAGE_KERNEL, PHYS_PAGE_NON_SWAPPABLE);
	if(!ppage)
	  FAILED_VERBOSE("[allocate_cpl0_stack] No more physical memory\n");

	map_virtual_page(thread->mother_team, vpage, ppage, 
			 VM_ACCESS_READ | VM_ACCESS_WRITE);
      }
    

    break;
      
  case USER_THREAD:
    stack_index = bsf_bitmap ((addr_t) free_cpl0_user_stack_bitmap.data,
			      free_cpl0_user_stack_bitmap.size);
    if (stack_index > free_cpl0_user_stack_bitmap.size * 8)
      return -1;
    thread->cpl0_stack_base_addr = CPL0_USER_STACK_VIRTUAL_ADDR 
      + stack_index * CPL0_USER_STACK_VIRTUAL_SIZE;
    set_bit_in_bitmap ((addr_t) free_cpl0_user_stack_bitmap.data,
		       stack_index, 0);		       

    for(vpage = (thread->cpl0_stack_base_addr +
		 CPL0_USER_STACK_VIRTUAL_SIZE -
		 CPL0_USER_STACK_PHYSICAL_SIZE);
	vpage < (thread->cpl0_stack_base_addr +
		 CPL0_USER_STACK_VIRTUAL_SIZE);
	vpage += PAGE_SIZE)
      {
	ppage = get_physical_page(PHYS_PAGE_KERNEL, PHYS_PAGE_NON_SWAPPABLE);
	if(!ppage)
	  FAILED_VERBOSE("[allocate_cpl0_stack] No more physical memory\n");

	map_virtual_page(thread->mother_team, vpage, ppage, 
			 VM_ACCESS_READ | VM_ACCESS_WRITE);
      }

    break;
    
  default:
    FAILED_VERBOSE("Unknown thread type\n");
  }

  return 0;
}

__init_text int init_cpl0_stacks_regions (void)
{
  /* Allocates place and initialize bitmaps */
  free_cpl0_kernel_stack_bitmap.size = (CPL0_KERNEL_STACK_VIRTUAL_REGION_SIZE 
					/ CPL0_KERNEL_STACK_VIRTUAL_SIZE) / 8;
  free_cpl0_user_stack_bitmap.size = (CPL0_USER_STACK_VIRTUAL_REGION_SIZE 
					/ CPL0_USER_STACK_VIRTUAL_SIZE) / 8;
  CONCEPTION_ASSERT(free_cpl0_user_stack_bitmap.size <= PAGE_SIZE);
  CONCEPTION_ASSERT(free_cpl0_kernel_stack_bitmap.size <= PAGE_SIZE);

  free_cpl0_kernel_stack_bitmap.data = 
    (addr_t) kmalloc(free_cpl0_kernel_stack_bitmap.size);
  CONCEPTION_ASSERT(free_cpl0_kernel_stack_bitmap.data);

  free_cpl0_user_stack_bitmap.data =
    (addr_t) kmalloc(free_cpl0_user_stack_bitmap.size);
  CONCEPTION_ASSERT(free_cpl0_user_stack_bitmap.data);

  memset((void *)free_cpl0_user_stack_bitmap.data, ~0, 
	 free_cpl0_user_stack_bitmap.size);
  memset((void *)free_cpl0_kernel_stack_bitmap.data, ~0, 
	 free_cpl0_kernel_stack_bitmap.size);

  return 0;
}

