/*
 * Copyright (C) 2000, Thomas PETAZZONI & David DECOTIGNY
 * http://kos.enix.org
 *
 * Functions needed to create and delete kernel threads.
 *
 * @(#) $Id: _task_kthread.c,v 1.1.1.1 2003/04/27 12:49:49 thomas Exp $
 */


#include <task/_task.h>
#include <kos/system.h>
#include <kos/assert.h>
#include <kmem/kmem.h>
#include <lib/string.h>
#include <kos/system.h>
#include <arch/mm/mm.h>
#include <kos/spinlock.h>

inline static void delete_kernel_thread_unsafe (thread_t *thread)
{
  if(thread == get_current_thread()) 
    {
      WARNING("Attempt to delete the current thread\n");
      return;
    }

  unregister_thread(thread);
  unallocate_cpl0_stack(thread);
  kfree_size_known(thread, sizeof(thread_t));
}

void thread_end_handler (void)
{
  cpu_state_t *next_cpu_state;
  thread_t *thread, *cur;
  k_ui32_t flags;

  write_spin_lock(kernel_thread_list_lock, flags);

  cur = current_thread;
  __dbg_printk(_B_RED "thr %s ended\n" _B_NORM, cur->name);

  cur->state = STOPPED;
  thread = reschedule_unsafe(cur);
  ASSERT_FATAL(cur != thread);

  /* Moving to a special private stack, in order to call the
   * delete_kernel_thread safely (because delete will destroy this
   * thread's stack) */
  asm volatile("movl %0, %%esp"::"r"(thread_end_handler_stack));

  next_cpu_state = thread->cpu_context.cpl0;  
  set_current_thread(thread);
  delete_kernel_thread_unsafe(cur);

  write_spin_unlock_remote(kernel_thread_list_lock, flags);
  cpl0_switch_no_return(next_cpu_state);
  
  /* No need to do do write_spin_unlock_local because we never get
     there (thread ended) */
}


thread_t *_create_kernel_thread(team_t *team,
				kernel_thread_fct_t *kt,
				char *name,
				void *data)
{
  thread_t *new_thread;
  k_ui32_t flags;

  if (team == NULL)
    team = get_current_team();

  new_thread = (thread_t *) kmalloc(sizeof(thread_t));

  if (!new_thread)
    {
      WARNING("No more memory to allocate thread\n");
      return NULL;
    }

  if (name)
    strncpy(new_thread->name, name, 32);
  else
    strncpy(new_thread->name, "<no name>", 32);

  new_thread->state = STOPPED;
  new_thread->type  = KERNEL_THREAD;

  /* Assign thread's team */
  new_thread->mother_team = team;

  write_spin_lock(kernel_thread_list_lock, flags);

  if (allocate_cpl0_stack (new_thread))
    {
      WARNING("No more CPL0 stacks !\n");
      write_spin_unlock(kernel_thread_list_lock, flags);
      return NULL;
    }

  register_thread(new_thread, team);
  init_kernel_thread_context (new_thread, kt, data);
  new_thread->state = READY;

  write_spin_unlock(kernel_thread_list_lock, flags);

  return new_thread;
}
