/*
 * Copyright (C) 2000, Thomas PETAZZONI
 * http://kos.enix.org
 *
 * Management of the array of sres, to create file descriptors.
 *
 * @(#) $Id: _task_sres.c,v 1.2 2003/04/27 20:55:55 thomas Exp $
 */

#include <kos/types.h>
#include <kos/assert.h>
#include <task/_task.h>
#include <karm/karm.h>

/* task_insert_sres
 *
 * Insert a sres somewhere in the sres_array of the team, and return
 * the file descriptor in the "fd" argument.
 *
 * sres : the sres to add
 * fd   : where to return the file descriptor (index in the array)
 *
 * Returns 0 on success, an error code on failure
 */
int task_insert_ures(struct ures *ures, /*out*/ int *fd)
{
  int i;
  team_t *team;
  thread_t *thread;

  ASSERT_FATAL(ures != NULL);
  ASSERT_FATAL(fd   != NULL);

  thread = get_current_thread();
  team   = thread->mother_team;

  /* What happens here if we get interrupted, and that during that
     time, the current team get destroyed ? 

     Maybe we should have a more global lock, that we could hold only
     to say that we are using a team. For example :

     take_global_lock();
     thread = get_current_thread();
     team = thread->mother_team;
     team->busy++;
     release_global_lock();

     take_semaphore_on_team();
     Do all the stuff
     release_semaphore_on_team();

     take_global_lock();
     team->busy--;
     release_global_lock();

     This should work, but isn't perfect at all : if the team
     destruction function sees that the team is busy (team->busy > 0),
     what does it do ? Return -EBUSY ? Wait for the team not to be
     busy ? But how ? Such a hack isn't really nice :

     while(team->busy > 0)
       {
         sleep(100);
       }
  */

  /* TODO : take a semaphore on team */

  /* Let's find the first unused slot */
  for (i = 0; i < MAX_FD_PER_TEAM; i++)
    {
      if(team->ures_array[i] == NULL)
	break;
    }

  /* If we didn't find a slot, exit */
  if(i >= MAX_FD_PER_TEAM) 
    {
      /* TODO : release semaphore on team */
      return -1; /* TODO : replace with error code, but which one ? */
    }
  
  team->ures_array[i] = ures;
  *fd = i;

  /* TODO : release semaphore on team */

  return 0;
}

/* task_replace_ures
 *
 * Replace the ures at index 'fd' in the array by the given ures
 *
 * fd   : the index to replace
 * ures : the ures that will be used to replace the old one
 *
 * Returns 0 on success, an error code on failure
 */
int task_replace_ures(int fd, struct ures *ures)
{
  team_t *team;
  thread_t *thread;

  if(fd < 0 || fd > MAX_FD_PER_TEAM)
    {
      return -1; /* TODO : return -EINVAL */
    }

  thread = get_current_thread();
  team   = thread->mother_team;

  /* Same remark as above */

  /* TODO : Take the semaphore on the team */

  if(team->ures_array[fd] == NULL)
    {
      return -1; /* TODO : return -ENOENT */
    }

  team->ures_array[fd] = ures;

  return 0; /* TODO : return -ENOENT */
}

/* task_delete_ures
 *
 * Delete the ures at the given index in the array.
 *
 * fd : the index in the array of the ures to delete.
 *
 * Returns 0 on success, an error code on failure
 */
int task_delete_ures(int fd)
{
  team_t *team;
  thread_t *thread;

  if(fd < 0 || fd > MAX_FD_PER_TEAM)
    {
      return -1; /* TODO : return -EINVAL */
    }

  thread = get_current_thread();
  team   = thread->mother_team;

  /* Same remark as above */

  /* TODO : Take the semaphore on the team */

  if(team->ures_array[fd] == NULL)
    {
      return -1; /* TODO : return -ENOENT */
    }

  team->ures_array[fd] = NULL;

  return 0; /* TODO : return -ENOENT */
}
