/*
 * Copyright (C) 2000, Thomas PETAZZONI
 * http://kos.enix.org
 *
 * Public declarations for task management (portable part)
 *
 * @(#) $Id: task.h,v 1.2 2003/04/27 20:55:55 thomas Exp $
 */


#ifndef __TASK_H__
#define __TASK_H__

#include <scheduler/scheduler.h>
#include <karm/ures.h>

#define MAX_KERNEL_THREAD_NUMBER  256
#define MAX_USER_THREAD_NUMBER   4096
#define MAX_TEAM_NUMBER          1024
#define MAX_FD_PER_TEAM           256

typedef struct team team_t;
typedef struct thread thread_t;

typedef void (kernel_thread_fct_t)(void *data);


#include <vmm/vmm.h>
#include <arch/mm/mm.h>
#include <arch/task/task.h>

#include <kos/spinlock.h>

struct team {

  /* each team has its own memory context */
  team_mm_context_t mm_context;

  /* each team has an array of struct sres, that allows to "create"
     file descriptors for the user */
  struct ures *ures_array[MAX_FD_PER_TEAM];

  /* a team is composed of a list one thread */
  struct thread  *thread_list;

  /* a team has a memory context, so it has a virtual region tree */
  struct team_address_space address_space;
  
  /* Next / Prev team in global (system-wide) list */
  struct team *prev, *next;
};

struct thread {
  char name[32];
  
  /* each thread is attached to a team */
  team_t    *mother_team;

  enum { KERNEL_THREAD, USER_THREAD } type;
  enum { READY, RUNNING, INTERRUPTED, STOPPED, 
	 DELETED, SLEEPING, WAITING,
	 WAITING_SEM, WAITING_WAIT_QUEUE } state;
  enum { CONTEXT_BLOCKED, CONTEXT_INTERRUPTED, CONTEXT_OBSOLETE } context_type;

  /* each thread has a process context */
  thread_cpu_context_t cpu_context;
  addr_t cpl0_stack_base_addr;

  union {
    
    struct {
      struct ksem   *sem;
      struct thread *prev;
      struct thread *next;
    } waiting_sem_thread;
    
    struct {
      struct kwaitqueue *kwq;
      struct thread     *prev;
      struct thread     *next;
    } waiting_waitqueue_thread;

    struct {
      jiffy_t expiration_time;
      struct thread *prev;
      struct thread *next;
    } sleeping_thread;
  } u;

  thread_t *next_global,  *prev_global;
  thread_t *next_in_team, *prev_in_team;
};

#define create_kernel_thread(t,k,d) _create_kernel_thread(t, k, #k, d)

void thread_end_handler (void);

thread_t * get_current_thread (void);
void       set_current_thread (thread_t *thread);

#define get_current_team() (get_current_thread()->mother_team)

team_t   * get_kernel_team (void);

thread_t * get_global_thread_list_head (void);

/* For create_user_thread & create_kernel_thread, team = NULL means we
   have to crete the thread in the current team */
thread_t * create_user_thread(team_t *team, addr_t eip, void *data);

thread_t *_create_kernel_thread(team_t *team,
				kernel_thread_fct_t *kt,
				char *name,
				void *data);



void delete_user_thread(thread_t *thread);
int  show_all_thread_info (void);
void delete_kernel_thread (thread_t *thread);

/* _task_team.c */
team_t *create_team(void);
int delete_team(team_t *team);

/* _task_sres.c */
result_t task_insert_ures(struct ures *ures, /*out*/ int *fd);
result_t task_replace_ures(int fd, struct ures *ures);
result_t task_delete_ures(int fd);

#endif /* __task_h__ */

