#include <kos/assert.h>

#include <vmm/_vmm.h>

virtual_region_t *cpl0_stack_region = NULL;

/* (visitor)
 * Returns TRUE whenever vr2 intersects vr1
 */
static int _check_overlapping(virtual_region_t *vr,
			      range_t *r)
{
  return ((vr->end > r->start)
	  && (vr->node.key < r->end));
}

/* (visitor)
 * Returns TRUE when vaddr is inside vr
 */
static int _vaddr_owned_by(virtual_region_t *vr,
			   addr_t vaddr)
{
  return ((vaddr >= vr->node.key) && (vaddr <= vr->end));
}

int as_init (team_address_space_t *as,
	     struct team *mother_team)
{
  if (!as)
    return -1;

  as->mother_team = mother_team;
  as->vr_tree     = NULL;

  return 0;
}

/* (resetter)
 * Free vr
 */
static int _vr_cleaner (virtual_region_t *vr,
			void *unused)
{
  UNUSED(unused);
  vr_del(vr);
  return 0; 
}

int as_cleanup(team_address_space_t *as)
{
  bst_reset ((tree_node_t *) as->vr_tree, 
	     (visitor_t) _vr_cleaner,
	     NULL);

  as->vr_tree = NULL;

  return 0;
}

/* (visitor)
 */
static int _as_dup_a_vr(virtual_region_t *vr,
			team_address_space_t *to_as)
{
  virtual_region_t *new_vr;

  new_vr = vr_dup(vr);
  
  if (!new_vr)
    return 1;

  as_add_vr(to_as, new_vr);

  return 0;
}
	       
int as_dup_vr_tree(team_address_space_t *from_as,
		   team_address_space_t *to_as)
{
  if (!from_as || !to_as)
    return -1;

  visit ((tree_node_t*) from_as->vr_tree,
	 (visitor_t) _as_dup_a_vr,
	 (void *) to_as);
	 
  return 0;
}


/* 0 On success */
/* Assumes that vr doesn't overlap any other vr in as */
int as_add_vr(team_address_space_t *as,
	      virtual_region_t *vr)
{
  if (!as || !vr)
    return -1;

  vr->node.left = vr->node.right = vr->node.up = NULL;

  if (! as->vr_tree)
    as->vr_tree = vr;
  else
    (as->vr_tree) = 
      (virtual_region_t *) insert_node( & vr->node,
					& as->vr_tree->node);
  
  DEBUG_PRINT("Region successfully added for [0x%x,0x%x].\n",
	      vr->node.key, vr->end);

  return 0;
}



/* Try to map [res_offset, res_offset+size] of res in as at vaddr_hint if
 * possible.
 * If must_respect_hint == FALSE
 *   if res would prefer some address => vaddr set to the latter
 *      (=> vaddr_hint ignored)
 *
 * With the resulting vaddr, make sure it doesn't overlap anything and
 * add it. Otherwise :
 * If must_respect_hint == FALSE
 *   choose a correct vaddr and add it.
 * else
 * return Error */
struct virtual_region *
as_map_kres(team_address_space_t *as,

	    struct kres *kres,
	    offset_t res_offset,
	    size_t  size,
	    
	    vaddr_t vaddr_hint,
	    bool_t must_respect_hint,
	    
	    sharing_type_t sharing_type,
	    access_right_t access_right)
{
  struct virtual_region *vr;
  range_t new_range;
  result_t result;

  ASSERT_RETURN_VAL(as && kres, NULL);

  /* Make sure the hint is page aligned */
  ASSERT_RETURN_VAL((vaddr_hint & (PAGE_SIZE-1)) == 0, NULL);

  if (must_respect_hint) {
    ASSERT_RETURN_VAL(vaddr_hint > 0, NULL);
  } else {
    struct virtual_region * vr_hint;
    vr_hint = kres_get_suitable_vr_if_exists (kres,
					      res_offset, size,
					      access_right, sharing_type);
    if (vr_hint)
      vaddr_hint = vr_hint->node.key;
  }

  if (vaddr_hint)
    {
      new_range.start = vaddr_hint;
      new_range.end   = vaddr_hint + size;
      
      /* Make sure that the new range won't overlap anything */
      if (as->vr_tree && visit(& as->vr_tree->node,
			       (visitor_t) _check_overlapping,
			       (void *)    & new_range))
	{
	if (must_respect_hint)
	  return NULL;
	
	/* Ignore the hint */
	vaddr_hint = 0;
	}
    }

  /* Here, either the hint is correct or not */
  if (! vaddr_hint) {
    WARNING("TBI: Find a suitable free interval in as\n");
    return NULL;
  }

  new_range.start = vaddr_hint;
  new_range.end   = vaddr_hint + size;

  vr = vr_new(& new_range, sharing_type, access_right,
	      kres, res_offset);
  if (!vr)
    return NULL;
  ASSERT_FATAL(! as_add_vr(as, vr));

  result = kres_add_mapping(kres, vr, FALSE);

  if(result < 0)
    {
      FAILED_VERBOSE("kres_add_mapping failed with %d", result);
    }

  return vr;
}


/*
 * returns the virtual_region owning vaddr
 */
inline static virtual_region_t *_get_vr_at_vaddr(team_address_space_t *as,
						 addr_t vaddr)
{
  if (!as || !as->vr_tree)
    return NULL;

  return (virtual_region_t*)
    visit(& as->vr_tree->node,
	  (visitor_t) _vaddr_owned_by,
	  (void *) vaddr);
}


/*
 * Page fault on vaddr
 * Returns 0 when could map vaddr
 * Otherwise -1 (=> must send SEGV to faulting thread)
 */
int as_page_fault(team_address_space_t *as,
		  vaddr_t vaddr,
		  pf_flags_t *pf)
{
  virtual_region_t *vr;

  vr = _get_vr_at_vaddr(as, vaddr);
  if (! vr)
      return -1;

  return vr_page_fault(vr, vaddr, pf);
}


int as_remove_vr(team_address_space_t *as,
		 virtual_region_t *vr)
{
  if (!as || !vr)
    return -1;

  ASSERT_FATAL(vr->mother_as == as);
  ASSERT_FATAL(vr->mother_as->vr_tree);

  vr->mother_as->vr_tree =
    (virtual_region_t *) delete_node(& vr->node,
				     & vr->mother_as->vr_tree->node);
  return 0;
}


/*
 * Tries to resize virtual region vr up to new_size.
 *
 * Returns 0 when OK
 */
int as_resize_vr(team_address_space_t *as,
		 virtual_region_t *vr,
		 size_t new_size)
{
  if (!as || !vr)
    return -1;

  ASSERT_FATAL(vr->mother_as == as);
  ASSERT_FATAL(vr->kres);

  if (new_size == (vr->end - vr->node.key))
    return 0;
  
  /* Don't need to check for overlapping when new_range is smaller
     than old_range */
  if (new_size > (vr->end - vr->node.key)) {
    size_t  old_size;
    range_t new_range;
    
    /* Prevents visit() below from detecting that the new_range
       overlaps the current range */
    old_size = vr->end - vr->node.key;
    vr->end = vr->node.key;

    new_range.start = vr->node.key;
    new_range.end   = new_range.start + new_size;

    /* Make sure that the new range won't overlap anything */
    if (as->vr_tree && visit(& as->vr_tree->node,
			     (visitor_t) _check_overlapping,
			     (void *)    & new_range))
      {
	/* Bad news => overlaps */
	vr->end = vr->node.key + old_size;
	return -1;
      }
  }

  if (kres_map_resize_notification(vr->kres, vr, new_size))
    return -1;

  /* Ok : everything went right : adding vr with new range */
  vr->end = vr->node.key + new_size;

  return 0;
}


void as_dump (team_address_space_t *as)
{
  __dbg_printk (_B_BLUE "*********** AS DUMP @0x%x ************\n",
		  (unsigned)as);

  visit ( (tree_node_t*) as->vr_tree,
	  (visitor_t) vr_dump,
	  NULL);

  __dbg_printk ("**********************************\n" _B_NORM);
}


__init_text int __init_as_engine()
{
  return 0;
}


