
#include <kos/assert.h>
#include <lib/string.h>
#include <liblist/liblist.h>

#include <kmem/kmem.h>
#include <vmm/_vmm.h>


static struct kslab_cache *vmm_vr_cache;
static struct kslab_cache *vmm_access_range_cache;

static int _access_range_cleaner(access_range_t *access_range,
				void *unused)
{
  UNUSED(unused);

  memset (access_range, 0, sizeof(access_range_t));
  ASSERT_FATAL(! kslab_cache_free(vmm_access_range_cache, access_range));

  return 0;
}

static inline int _vr_del(struct virtual_region *vr)
{
  if (! vr)
    return -1;

  bst_reset ((tree_node_t *) vr->access_range_tree,
	     (visitor_t) _access_range_cleaner,
	     NULL);

  CONCEPTION_ASSERT(vr->kres);
  ASSERT_FATAL(! kres_remove_mapping(vr->kres, vr));

  memset(vr, 0x0, sizeof(struct virtual_region));
  ASSERT_FATAL(! kslab_cache_free(vmm_vr_cache, vr));
  return 0;
}


struct virtual_region *
vr_new(range_t *vaddr_range,

       sharing_type_t sharing_type,
       access_right_t access_right,

       struct kres *kres,
       offset_t res_offset)
{
  struct virtual_region *vr;
  struct access_range *initial_access_range;

  CONCEPTION_ASSERT(vaddr_range && kres);

  vr = kslab_cache_alloc(vmm_vr_cache);
  if (! vr)
    return NULL;

  vr->end               = vaddr_range->end;
  vr->sharing_type      = sharing_type;

  initial_access_range = kslab_cache_alloc(vmm_access_range_cache);
  if (! initial_access_range)
    return NULL;

  initial_access_range->node.left    =
    initial_access_range->node.right =
    initial_access_range->node.up    = NULL;
  initial_access_range->node.key     = vr->node.key;
  initial_access_range->end          = vr->end;
  initial_access_range->access_right = access_right;

  vr->mother_as   = NULL;

  vr->node.key  = vaddr_range->start;
  vr->node.left = vr->node.right = vr->node.up = NULL;
  vr->prev_in_kres = vr->next_in_kres = NULL;

  vr->kres   = kres;
  vr->kres_offset   = res_offset;
  vr->access_range_tree = initial_access_range;

  return vr;
}


static int _access_range_dup_visitor (access_range_t *access_range, 
				      virtual_region_t *to_vr)
{
  access_range_t *new_access_range;

  new_access_range = kslab_cache_alloc(vmm_access_range_cache);

  if (! vmm_access_range_cache)
    return 1;
  
  memset (new_access_range, 0, sizeof(access_range_t));

  new_access_range->node.key     = access_range->node.key;
  new_access_range->access_right = access_range->access_right;
  new_access_range->end          = access_range->end;

  if (! to_vr->access_range_tree)
    to_vr->access_range_tree = new_access_range;
  else
    to_vr->access_range_tree =
      (access_range_t *) insert_node (& new_access_range->node,
				      & to_vr->access_range_tree->node);

  return 0;
}

/* allocates+duplicates a virtual region. the new virtual region is
   ready to be inserted in the new tree. */
virtual_region_t *vr_dup(virtual_region_t *vr)
{
  virtual_region_t *new_vr;
  
  if (!vr)
    return NULL;

  new_vr = kslab_cache_alloc(vmm_vr_cache);
  
  if (!new_vr)
    return NULL;

  memcpy(new_vr, vr, sizeof(virtual_region_t));
  new_vr->node.left = new_vr->node.right = new_vr->node.up = NULL;
  new_vr->prev_in_kres = new_vr->next_in_kres = NULL;  

  visit ((tree_node_t*) new_vr->access_range_tree,
	 (visitor_t) _access_range_dup_visitor,
	 (void *) new_vr);

  kres_add_mapping(new_vr->kres, new_vr, TRUE);
  
  return new_vr;
}


int vr_del(struct virtual_region *vr)
{
  return _vr_del(vr);
}


/* (visitor)
 * Returns TRUE when vaddr is inside ar
 */
static int _vaddr_owned_by(access_range_t *ar,
			   addr_t vaddr)
{
  return ((vaddr >= ar->node.key) && (vaddr <= ar->end));
}


/*
 * returns the virtual_region owning vaddr
 */
inline static access_range_t *_get_ar_at_vaddr(virtual_region_t *vr,
					       addr_t vaddr)
{
  if (!vr || !vr->access_range_tree)
    return NULL;

  return (access_range_t*)
    visit(& vr->access_range_tree->node,
	  (visitor_t) _vaddr_owned_by,
	  (void *) vaddr);
}


/*
 * Page fault on vaddr
 * Returns 0 when could map vaddr
 * Otherwise -1 (=> must send SEGV to faulting thread)
 */
int vr_page_fault(virtual_region_t *vr,
		  vaddr_t vaddr,
		  pf_flags_t *pf)
{
  access_range_t *ar;

  ar = _get_ar_at_vaddr(vr, vaddr);
  if (! ar)
    return -1;

  if ((pf->access & ar->access_right) != pf->access) {
    /* Unauthorized access to this area */
    return -1;
  }

  CONCEPTION_ASSERT(vr->kres);
  return kres_page_fault(vr->kres, vr, vaddr, ar, pf);
}


static inline int _access_range_dump_visitor (access_range_t *access_range, 
					      void *unused)
{
  struct {
    char r, w, x, zero;
  } rwx;

  UNUSED(unused);

#define CASE_STR(esac) case esac: str = # esac; break

  if (access_range->access_right | VM_ACCESS_READ)
    rwx.r = 'R';
  else
    rwx.r = '.';
  if (access_range->access_right | VM_ACCESS_WRITE)
    rwx.w = 'W';
  else
    rwx.w = '.';
  if (access_range->access_right | VM_ACCESS_EXEC)
    rwx.x = 'X';
  else
    rwx.x = '.';
  rwx.zero ='\0';

  __dbg_printk("Range [0x%x ; 0x%x] : %s",
	       access_range->node.key,
	       access_range->end,
	       (char*) &rwx);

  return 0;
}

int access_range_dump (access_range_t *access_range)
{
  _access_range_dump_visitor (access_range, NULL);
  return 0;
}

static inline int _vr_dump_visitor (virtual_region_t *vr, void *unused)
{
  char *str;
  UNUSED(unused);

  __dbg_printk("Virtual region at 0x%x : [0x%x,0x%x], ", 
	       (unsigned)vr, vr->node.key, vr->end);
  __dbg_printk("Mother as=0x%x, ",
	       (unsigned)vr->mother_as);
  __dbg_printk("Kernel resource 0x%x (kres_offset=0x%x, %s on PT), ",
	       (unsigned)vr->kres,
	       vr->kres_offset,
	       vr->is_alone_on_pt?"alone":"not alone");

#define CASE_STR(esac) case esac: str = # esac; break

  switch (vr->sharing_type)
    {
      CASE_STR(MAP_PRIVATE);
      CASE_STR(MAP_SHARED);
    default:
      str = "unknown";
    }
  __dbg_printk ("Sharing=%s\n", str);

  visit (& vr->access_range_tree->node,
	 (visitor_t) _access_range_dump_visitor,
	 NULL);
  
  return TRUE;
}

int vr_dump_tree (virtual_region_t *vr_root)
{
  visit (& vr_root->node,
	 (visitor_t) _vr_dump_visitor,
	 NULL);
  return 0;
}

int vr_dump (virtual_region_t *vr)
{
  _vr_dump_visitor (vr, NULL);
  return 0;
}

__init_text int __init_vr_engine()
{
  vmm_vr_cache = kslab_cache_create("vmm_vr_cache",
				    sizeof(struct virtual_region),
				    0, 0, 0);
  if (! vmm_vr_cache)
    return -1;

  vmm_access_range_cache = kslab_cache_create("vmm_access_range_cache",
					      sizeof(struct virtual_region),
					      0, 0, 0);
  if (! vmm_access_range_cache)
    return -1;


  return 0;
}
