#ifndef __VMM_H__
#define __VMM_H__

#include <kos/types.h>
#include <libbst/libbst.h>
#include <arch/mm/mm.h>
#include <karm/kres.h>

typedef enum { MAP_PRIVATE=0x20, MAP_SHARED } sharing_type_t;

typedef struct pf_flags
{
  enum { PF_PRESENT=0x10, PF_NOT_PRESENT } page_present;

  access_right_t access; 

  enum { PF_ACCESS_USER=0x20, PF_ACCESS_SUPERVISOR } who;
} pf_flags_t;

typedef struct team_address_space
{
  struct team *mother_team;

  /* Tree of virtual regions */
  struct virtual_region  *vr_tree;
 
} team_address_space_t;

typedef struct access_range
{
  // the key is the range start addr 
  tree_node_t node;

  vaddr_t end;
  access_right_t access_right;
} access_range_t;


typedef struct virtual_region 
{
  // the key is the region start address
  tree_node_t node;

  vaddr_t                   end;
  struct team_address_space *mother_as;

  sharing_type_t            sharing_type;

  access_range_t            *access_range_tree;

  bool_t                    is_alone_on_pt;

  /* The mapped kres */
  struct kres *kres;

  /* The offset in the ures which correspond to the beginning of the
     virtual region */
  offset_t kres_offset;
  
  /* All virtual regions that map the same kres are linked together
     (in order to be able to do PT sharing) */
  struct fres *prev_in_kres, *next_in_kres;
} virtual_region_t;


typedef struct vmm_kres_drv {
  /* Returns TRUE when 
     vaddr has been successfully mapped
     Returns FALSE otherwise (=> SEGV will be sent to faulting thread */
  bool_t (*map)(struct virtual_region*,
		vaddr_t vaddr,
		access_range_t *ar,
		pf_flags_t *pf);
  size_t (*sync)(struct virtual_region*,
		 vaddr_t vaddr_start, size_t sz);
} vmm_kres_drv_t;


#include <task/task.h>

/* ****************************************************
 * Team address space
 */

int as_init (team_address_space_t *as,
	     struct team *mother_team);

int as_cleanup(team_address_space_t *as);

int as_dup_vr_tree(team_address_space_t *from_as,
		   team_address_space_t *to_as);

/* Try to map [sr_offset, sr_offset+size] of sr in as at vaddr_hint if possible.
 * If must_respect_hint == FALSE
 *   if sr would prefer some address => vaddr set to the latter (=> vaddr_hint ignored)
 *
 * With the resulting vaddr, make sure it doesn't overlap anything and
 * add it. Otherwise :
 * If must_respect_hint == FALSE
 *   choose a correct vaddr and add it.
 * else
 *   return Error
 */
struct virtual_region *
as_map_kres(team_address_space_t *as,
	    struct kres *kres,
	    offset_t res_offset,
	    size_t  size,
	    vaddr_t vaddr_hint,
	    bool_t must_respect_hint,
	    sharing_type_t sharing_type,
	    access_right_t access_right);

int as_remove_vr(team_address_space_t *as,
		 virtual_region_t *vr);


/*
 * Page fault on vaddr
 * Returns 0 when could map vaddr
 * Otherwise -1 (=> must send SEGV to faulting thread) */
int as_page_fault(team_address_space_t *as,
		  vaddr_t vaddr,
		  pf_flags_t *pf);

/*
 * Tries to resize virtual region vr up to new_size.
 *
 * Returns 0 when OK
 */
int as_resize_vr(team_address_space_t *as,
		 virtual_region_t *vr,
		 size_t new_size);

void as_dump (team_address_space_t *as);


/* ****************************************************
 * Virtual region
 */

struct virtual_region *
vr_new(range_t *vaddr_range,
       sharing_type_t sharing_type,
       access_right_t access_right,
       struct kres *kres,
       offset_t shadow_offset);

int vr_del(struct virtual_region *vr);


/* ****************************************************
 * For task module
 */

int init_team_as  (struct team *team);

/* When team is NULL, map/unmap in current team */
int map_virtual_page(struct team* dest_team,
		     addr_t virt, addr_t phys,
		     access_right_t access_rights);
int unmap_virtual_range(struct team* dest_team,
			range_t *range);
int unmap_virtual_page(struct team* dest_team,
		       vaddr_t vaddr);
int protect_virtual_page(struct team* dest_team,
			 vaddr_t vaddr,
			 access_right_t access_rights);

#endif
