#ifndef __INTERNALS_H__
#define __INTERNALS_H__

#include <kos/system.h>
#include <arch/mm/mm.h>

/***
 * GDT segments IDs
 */

#define SEG_NB               7
#define SEG_NULL_ID          0
#define SEG_CODE_KERNEL_ID   1
#define SEG_DATA_KERNEL_ID   2
#define SEG_CODE_USER_ID     3
#define SEG_DATA_USER_ID     4
#define SEG_TSS              5
#define SEG_DOUBLE_FAULT_TSS 6


/***
 * PDE Indexes of various special regions inside each virtual address
 * space
 */

/* 4M for mapping the PTs of the current team. MUST be consistent with
   loader/mm/mm.h */
#define CURRENT_PT_AREA_INDEX  509

/* 4M for mapping all the PD of all the teams. Needed for
   synchronizing the PTs of the kernel space among all the teams. */
#define PD_TABLE_MAPPING_INDEX 510

/* 4M for mapping all the PTs of a foreign team inside current
   team. */
#define FOREIGN_PT_AREA_INDEX  511


/***
 * Helpers for accessing the pde/pte of the special virtual mappings
 */
#define VADDR_AT_PD_INDEX(index_in_pd) \
  (1UL*(index_in_pd)*1024*PAGE_SIZE)
#define CURRENT_PT_AREA_START  VADDR_AT_PD_INDEX(CURRENT_PT_AREA_INDEX)
#define PD_TABLE_MAPPING_START VADDR_AT_PD_INDEX(PD_TABLE_MAPPING_INDEX)
#define FOREIGN_PT_AREA_START  VADDR_AT_PD_INDEX(FOREIGN_PT_AREA_INDEX)

#define CURRENT_PD_VADDR \
  (CURRENT_PT_AREA_START + CURRENT_PT_AREA_INDEX*PAGE_SIZE)
#define CURRENT_PDE(index_in_pd) \
  (*((page_entry_t*)(CURRENT_PD_VADDR + (index_in_pd)*sizeof(page_entry_t))))
#define CURRENT_PDE_PADDR(index_in_pd) \
  (CURRENT_PDE(index_in_pd) & PAGE_ADDR_MASK)
#define CURRENT_PD_PADDR \
  (CURRENT_PDE_PADDR(CURRENT_PT_AREA_INDEX))
#define FOREIGN_PD_PADDR \
  (CURRENT_PDE_PADDR(FOREIGN_PT_AREA_INDEX))

/* The array of page entries (phys addr | flags) in current pd_mapping
   zone of any other PD */
#define PT_OF_ALL_PD_START \
  (CURRENT_PT_AREA_START + PD_TABLE_MAPPING_INDEX*PAGE_SIZE)
/* Get a precise PD (phys addr | flags) of the PD of any other team */
#define PD_ANY_TEAM(team_index) \
  (*((page_entry_t*)(PT_OF_ALL_PD_START + (team_index)*sizeof(page_entry_t))))
/* Get the address of a precise PDE of any other team */
#define VADDR_PDE_ANY_TEAM(team_index,index_in_pd) \
  ((page_entry_t*)(PD_TABLE_MAPPING_START + (team_index)*PAGE_SIZE + (index_in_pd)*sizeof(k_ui32_t)))

#define FOREIGN_PD_VADDR \
  (CURRENT_PT_AREA_START + FOREIGN_PT_AREA_INDEX*PAGE_SIZE)
#define FOREIGN_PDE(index_in_pd) \
  (*((page_entry_t*)(FOREIGN_PD_VADDR + (index_in_pd)*sizeof(page_entry_t)))
#define FOREIGN_PDE_PADDR(index_in_pd) \
  (FOREIGN_PDE(index_in_pd) & PAGE_ADDR_MASK)

/***
 * Assertions
 */
#if (CURRENT_PT_AREA_INDEX > KERNEL_MEMORY_NUMBER_OF_PT)
#  error "Invalid CURRENT_PT_AREA_INDEX (should be <= KERNEL_MEMORY_NUMBER_OF_PT)"
#endif
#if (PD_TABLE_MAPPING_INDEX > KERNEL_MEMORY_NUMBER_OF_PT)
#  error "Invalid PD_TABLE_MAPPING_INDEX (should be <= KERNEL_MEMORY_NUMBER_OF_PT)"
#endif
#if (FOREIGN_PT_AREA_INDEX > KERNEL_MEMORY_NUMBER_OF_PT)
#  error "Invalid FOREIGN_PT_AREA_INDEX (should be <= KERNEL_MEMORY_NUMBER_OF_PT)"
#endif

#if (CORE_KERNEL_VIRTUAL_ADDR < KERNEL_MEMORY_START) \
    || (CORE_KERNEL_VIRTUAL_ADDR+CORE_KERNEL_VIRTUAL_SIZE > KERNEL_MEMORY_END)
#  error "Kernel is located outside of kernel virtual address space"
#endif
#if (CPL0_KERNEL_STACK_VIRTUAL_ADDR < KERNEL_MEMORY_START) \
    || (CPL0_KERNEL_STACK_VIRTUAL_ADDR+CPL0_KERNEL_STACK_VIRTUAL_REGION_SIZE \
        > KERNEL_MEMORY_END)
#  error "CPL0 stacks located outside of kernel virtual address space"
#endif

#if (KERNEL_MEMORY_END > USER_SPACE_START)
#  error "User space starts before kernel ends !"
#endif

#endif /* __INTERNALS_H__ */
