#include <kmem/kmem.h>
#include <kos/assert.h>
#include <kos/macros.h>
#include <kos/spinlock.h>
#include <liblist/liblist.h>

/* I know it's forbidden to include private headers, but here, I think
   it's much better to include a private file that putting in the
   public header things that are only needed here, such as
   chained_page_range. */
#include <kmem/_kvmem.h>

#include "_mm.h"

static struct kslab_cache *_rmap_cache;

static int _nop_pre_add_rmapping(struct rmap_session_s *rmap_session);
static int _nop_do_add_rmapping(struct rmap_session_s *rmap_session,
				k_ui32_t index_team, gpfme_t *gpfme, 
				vaddr_t virt, bool_t is_pt);
static int _nop_post_add_rmapping(struct rmap_session_s *);

static int _nop_pre_del_rmapping(struct rmap_session_s *rmap_session);
static int _nop_do_del_rmapping(struct rmap_session_s *rmap_session,
				k_ui32_t index_team, gpfme_t *gpfme, 
				vaddr_t virt);
static int _nop_post_del_rmapping(struct rmap_session_s *);

/***********************************************************************
 * Initial rmap operations, used until the kslab is initialised:
 * only update the reference counters
 */
struct _rmap_ops_s _rmap_ops = {
  _nop_pre_add_rmapping,
  _nop_do_add_rmapping,
  _nop_post_add_rmapping,

  _nop_pre_del_rmapping,
  _nop_do_del_rmapping,
  _nop_post_del_rmapping
};


__init_text static int
 _nop_pre_add_rmapping(struct rmap_session_s *rmap_session)
{
  UNUSED(rmap_session);
  return 0;
}


__init_text static int
_nop_do_add_rmapping(struct rmap_session_s *rmap_session,
		     k_ui32_t index_team, gpfme_t *gpfme, 
		     vaddr_t virt, bool_t is_pt)
{
  UNUSED(rmap_session);
  UNUSED(index_team);
  UNUSED(is_pt);

  CONCEPTION_ASSERT(gpfme && virt);

  CONCEPTION_ASSERT(gpfme->mapping_list == NULL);
  gpfme->ref_cnt ++;

  return 0;
}

__init_text static int
_nop_post_add_rmapping(struct rmap_session_s *rmap_session)
{
  UNUSED(rmap_session);
  return 0;
}

__init_text static int
_nop_pre_del_rmapping(struct rmap_session_s *rmap_session)
{
  UNUSED(rmap_session);
  return 0;
}

__init_text static int
_nop_do_del_rmapping(struct rmap_session_s *rmap_session,
		     k_ui32_t index_team, gpfme_t *gpfme, 
		     vaddr_t virt)
{
  UNUSED(index_team);
  UNUSED(rmap_session);

  CONCEPTION_ASSERT(gpfme && virt);
  CONCEPTION_ASSERT((gpfme->ref_cnt > 0)
		    && (gpfme->mapping_list == NULL));
  gpfme->ref_cnt --;

  return 0;
}


__init_text static int
_nop_post_del_rmapping(struct rmap_session_s *rmap_session)
{
  UNUSED(rmap_session);
  return 0;
}


/***********************************************************************
 * Real rmap operations: called once kslab is initialised
 */

/*
 * Some explanations concerning rmap_session and prepare/do/end
 * add/del callbacks
 *
 * Problem description: basically, these callbacks can ONLY be called
 * by vmap. During the vmap operations, the gpfm, and the mm_context
 * or kernel_space locks are owned by the calling thread. BUT, the
 * rmap need new rmap structures, which implies calling the kslab
 * facilities. BUT the kslab relies on the vmap. So that we have the
 * following deadlock in the case the original page to be mapped was a
 * kernel page: vmap -> lock(kernel_space) -> rmap -> kslab -> vmap ->
 * lock(kernel_space)
 *
 * Possibile solutions:
 * - call the rmap functions outside the lock => BAD idea because the
 *   swapper might appear between a new page vmap'ped and the page
 *   being rmap'ped => the swapper won't be able to consider the new
 *   page.
 * - Create a rmap pool management API. This is redundant with the
 *   kslab, and does not avoid the deadlock problem, because any new
 *   pool page that needs be allocated, needs also be vmap'ped...
 * - "pre-allocate" the rmap structures BEFORE the lock, and release
 *   any unused rmap after the rmap_add/del and after the lock.
 *
 * => We implemented the last approach.
 *
 * Implementation considerations:
 * - vmap needs at most allocate a normal page AND a PT while holding
 *   the lock. Never more, maybe less (PT unused).
 * - It would be unadapted to always pre-allocate the rmaps for both
 *   the page AND the potential PT to allocate, precisely because a
 *   new PT is scarcely needed => we use the spare_pt_rmap rmap data
 *   to store any previously allocated but unused PT rmap structure.
 */

static SPINLOCK(spare_pt_rmap_lock);
static struct mapping_s* spare_pt_rmap = NULL;

static int
_pre_add_rmapping(struct rmap_session_s *rmap_session)
{
  k_ui32_t flags;

  CONCEPTION_ASSERT(rmap_session);

  /* If any spared PT rmapping is available, use it (this prevents
     from allocating another one) */
  write_spin_lock(spare_pt_rmap_lock, flags);
  rmap_session->pt_rmap = spare_pt_rmap;
  spare_pt_rmap = NULL;
  write_spin_unlock(spare_pt_rmap_lock, flags);

  /* If spare PT rmap was not available, allocate one */
  if (rmap_session->pt_rmap == NULL)
    rmap_session->pt_rmap = kslab_cache_alloc(_rmap_cache);

  /* Allways get a "normal" page mapping */
  rmap_session->page_rmap = kslab_cache_alloc(_rmap_cache);

  CONCEPTION_ASSERT(rmap_session->pt_rmap && rmap_session->page_rmap);

  return 0;
}


static int
_do_add_mapping(struct rmap_session_s *rmap_session,
		k_ui32_t index_team, gpfme_t *gpfme, 
		vaddr_t virt, bool_t is_pt)
{
  struct mapping_s* rmap;

  CONCEPTION_ASSERT(rmap_session && gpfme && virt);

  /* Retrieve the right rmap to update from the rmap_session structure */
  if (is_pt)
    {
      CONCEPTION_ASSERT(rmap_session->pt_rmap);
      rmap = rmap_session->pt_rmap;
      rmap_session->pt_rmap = NULL; /* mark it as used */
    }
  else
    {
      CONCEPTION_ASSERT(rmap_session->page_rmap);
      rmap = rmap_session->page_rmap;
      rmap_session->page_rmap = NULL; /* mark it as used */
    }

  CONCEPTION_ASSERT(rmap);

  /* Actually update the rmap */
  rmap->index_team = index_team;
  rmap->index_pde  = va_to_pde(virt);
  rmap->index_pte  = va_to_pte(virt);
  rmap->is_pt      = is_pt;
  rmap->next       = NULL;
  
  /* Add the new rmap to the gpfme */
  gpfme->ref_cnt++;
  if(gpfme->mapping_list != NULL)
    rmap->next = gpfme->mapping_list;
  gpfme->mapping_list = rmap;

  return 0;
}


static int
_post_add_rmapping(struct rmap_session_s *rmap_session)
{
  k_ui32_t flags;
  CONCEPTION_ASSERT(rmap_session);

  /* Reclaim unused PT rmap: if spare_pt_rmap is free, don't free the
     PT rmap => assign it to spare_pt_rmap instead for future use by
     another add_rmapping */
  if (rmap_session->pt_rmap)
    {

      /* Try to assign the PT rmap to release to the spare PT */
      write_spin_lock(spare_pt_rmap_lock, flags);
      if (spare_pt_rmap == NULL)
	{
	  /* Ok, assign unused PT rmap to spare PT */
	  spare_pt_rmap = rmap_session->pt_rmap;
	  rmap_session->pt_rmap = NULL;
	}
      /* else: release unused PT rmap manually -> see below */
      write_spin_unlock(spare_pt_rmap_lock, flags);

      /* If spare PT rmap was already assigned, free the current PT rmap */
      if (rmap_session->pt_rmap != NULL)
	kslab_cache_free(_rmap_cache, rmap_session->pt_rmap);
    }

  /* Should never occur (the _do_add_mapping() should normally have
     always been called between _pre_add/_post_add, so that the
     "normal" page_rmap is always used */
  if (rmap_session->page_rmap)
    {
      write_spin_lock(spare_pt_rmap_lock, flags);
      if (spare_pt_rmap == NULL)
	{
	  /* Ok, assign unused rmap to spare PT */
	  spare_pt_rmap = rmap_session->page_rmap;
	  rmap_session->page_rmap = NULL;
	}
      /* else: release unused PT rmap manually -> see below */
      write_spin_unlock(spare_pt_rmap_lock, flags);

      if (rmap_session->page_rmap)
	kslab_cache_free(_rmap_cache, rmap_session->page_rmap);
    }

  return 0;
}


static int
_pre_del_rmapping(struct rmap_session_s *rmap_session)
{
  CONCEPTION_ASSERT(rmap_session);

  /* Indicate that none is to be freed by the _post_del_rmap() */
  rmap_session->page_rmap = NULL;
  rmap_session->pt_rmap   = NULL;

  return 0;
}

static int _do_del_mapping(struct rmap_session_s *rmap_session,
			   k_ui32_t index_team, gpfme_t *gpfme,
			   vaddr_t virt)
{
  struct mapping_s *current, *prev;
  int index_pde, index_pte;

  CONCEPTION_ASSERT(rmap_session && gpfme && virt);

  index_pde = va_to_pde(virt);
  index_pte = va_to_pte(virt);

  CONCEPTION_ASSERT(gpfme->ref_cnt > 0);

  prev = NULL;
  for ( current = gpfme->mapping_list ;
	current != NULL ;
	prev = current, current = current->next)
    if ( (current->index_team == index_team)
	 && (current->index_pde == index_pde)
	 && (current->index_pte == index_pte) )
      break;

  /* We did not find the requested mapping ! */
  CONCEPTION_ASSERT(current);

  /* Del the mapping */
  if (prev != NULL)
    prev->next = current->next;
  else
    gpfme->mapping_list = current->next;

  gpfme->ref_cnt --;

  /* Mark the rmap structure to be freed by _post_del_rmap() */
  if (current->is_pt) 
    {
      CONCEPTION_ASSERT(rmap_session->pt_rmap == NULL);
      rmap_session->pt_rmap = current;
    }
  else
    {
      CONCEPTION_ASSERT(rmap_session->page_rmap == NULL);
      rmap_session->page_rmap = current;
    }

  return 0;
}

/* Actually delete any "marked to be deleted" PT/page rmap */
static int
_post_del_rmapping(struct rmap_session_s *rmap_session)
{
  CONCEPTION_ASSERT(rmap_session);

  /* The rmap_session's PT is to be deleted */
  if (rmap_session->pt_rmap)
    {
      k_ui32_t flags;

      /* Try to assign the PT rmap to release to the spare PT */
      write_spin_lock(spare_pt_rmap_lock, flags);
      if (spare_pt_rmap == NULL)
	{
	  /* Ok, assign PT rmap to spare PT */
	  spare_pt_rmap = rmap_session->pt_rmap;
	  rmap_session->pt_rmap = NULL;
	}
      /* else: release PT rmap manually -> see below */
      write_spin_unlock(spare_pt_rmap_lock, flags);

      /* If spare PT rmap was already assigned, free the current PT rmap */
      if (rmap_session->pt_rmap != NULL)
	kslab_cache_free(_rmap_cache, rmap_session->pt_rmap);
    }

  /* The normal page rmap is to be deleted */
  if (rmap_session->page_rmap)
    kslab_cache_free(_rmap_cache, rmap_session->page_rmap);

  return 0;
}

/* When gpfme not mapped (but set to ref_cnt=1 in _pmm_init), return
   the physical address of the page for it to be further
   suppressed. Or return 0 (mapped Ok, or not mapped). */
__init_text static int _check_empty(gpfme_t* gpfme, void* custom_param)
{
  UNUSED(gpfme);
  UNUSED(custom_param);

  if (gpfme->ref_cnt == 0)
    /* Nobody */
    return 0;

  if (gpfme->mapping_list)
    /* Ok, correcly rmapped */
    return 0;

  WARNING(_B_RED "gpfme for paddr=0x%x, type=%d, swap_status=%d, ref_cnt=%d not rmapped => suppression in progress\n" _B_NORM,
	  gpfme->address, gpfme->flags.page_type,
	  gpfme->flags.swap_status, gpfme->ref_cnt);

  /* Mark gpfme to be suppressed */
  gpfme->ref_cnt = 0;
  return gpfme->address;
}

__init_text static int _rmap_update_vpage(vaddr_t vaddr, bool_t is_pt)
{
  gpfme_t* gpfme;
  k_ui32_t flags;
  struct rmap_session_s rmap_session;
  
  _rmap_ops.pre_add_rmapping(& rmap_session);

  gpfme = get_gpfme_at_virt_addr(vaddr, & flags);
  if ((!gpfme) || (gpfme->mapping_list != NULL))
    {
      gpfme_unlock(gpfme, & flags);
      _rmap_ops.post_add_rmapping(& rmap_session);
      return 0;
    }

  /* Got one */
  _rmap_ops.do_add_rmapping(& rmap_session, PRIMARY_TEAM_PD_ID, gpfme,
			    vaddr, is_pt);
  gpfme->ref_cnt = 1;
  gpfme_unlock(gpfme, & flags);
  _rmap_ops.post_add_rmapping(& rmap_session);

  return 0;
}

/*
 * Assumes that called during the init phase (no lock acquired !!!)
 */
__init_text static int _rmap_update_kernel_area(void)
{
  int i;
  vaddr_t vaddr;
  paddr_t ppage_to_suppress;
  struct chained_page_range *page_range;
  
  /* We only check the 512 first entries, since we are at the
     beginning of the initialisation, there's not PT in the user space */
  for (i = 0 ; i < 512 ; i++)
    {
      page_entry_t *pd = (page_entry_t*) CURRENT_PD_VADDR;

      if(IS_UNMAPPED(pd[i]))
	continue;
  
      vaddr = CURRENT_PT_AREA_START + i*PAGE_SIZE;
      _rmap_update_vpage(vaddr, TRUE);
    }

  /* For each used range of the kvmem list, we add all pages */
  list_foreach(__kvmem_get_used_page_range_list(), page_range, i)
    {
      for (vaddr = page_range->start;
	   vaddr < (page_range->start + page_range->nb_pages * PAGE_SIZE);
	   vaddr += PAGE_SIZE)
	{
	  _rmap_update_vpage(vaddr, FALSE);
	}
    }

  /* Free any unmapped gpfme */
  do {
    ppage_to_suppress = 
      _gpfm_visit_list_unsafe(PHYS_PAGE_USER, _check_empty, NULL);
    if (ppage_to_suppress != 0)
      put_physical_page(ppage_to_suppress);
  } while (ppage_to_suppress != 0);

  do {
    ppage_to_suppress = 
      _gpfm_visit_list_unsafe(PHYS_PAGE_KERNEL, _check_empty, NULL);
    if (ppage_to_suppress != 0)
      put_physical_page(ppage_to_suppress);
  } while (ppage_to_suppress != 0);

  do {
    ppage_to_suppress = 
      _gpfm_visit_list_unsafe(PHYS_PAGE_HW_MAPPING, _check_empty, NULL);
    if (ppage_to_suppress != 0)
      put_physical_page(ppage_to_suppress);
  } while (ppage_to_suppress != 0);

  return 0;
}


__init_text int _rmap_init(kernel_parameter_t *kp)
{
  UNUSED(kp);

  /* Make sure we have enough margin to eventually allocate 2 pages :
     one for an enlargement of the rmap slab cache, one for an
     eventual PT associated with it. */
#define RMAP_SLAB_MARGIN 16 /* 2 should be enough. 16 is also fine ;) */
  _rmap_cache = kslab_cache_create("rmap", sizeof(struct mapping_s), 0, 0,
				   RMAP_SLAB_MARGIN);

  if (_rmap_cache == NULL)
    FAILED_VERBOSE("(init_rmap) _rmap.c allocation failed\n");

  _rmap_ops.pre_add_rmapping  = _pre_add_rmapping;
  _rmap_ops.do_add_rmapping   = _do_add_mapping;
  _rmap_ops.post_add_rmapping = _post_add_rmapping;

  _rmap_ops.pre_del_rmapping  = _pre_del_rmapping;
  _rmap_ops.do_del_rmapping   = _do_del_mapping;
  _rmap_ops.post_del_rmapping = _post_del_rmapping;
  
  _rmap_update_kernel_area();

  return 0;
}

