/*
 * Copyright (C) 2000, Thomas PETAZZONI
 * http://kos.enix.org
 *
 *
 * @(#) $Id: _team_mm_context.c,v 1.1.1.1 2003/04/27 12:49:49 thomas Exp $
 */

#include <loader/mod.h>

#include <debug/debug.h>
#include <kos/assert.h>

#include <pmm/pmm.h>
#include <task/task.h>
#include <arch/lib/lib.h>
#include <lib/string.h>
#include <kos/spinlock.h>

#include "_mm.h"

/*
 * For each team, pd[PD_TABLE_MAPPING_INDEX] = ptr to the PT containing
 * the addresses of all the PDs.
 */
static page_entry_t *pd_table_mapper;

static k_ui32_t team_mm_contexts_bitmap[ALIGN_SUP(MAX_TEAM_NUMBER, NBITS(k_ui32_t))/NBITS(k_ui32_t)];

/* Task address space allocator. It does not just allocate a physical
   page for the PD of the task, but also write the address of the PD
   in a special PT, so as the PD to be accessible for the kernel. This
   is necessary in order to avoid context switch when maintaining
   kernel space coherency */

int arch_team_mm_context_construct(team_mm_context_t* mm_ctxt)
{
  unsigned int index;
  k_ui32_t flags;
  page_entry_t *pd;

  CONCEPTION_ASSERT(mm_ctxt != NULL);

  write_spin_lock(kernel_team_pd_bitmap_lock, flags);
  index =  bsf_bitmap((addr_t) team_mm_contexts_bitmap,
		      sizeof(team_mm_contexts_bitmap));
  if(index >= NBITS(team_mm_contexts_bitmap))
    {
      write_spin_unlock(kernel_team_pd_bitmap_lock, flags);
      return -1;
    }

  /* Mark this slot as used in the bitmap */
  set_bit_in_bitmap((addr_t) team_mm_contexts_bitmap, index, 0);
  write_spin_unlock(kernel_team_pd_bitmap_lock, flags);
  
  /* Allocate new PD for the team */
  mm_ctxt->pd = get_physical_page(PHYS_PAGE_KERNEL, PHYS_PAGE_NON_SWAPPABLE);
  __dbg_printk("Physical page allocated for this PD is 0x%x\n", 
	       mm_ctxt->pd);
  
  if(!(mm_ctxt->pd))
    return -1;

  /* Insert the address of the PD in the correct slot of the list,
     making this PD available in all teams at
     PD_TABLE_MAPPING_START+index*PAGE_SIZE == pd */
  mm_ctxt->pd_index_in_pd_table = index;
  pd_table_mapper[index]  = (page_entry_t) mm_ctxt->pd;
  pd_table_mapper[index] |= (PAGE_SUPERVISOR | PAGE_RW | PAGE_PRESENT);

  memset((char *) (PD_TABLE_MAPPING_START + index*PAGE_SIZE), 0, PAGE_SIZE);

  /* Copy the contents of kernel team 0 PD into newly created team's PD */
  memcpy((char *) (PD_TABLE_MAPPING_START + index*PAGE_SIZE),
	 (char *) (PD_TABLE_MAPPING_START + PRIMARY_TEAM_PD_ID*PAGE_SIZE),
	 KERNEL_MEMORY_NUMBER_OF_PT * sizeof(k_ui32_t));

  /* Enable mirroring on the new PD. This MUST be done in order to
     always have the PD of the current team mapped at the same
     (CURRENT_PT_AREA_INDEX) PDE.
     This MUST be done AFTER the memcpy above to avoid having same
     mirroring as kernel team 0.
  */
  pd = (page_entry_t *) (PD_TABLE_MAPPING_START + index*PAGE_SIZE);
  pd[CURRENT_PT_AREA_INDEX] = (page_entry_t) (mm_ctxt->pd);
  pd[CURRENT_PT_AREA_INDEX] |= (PAGE_SUPERVISOR | PAGE_RW | PAGE_PRESENT);

  spinlock_init(mm_ctxt->user_space_lock);

  return 0;
}


int arch_team_mm_context_destruct(team_mm_context_t* mm_ctxt)
{
  int index;
  paddr_t pd;
  k_ui32_t flags;
  RETURN_VAL_IF_FAIL_VERBOSE(mm_ctxt != NULL, -1);

  /* Make sure we don't free the main/first/parent/ancestor kernel
     team */
  index = mm_ctxt->pd_index_in_pd_table;
  pd    = mm_ctxt->pd;
  CONCEPTION_ASSERT(index != PRIMARY_TEAM_PD_ID);

  /* Make sure the pd was marked allocated */
  write_spin_lock(kernel_team_pd_bitmap_lock, flags);
  CONCEPTION_ASSERT(get_bit_in_bitmap((addr_t) team_mm_contexts_bitmap,
				      index) == 0);

  /* Mark team's PD as free (put_physical_page is below) */
  mm_ctxt->pd = 0;
  mm_ctxt->pd_index_in_pd_table = -1;

  /* Mark team's PD deallocated */
  pd_table_mapper[index] = 0;
  set_bit_in_bitmap((addr_t) team_mm_contexts_bitmap, index, 1);
  write_spin_unlock(kernel_team_pd_bitmap_lock, flags);

  /* We do this outside spinlock */
  put_physical_page(pd);

  return 0;
}

int arch_init_kteam_mm_context(team_mm_context_t* mm_ctxt)
{
  /* Set the PHYSICAL address of the current PD */
  mm_ctxt->pd = CURRENT_PD_PADDR;
  mm_ctxt->pd_index_in_pd_table = PRIMARY_TEAM_PD_ID;

  /* The first entry of the PT is used for the current team */
  pd_table_mapper[PRIMARY_TEAM_PD_ID]  = (page_entry_t)mm_ctxt->pd;
  pd_table_mapper[PRIMARY_TEAM_PD_ID] |= (PAGE_SUPERVISOR | PAGE_RW | PAGE_PRESENT);
  
  set_bit_in_bitmap((addr_t) team_mm_contexts_bitmap,
		    PRIMARY_TEAM_PD_ID, 0);

  spinlock_init(mm_ctxt->user_space_lock);

  return 0;
}

__init_text int _init_pd_table_mapper(void)
{
  addr_t *pd = (page_entry_t*) CURRENT_PD_VADDR;
  addr_t ppage;

  /* Set the bitmap to correct values */
  memset((char *) team_mm_contexts_bitmap,
	 ~0,
	 sizeof(team_mm_contexts_bitmap));

  /* Allocate page for the new PT, which will be used to maintain the
     list of the addresses of all PDs of all teams */
  ppage = get_physical_page(PHYS_PAGE_KERNEL, PHYS_PAGE_NON_SWAPPABLE);
  CONCEPTION_ASSERT(ppage != 0);

  /* Make this PT available */
  pd[PD_TABLE_MAPPING_INDEX]  = ppage;
  pd[PD_TABLE_MAPPING_INDEX] |= (PAGE_RW | PAGE_SUPERVISOR | PAGE_PRESENT);
  pd_table_mapper = (page_entry_t *) (CURRENT_PT_AREA_START
				      + PD_TABLE_MAPPING_INDEX*PAGE_SIZE);
  memset((void *) pd_table_mapper, 0, PAGE_SIZE);

  return 0;
}
