/*
 * Copyright (C) 2000, Thomas PETAZZONI
 * http://kos.enix.org
 *
 * Specific part of memory management for x86 architecture : paging.
 * 
 * @(#) $Id: _vmap.c,v 1.1.1.1 2003/04/27 12:49:49 thomas Exp $
 */

#include <kos/types.h>
#include <kos/assert.h>
#include <lib/string.h>
#include <lib/stdio.h>

#include <loader/mod.h>
#include <pmm/pmm.h>
#include <debug/debug.h>
#include <kos/macros.h>
#include <kos/spinlock.h>
#include <vmm/vmm.h>
#include <liblist/liblist.h>

#include "_mm.h"

/* For opcode 'invlpg', operand must be a virtual memory address */
#define invlpg(vaddr) \
  do { \
       __asm__ __volatile__("invlpg %0"::"m"(*(unsigned *)vaddr)); \
  } while(0)

#define flush_tlb() \
  do { \
	unsigned long tmpreg; \
       __asm__ __volatile__("movl %%cr3,%0\n\tmovl %0,%%cr3" :"=r" (tmpreg) : :"memory"); \
  } while (0)

#define _translate_vm_entry(phys_addr,access_rights) \
            (((phys_addr) & PAGE_ADDR_MASK) | ((access_rights) & PAGE_FLAGS_MASK))

spinlock_t arch_remap_tmp_page;

/* Translate the access_right_t at vaddr into an x86 PT access flag */
inline static k_ui32_t
_translate_access_rights(vaddr_t vaddr,
			 access_right_t access_right)
{
  k_ui32_t flags = 0;

  if (IS_KERNEL_PAGE(vaddr))
    flags |= PAGE_SUPERVISOR;
  else
    flags |= PAGE_USER;

  if (access_right | VM_ACCESS_WRITE)
    flags |= PAGE_RW;
  else
    flags |= PAGE_RO;

  return flags;
}

inline static team_mm_context_t*
_setup_mm_ctxt(/*in*/team_mm_context_t* dest_mm_ctxt,
	       /*out*/int*    pd_index_in_pd_table,
	       /*out*/bool_t* is_foreign_mm_context)
{
  // TBI("Maybe race conditions: make sure we are modifying address space of a not-to-be-deleted team, because this setup takes place outside any lock");

  *is_foreign_mm_context = FALSE;

  if(dest_mm_ctxt == NULL) {
    /* In the early phases of the initialization, there's no team */
    if (get_current_thread())
      {
	dest_mm_ctxt          = & get_current_team()->mm_context;
	*pd_index_in_pd_table = dest_mm_ctxt->pd_index_in_pd_table;
      }
    else
      *pd_index_in_pd_table   = PRIMARY_TEAM_PD_ID;
  } else {
    *pd_index_in_pd_table     = dest_mm_ctxt->pd_index_in_pd_table;
    *is_foreign_mm_context    = TRUE;
  }

  return dest_mm_ctxt;
}

/* All PD's corresponding to the kernel virtual space (4k-2G) must be
   synchronized between all teams. This function forces synchronisation
   of the PDE which has been modified. This function MUST be called
   with kernel_space_lock locked */

static void _sync_thread_pd_unsafe(int index_in_pd, page_entry_t pde)
{
  k_ui32_t flags;
  team_t *team, *current_team;
  thread_t *thread;
  int nb;
  page_entry_t *pde_to_update;

  thread = get_current_thread();
  if(thread == NULL)
    return;

  current_team = thread->mother_team;

  read_spin_lock(kernel_team_list_lock, flags);
  list_foreach(current_team, team, nb)
    {
      pde_to_update = VADDR_PDE_ANY_TEAM(team->mm_context.pd_index_in_pd_table,
					 index_in_pd);
      *pde_to_update = pde;
    }
  read_spin_unlock(kernel_team_list_lock, flags);

  return;
}

inline static int _map_foreign_pd_unsafe(team_mm_context_t* dest_mm_ctxt)
{
  page_entry_t* my_pd;

  CONCEPTION_ASSERT(dest_mm_ctxt);

  my_pd = (page_entry_t*)CURRENT_PD_VADDR;

  if ((!IS_UNMAPPED(CURRENT_PDE(FOREIGN_PT_AREA_INDEX)))
      && (CURRENT_PDE_PADDR(FOREIGN_PT_AREA_INDEX) == dest_mm_ctxt->pd))
    {
      /* Foreign PTs are mapped onto the FOREIGN_PT_AREA, and they
         appear to be the PTs that are wanted */
      return 0;
    }

  /* Make sure we are mapping the correct foreign PD */
  CONCEPTION_ASSERT((PD_ANY_TEAM(dest_mm_ctxt->pd_index_in_pd_table)
		     & PAGE_ADDR_MASK) == dest_mm_ctxt->pd);

  /* Actually map the foreign PD */
  my_pd[FOREIGN_PT_AREA_INDEX]
    = PD_ANY_TEAM(dest_mm_ctxt->pd_index_in_pd_table);
  invlpg((vaddr_t)FOREIGN_PD_VADDR);

  /* WARNING: the individual TLB entries for the PT in the
     FOREIGN_AREA are not invalidated, and must be updated
     individually */

  return 0;
}


inline static int _map_virtual_page_unsafe(struct rmap_session_s *rmap_session,
					   int pd_index_in_pd_table,
					   bool_t is_foreign_mm_context,
					   team_mm_context_t* dest_mm_ctxt,
					   vaddr_t virt, addr_t phys,
					   access_right_t access_rights)
{
  gpfme_t* gpfme;
  page_entry_t *pd;
  page_entry_t *pt;
  k_ui32_t gpfm_lock_flags;

  if (is_foreign_mm_context && !IS_KERNEL_PAGE(virt))
    {
      _map_foreign_pd_unsafe(dest_mm_ctxt);
      pd = (page_entry_t*)FOREIGN_PD_VADDR;
      pt = (page_entry_t*)(FOREIGN_PT_AREA_START + va_to_pde(virt)*PAGE_SIZE);
      invlpg((vaddr_t)pt);
    }
  else
    {
      pd = (page_entry_t*)CURRENT_PD_VADDR;
      pt = (page_entry_t*)(CURRENT_PT_AREA_START + va_to_pde(virt)*PAGE_SIZE);
    }

  /* No PT yet */
  if(IS_UNMAPPED(pd[va_to_pde(virt)]))
    {
      paddr_t new_pt;
      k_ui32_t pde_flags;

      /* Deadlock warning : get_physical_page acquire on GPFME lists */
      new_pt = get_physical_page(PHYS_PAGE_KERNEL, PHYS_PAGE_NON_SWAPPABLE);
      CONCEPTION_ASSERT(new_pt);
      
      gpfme = get_gpfme_at_phys_addr(new_pt, & gpfm_lock_flags);
      CONCEPTION_ASSERT(gpfme);

      /* Add the rmap to this PT */
      _rmap_ops.do_add_rmapping(rmap_session, pd_index_in_pd_table, gpfme,
				(vaddr_t)pt, TRUE);
      gpfme_unlock(gpfme, & gpfm_lock_flags);

      /* Init the PT flags */
      pde_flags = PAGE_RW | PAGE_PRESENT;
      if (! IS_KERNEL_PAGE(virt))
	pde_flags |= (PAGE_USER);

      /* Update the address space */
      pd[va_to_pde(virt)] = _translate_vm_entry(new_pt, pde_flags);
      
      /* Update local TLB */
      // TBI("Propagate this TLB update to the other processors, if they are executing a thread in the same team");
      invlpg(pt);

      /* Reset this new PT */
      memset((void*)pt, 0x0, PAGE_SIZE);

      /* Update the other teams' address space when this new PT is a
         kernel PT */
      if (IS_KERNEL_PAGE(virt))
	{
	  /* Deadlock warning : _syn_thread_pd acquires lock on team list */
	  _sync_thread_pd_unsafe(va_to_pde(virt), pd[va_to_pde(virt)]);
	}
    }

  gpfme = get_gpfme_at_phys_addr(phys, & gpfm_lock_flags);
  CONCEPTION_ASSERT(gpfme);

  /* Add the rmap to this page. Any needed rmap structure are already
     available in rmap_session => no eager & egg problem
     kslab/vmap/kslab */
  _rmap_ops.do_add_rmapping(rmap_session, pd_index_in_pd_table,
			    gpfme, virt, FALSE);
  gpfme_unlock(gpfme, & gpfm_lock_flags);
      
  /* Effective mapping of the page */
  /* Accessing PT might generate a #PF, but we are sure that the PT is
     swapped. And in this case, we dare (aka we hope) assume that the
     #PF handler will handle this. BUT this seems to raise a lot of
     questions regarding synchronization with #PF mapping this PT
     (same spinlock locked as here). */
  pt[va_to_pte(virt)] =
    _translate_vm_entry(phys,
			_translate_access_rights(virt, access_rights)
			| PAGE_PRESENT);

  /* Update local TLB */
  // TBI("Propagate this TLB update to the other processors, if they
  // are executing a thread in the same team");
  invlpg(virt);

  return 0;
}



/*
 * When dest_mm_ctxt == NULL, then map virtual page in current team's space
 * virt AND phys MUST be page-aligned !!!
 * Returns -1 when virt is not owned by any virtual region of the current team
 */ 
int arch_map_virtual_page(team_mm_context_t* dest_mm_ctxt,
			  vaddr_t virt, paddr_t phys,
			  access_right_t access_rights)
{
  int pd_index_in_pd_table;
  struct rmap_session_s rmap_session;
  bool_t is_foreign_mm_context;
  k_ui32_t as_lock_flags;
  int ret;

  dest_mm_ctxt = _setup_mm_ctxt(dest_mm_ctxt, & pd_index_in_pd_table,
				& is_foreign_mm_context);

  /* Reserve rmap structures for the page and (maybe) the new PT */
  _rmap_ops.pre_add_rmapping(& rmap_session);

  /* Acquire lock */
  if(IS_KERNEL_PAGE(virt))
    {
      /* get lock for kernel space (needs to be BEFORE the test to
	 make sure another thread to not map this PT between the test
	 and the real mapping of the PT) */
      write_spin_lock(kernel_space_lock, as_lock_flags);
    }
  else
    {
      /* Cannot map to user space during init ! (no real dest_mm_ctxt) */
      CONCEPTION_ASSERT(get_current_thread());

      // get lock for this user space
      write_spin_lock(dest_mm_ctxt->user_space_lock, as_lock_flags);
    }

  ret = _map_virtual_page_unsafe(& rmap_session, 
				 pd_index_in_pd_table,
				 is_foreign_mm_context,
				 dest_mm_ctxt,
				 virt, phys, access_rights);

  /* Release lock */
  if (IS_KERNEL_PAGE(virt))
    {
      // release lock for kernel space
      write_spin_unlock(kernel_space_lock, as_lock_flags);
    }
  else
    {
      // release lock for this user space
      write_spin_unlock(dest_mm_ctxt->user_space_lock, as_lock_flags);
    }

  /* Release any unused pre-allocated rmapping */
  _rmap_ops.post_add_rmapping(& rmap_session);
  
  return ret;
}

inline page_entry_t _get_pte_unsafe(vaddr_t virt)
{
  page_entry_t *pd;
  page_entry_t *pt;

  pd = (page_entry_t*) CURRENT_PD_VADDR;
  if(IS_UNMAPPED(pd[va_to_pde(virt)]))
    return 0;
  
  pt = (page_entry_t*) (CURRENT_PT_AREA_START + va_to_pde(virt)*PAGE_SIZE);
  if(IS_UNMAPPED(pt[va_to_pte(virt)]))
    {
      return 0;
    }

  /* Accessing PT might generate a #PF, but we are sure that the PT is
     swapped. And in this case, we dare (aka we hope) assume that the
     #PF handler will handle this. BUT this seems to raise a lot of
     questions regarding synchronization with #PF mapping this PT
     (same spinlock locked as here). */
  return pt[va_to_pte(virt)];
}


/* Returns 0 when virt_addr isn't mapped */
paddr_t arch_get_paddr_at_vaddr(vaddr_t virt)
{
  k_ui32_t flags;

  if(IS_KERNEL_PAGE(virt))
    {
      page_entry_t pte;
      read_spin_lock(kernel_space_lock, flags);
      pte = _get_pte_unsafe(virt);
      read_spin_unlock(kernel_space_lock, flags);
      return (pte & PAGE_ADDR_MASK);
    }
  else
    {
      team_mm_context_t *mm_ctxt;
      page_entry_t pte;

      if(get_current_thread())
	mm_ctxt = & get_current_team()->mm_context;
      else
	FAILED_VERBOSE("Not supposed to access User space during Init !\n");

      /* We're not supposed to access user space during init (no real
         mm_ctxt) */
      CONCEPTION_ASSERT(mm_ctxt);
      read_spin_lock(mm_ctxt->user_space_lock, flags);
      pte = _get_pte_unsafe(virt);
      read_spin_unlock(mm_ctxt->user_space_lock, flags);
      return (pte & PAGE_ADDR_MASK);
    }
  
  return 0;
}


static inline vpage_status_t
_get_vpage_status_unsafe(team_mm_context_t* dest_mm_ctxt,
			 vaddr_t vaddr)
{
  page_entry_t *pd;
  page_entry_t *pt;
  int    pd_index_in_pd_table;
  bool_t is_foreign_mm_context;

  dest_mm_ctxt = _setup_mm_ctxt(dest_mm_ctxt, & pd_index_in_pd_table,
				& is_foreign_mm_context);

  if (is_foreign_mm_context && !IS_KERNEL_PAGE(vaddr))
    {
      _map_foreign_pd_unsafe(dest_mm_ctxt);
      pd = (page_entry_t*)FOREIGN_PD_VADDR;
      pt = (page_entry_t*)(FOREIGN_PT_AREA_START + va_to_pde(vaddr)*PAGE_SIZE);
      invlpg((vaddr_t)pt);
    }
  else
    {
      pd = (page_entry_t*)CURRENT_PD_VADDR;
      pt = (page_entry_t*)(CURRENT_PT_AREA_START + va_to_pde(vaddr)*PAGE_SIZE);
    }

  /* If PD not present AND not swapped => page is unmapped */
  if (IS_UNMAPPED(pd[va_to_pde(vaddr)]))
    return PHYS_PAGE_UNMAPPED;

  /* Accessing PT might generate a #PF, but we are sure that the PT is
     swapped. And in this case, we dare (aka we hope) assume that the
     #PF handler will handle this. BUT this seems to raise a lot of
     questions regarding synchronization with #PF mapping this PT
     (same spinlock locked as here). */
  if(!IS_PRESENT(pt[va_to_pte(vaddr)]))
    {
      if(IS_SWAPPED(pt[va_to_pte(vaddr)]))
	return PHYS_PAGE_SWAPPED;
      return PHYS_PAGE_UNMAPPED;
    }

  return PHYS_PAGE_PRESENT;
}


vpage_status_t arch_get_vpage_status(team_mm_context_t* dest_mm_ctxt,
				     vaddr_t vaddr)
{ 
  k_ui32_t flags;
  bool_t is_foreign_mm_context;
  int pd_index_in_pd_table;  

  dest_mm_ctxt = _setup_mm_ctxt(dest_mm_ctxt, & pd_index_in_pd_table,
				& is_foreign_mm_context);

  if(IS_KERNEL_PAGE(vaddr))
    {
      vpage_status_t vpage_status;
      read_spin_lock(kernel_space_lock, flags);
      vpage_status = _get_vpage_status_unsafe(dest_mm_ctxt, vaddr);
      read_spin_unlock(kernel_space_lock, flags);
      return vpage_status;
    }
  else
    {
      vpage_status_t vpage_status;
      /* Cannot access user space during init */
      CONCEPTION_ASSERT(dest_mm_ctxt);
      read_spin_lock(dest_mm_ctxt->user_space_lock, flags);
      vpage_status = _get_vpage_status_unsafe(dest_mm_ctxt, vaddr);
      read_spin_unlock(dest_mm_ctxt->user_space_lock, flags);
      return vpage_status;
    }
  return 0;
}


/* Return paddr of page to put_physical_page() or NULL */
inline static paddr_t
_unmap_virtual_page_unsafe(struct rmap_session_s *rmap_session,
			   int pd_index_in_pd_table,
			   bool_t is_foreign_mm_context,
			   team_mm_context_t* dest_mm_ctxt,
			   vaddr_t vaddr)
{
  paddr_t ret_paddr = 0;
  page_entry_t *pt;
  page_entry_t *pd;

  if (is_foreign_mm_context && !IS_KERNEL_PAGE(vaddr))
    {
      _map_foreign_pd_unsafe(dest_mm_ctxt);
      pd = (page_entry_t*)FOREIGN_PD_VADDR;
      pt = (page_entry_t*)(FOREIGN_PT_AREA_START + va_to_pde(vaddr)*PAGE_SIZE);
      invlpg((vaddr_t)pt);
    }
  else
    {
      pd = (page_entry_t*)CURRENT_PD_VADDR;
      pt = (page_entry_t*)(CURRENT_PT_AREA_START + va_to_pde(vaddr)*PAGE_SIZE);
    }

  if (IS_UNMAPPED(pd[va_to_pde(vaddr)]))
    {
      return 0;
    }

  /* Page is physically present in RAM */
  if (IS_PRESENT(pt[va_to_pte(vaddr)]))
    {
      gpfme_t *gpfme;
      k_ui32_t gpfm_lock_flags;
      paddr_t paddr;

      /* Get the phys addr and gpfme for the page */
      paddr = pt[va_to_pte(vaddr)] & PAGE_ADDR_MASK;
      
      pt[va_to_pte(vaddr)] = 0;
      /* We only invalidate this page entry in the cache, and not all
	 entries in the TLB */
      // TBI("Propagate this TLB update to the other processors, if they are executing a thread in the same team");
      invlpg(vaddr);

      gpfme = get_gpfme_at_phys_addr(paddr, & gpfm_lock_flags);
      CONCEPTION_ASSERT(gpfme);

      /* rmap ranges to be deleted are be marked "to be deleted", and
         will actually be released outside of the lock in the unmap
         function => no eager & egg problem kslab/vmap/kslab */
      _rmap_ops.do_del_rmapping(rmap_session, pd_index_in_pd_table,
				gpfme, vaddr);

      if (gpfme->ref_cnt == 0)
	ret_paddr = paddr;

      gpfme_unlock(gpfme, & gpfm_lock_flags);
    }
  else
    {
      TBI("Unmap a swapped page\n");
      return 0;
    }

  return ret_paddr;
}


int arch_unmap_virtual_page(team_mm_context_t* dest_mm_ctxt,
			    vaddr_t vaddr)
{
  paddr_t paddr;
  int pd_index_in_pd_table;
  k_ui32_t flags;
  bool_t is_foreign_mm_context;
  struct rmap_session_s rmap_session;

  dest_mm_ctxt = _setup_mm_ctxt(dest_mm_ctxt, & pd_index_in_pd_table,
				& is_foreign_mm_context);

  /* Init rmap session: rmap range to be deleted will be marked "to be
     deleted by the _unmap below, and actually released outside of the
     lock */
  _rmap_ops.pre_del_rmapping(& rmap_session);

  /* Get the lock */
  if(IS_KERNEL_PAGE(vaddr))
    {
      write_spin_lock(kernel_space_lock, flags);
    }
  else
    {
      /* Cannot access user space during initizalition ! */
      CONCEPTION_ASSERT(dest_mm_ctxt);      
      write_spin_lock(dest_mm_ctxt->user_space_lock, flags);
    }

  /* Actually unmap the page */
  paddr = _unmap_virtual_page_unsafe(& rmap_session, pd_index_in_pd_table,
				     is_foreign_mm_context,
				     dest_mm_ctxt, vaddr);

  /* Release the lock */
  if(IS_KERNEL_PAGE(vaddr))
    {
      write_spin_unlock(kernel_space_lock, flags);
    }
  else
    {
      write_spin_unlock(dest_mm_ctxt->user_space_lock, flags);
    }
  
  /* Delete any rmapping marked "to be deleted" */
  _rmap_ops.post_del_rmapping(& rmap_session);
  if (paddr != 0)
    put_physical_page(paddr);

  return 0;
}

int arch_remap_virtual_page(team_mm_context_t *dest_mm_ctxt,
			    vaddr_t vaddr, paddr_t to)
{
  int    pd_index_in_pd_table;
  bool_t is_foreign_mm_context;
  addr_t *pd;
  addr_t *pt;
  addr_t *pt_tmp;
  addr_t tmp_map_addr = CORE_KERNEL_VIRTUAL_ADDR - _get_gpfm_ram_map_size() - PAGE_SIZE;
  gpfme_t *gpfme_from;
  gpfme_t *gpfme_to;
  k_ui32_t as_flags;
  k_ui32_t tmp_flags;
  k_ui32_t gpfm_flags = 0;
  struct rmap_session_s add_session, del_session;
  
  dest_mm_ctxt = _setup_mm_ctxt(dest_mm_ctxt, & pd_index_in_pd_table,
				& is_foreign_mm_context);

  _rmap_ops.pre_del_rmapping(& del_session);  
  _rmap_ops.pre_add_rmapping(& add_session);  

  write_spin_lock(arch_remap_tmp_page, tmp_flags);
  
  if(IS_KERNEL_PAGE(vaddr))
    {
      write_spin_lock(kernel_space_lock, as_flags);
    }
  else
    {
      /* Cannot access user space during initizalition ! */
      CONCEPTION_ASSERT(dest_mm_ctxt);      
      write_spin_lock(dest_mm_ctxt->user_space_lock, as_flags);
    }

  if (is_foreign_mm_context && !IS_KERNEL_PAGE(vaddr))
    {
      _map_foreign_pd_unsafe(dest_mm_ctxt);
      pd = (page_entry_t*)FOREIGN_PD_VADDR;
      pt = (page_entry_t*)(FOREIGN_PT_AREA_START + va_to_pde(vaddr)*PAGE_SIZE);
      invlpg((vaddr_t)pt);
    }
  else
    {
      pd = (page_entry_t*)CURRENT_PD_VADDR;
      pt = (page_entry_t*)(CURRENT_PT_AREA_START + va_to_pde(vaddr)*PAGE_SIZE);
    }


  /* A PT for this page must have been allocated during init */
  CONCEPTION_ASSERT(! IS_UNMAPPED(pd[va_to_pde(tmp_map_addr)]));

  gpfme_from = get_gpfme_at_phys_addr(_get_pte_unsafe(vaddr) & PAGE_ADDR_MASK,
				      & gpfm_flags);
  gpfme_unlock(gpfme_from, &gpfm_flags);
  
  gpfme_to   = get_gpfme_at_phys_addr(to, & gpfm_flags);
  gpfme_unlock(gpfme_to, &gpfm_flags);

  pt_tmp = (page_entry_t*) (CURRENT_PT_AREA_START +
			    va_to_pde(tmp_map_addr) * PAGE_SIZE);

  pt_tmp[va_to_pte(tmp_map_addr)] = to | PAGE_SUPERVISOR | PAGE_RW | PAGE_PRESENT;

  memcpy((void *) tmp_map_addr, (void *) vaddr, PAGE_SIZE);
  
  pt[va_to_pte(vaddr)] = (to & PAGE_ADDR_MASK) | 
    (pt[va_to_pte(vaddr)] & PAGE_FLAGS_MASK);
  invlpg(vaddr);
  
  _rmap_ops.do_del_rmapping(& del_session, pd_index_in_pd_table, 
			    gpfme_from, vaddr);  
  _rmap_ops.do_add_rmapping(& add_session, pd_index_in_pd_table, 
			    gpfme_to, vaddr, FALSE);  
  
  if(IS_KERNEL_PAGE(vaddr))
    {
      write_spin_unlock(kernel_space_lock, as_flags);
    }
  else
    {
      /* Cannot access user space during initizalition ! */
      CONCEPTION_ASSERT(dest_mm_ctxt);      
      write_spin_unlock(dest_mm_ctxt->user_space_lock, as_flags);
    }

  write_spin_unlock(arch_remap_tmp_page, tmp_flags);
  
  _rmap_ops.post_del_rmapping(& del_session);  
  _rmap_ops.post_add_rmapping(& add_session);  

  if(gpfme_from->ref_cnt == 0)
    {
      __dbg_printk("page is not used anymore => free it!\n");
      put_physical_page(gpfme_from->address);
    }
  
  return 0;
}

inline static int _protect_virtual_page_unsafe(team_mm_context_t *dest_mm_ctxt,
					       bool_t is_foreign_mm_context,
					       vaddr_t vaddr,
					       access_right_t access_rights)
{
  addr_t *pd, *pt;

  if (is_foreign_mm_context && !IS_KERNEL_PAGE(vaddr))
    {
      _map_foreign_pd_unsafe(dest_mm_ctxt);
      pd = (page_entry_t*)FOREIGN_PD_VADDR;
      pt = (page_entry_t*)(FOREIGN_PT_AREA_START + va_to_pde(vaddr)*PAGE_SIZE);
      invlpg((vaddr_t)pt);
    }
  else
    {
      pd = (page_entry_t*)CURRENT_PD_VADDR;
      pt = (page_entry_t*)(CURRENT_PT_AREA_START + va_to_pde(vaddr)*PAGE_SIZE);
    }

  if(IS_UNMAPPED(pd[va_to_pde(vaddr)]))
    return -1;

  if(IS_UNMAPPED(pt[va_to_pte(vaddr)]))
    return -1;
  
  pt[va_to_pte(vaddr)] = 
    (pt[va_to_pte(vaddr)] & (~PAGE_RW)) |
    (_translate_access_rights(vaddr, access_rights) & (PAGE_RW));

  if(!is_foreign_mm_context)
    invlpg(vaddr);

  return 0;
}


/* Change the flags for the (mapped) vpage */
int arch_protect_virtual_page(team_mm_context_t* dest_mm_ctxt,
			      vaddr_t vaddr,
			      access_right_t access_rights)
{
  int ret;
  k_ui32_t as_flags;
  bool_t is_foreign_mm_context;
  k_ui32_t pd_index_in_pd_table;

  dest_mm_ctxt = _setup_mm_ctxt(dest_mm_ctxt, & pd_index_in_pd_table,
				& is_foreign_mm_context);
  
  if(IS_KERNEL_PAGE(vaddr))
    {
      write_spin_lock(kernel_space_lock, as_flags);
    }
  else
    {
      /* Cannot access user space during initizalition ! */
      CONCEPTION_ASSERT(dest_mm_ctxt);      
      write_spin_lock(dest_mm_ctxt->user_space_lock, as_flags);
    }

  ret = _protect_virtual_page_unsafe(dest_mm_ctxt, is_foreign_mm_context,
				     vaddr, access_rights);

  if(IS_KERNEL_PAGE(vaddr))
    {
      write_spin_unlock(kernel_space_lock, as_flags);
    }
  else
    {
      /* Cannot access user space during initizalition ! */
      CONCEPTION_ASSERT(dest_mm_ctxt);      
      write_spin_unlock(dest_mm_ctxt->user_space_lock, as_flags);
    }
  
  return ret;
}



__init_text int _remap_init(void)
{
  addr_t tmp_map_addr = CORE_KERNEL_VIRTUAL_ADDR - _get_gpfm_ram_map_size();
  paddr_t paddr;
  addr_t *pd, *pt;

  pd = (page_entry_t*)CURRENT_PD_VADDR;
  pt = (page_entry_t*)(CURRENT_PT_AREA_START + va_to_pde(tmp_map_addr)*PAGE_SIZE);

  if(IS_UNMAPPED(pd[va_to_pde(tmp_map_addr)]))
    {
      paddr = get_physical_page(PHYS_PAGE_KERNEL, PHYS_PAGE_NON_SWAPPABLE);
      pd[va_to_pde(tmp_map_addr)] = _translate_vm_entry(paddr, PAGE_SUPERVISOR | PAGE_RW | PAGE_PRESENT);
    }

  spinlock_init(arch_remap_tmp_page);
  
  return 0;
}
